#ifndef __PROGRAM_CC__
#define __PROGRAM_CC__

#include <iostream>
#include <iterator>
#include <sstream>
#include <fstream>
#include <set>
#include <sys/time.h>

#include "Program.h"
#include "Application.h"
#include "Exception.h"
#include "Output.h"
#include "Checker.h"
#include "Result.h"

#include "Value.h"
#include "ModuleGeneric.h"
#include "ModuleLocal.h"
#include "ModuleBelief.h"
#include "ModuleAction.h"
#include "ModuleProblem.h"

#include "modules/Plugin.h"
#include "modules/Math.h"
#include "modules/Bel.h"

#include "Logic.h"

#define KEY_UP 72
#define KEY_DOWN 80
#define KEY_LEFT 75
#define KEY_RIGHT 77
#define KEY_ENTER 13

namespace Inter4ql {

	Application* Application::app_instance;

	Application::Application(std::istream *i, std::ostream *o, std::ostream *e) {
		std::cout << "\tInterpreter for 4ql, version 5.2, http://www.4ql.org/" << std::endl;
		this->output = new Inter4ql::Output(o, e);
		if (i)
			this->set_input(i);
		else
			this->set_input(&std::cin);
		this->modules = new std::vector<ModuleGeneric *>();
		this->actions = new std::vector<ModuleGeneric *>();
		this->problems = new std::vector<ModuleGeneric *>();
		this->checker = new Checker(this->modules);
		/* hardcoded modules */
		this->variable_space = new VariableSpace();
		this->modules->push_back(static_cast<ModuleGeneric *>(new Plugin(this->variable_space)));
		this->modules->push_back(static_cast<ModuleGeneric *>(new Math(this->variable_space)));
		this->modules->push_back(static_cast<ModuleGeneric *>(new Bel(this->variable_space, this->modules)));

		Application::app_instance = this;
	}

	Application::~Application() {
		delete this->output;
		for(std::vector<ModuleGeneric*>::iterator it = this->modules->begin(); it != this->modules->end(); it++) {
			if((*it)->get_module_type() == module_type::LOCAL) {
				delete (ModuleLocal*) (*it);
			} else
			if((*it)->get_module_type() == module_type::BEL) {
				delete (Bel*) (*it);
			} else
			if((*it)->get_module_type() == module_type::BELIEF) {
				delete (ModuleBelief*) (*it);
			} else
			if((*it)->get_module_type() == module_type::EXTERNAL) {
				delete (ModuleExternal*) (*it);
			} else
			if((*it)->get_module_type() == module_type::INTERNAL) {
				delete (ModuleInternal*) (*it);
			}
		}
		for(std::vector<ModuleGeneric*>::iterator it = this->actions->begin(); it != this->actions->end(); it++) {
			delete (ModuleAction*) (*it);
		}
		for(std::vector<ModuleGeneric*>::iterator it = this->problems->begin(); it != this->problems->end(); it++) {
			delete (ModuleProblem*) (*it);
		}
		delete this->modules;
		delete this->actions;
		delete this->problems;
		delete this->checker;
		delete this->variable_space;
		std::cout << "Thanks for using!\n";
	}

	void Application::start() {
		while (this->parse_line());
		return;
	}

	void Application::add_to_variable_space(Program *p, VariableSpace* vs) {
		for (ModuleGeneric * mg : *p->get_modules()) {
			if (mg->get_module_type() != module_type::LOCAL)
				continue;

			ModuleLocal *m = static_cast<ModuleLocal*>(mg);
			std::vector<Relation*>* relations = m->get_relations();

			for (const auto& rule: * m->get_rules()) {
				this->add_to_variable_space(rule->get_head(), vs, relations);

				for (const auto& conjunction : * rule->get_body())
					for (Fact* curr : *conjunction) {
						std::vector<Relation*>* relationsBody = NULL;

						std::string target_module_name = curr->get_module();
						if(target_module_name != m->get_name()) {
							for (const auto& mod : *p->get_modules()) {
								if(mod->get_name() == target_module_name) {
									relationsBody = mod->get_relations();
									break;
								}
							}
							if(relationsBody == NULL) {
								for (const auto& mod : *this->modules) {
									if(mod->get_name() == target_module_name) {
										relationsBody = mod->get_relations();
										break;
									}
								}
							}
							if(relationsBody == NULL) {
								throw new Exception(std::string("Module "+target_module_name+" could not be found!\n"));
							}
						} else {
							relationsBody = relations;
						}
						this->add_to_variable_space(curr, vs, relationsBody);
					}
			}

			for (const auto& fact: *m->get_facts()) {
				this->add_to_variable_space(fact, vs, relations);
			}
		}
	}

	bool Application::add_to_variable_space(Fact *r, VariableSpace* vs, std::vector<Relation*>* relations) {
		bool did_change = false;
		if (r->get_value() == Inter4ql::logic_type::type::FALSE)
			return false;
		for (int i = 0; i < r->get_params()->size(); i++) {
			if (r->get_params()->at(i)->get_type() == variable_type::VARIABLE) {
				continue;
			}

			std::string domain = "";
			for(const auto& relation : *relations) {
				if(relation->get_name() == r->get_relation()) {
					domain = relation->get_params()->at(i)->get_name();
					break;
				}
			}
			if(domain == "") {
				throw new Exception(std::string("Relation "+r->get_relation()+" could not be found!\n"));
			}
			Value *v = r->get_params()->at(i);
			r->get_params()->at(i) = vs->get_pointer(v, r->get_module(), domain, did_change);
			if (v != r->get_params()->at(i))
				delete v;
		}
		return did_change;
	}

	int Application::parse_line() {
		try {
			std::string line = "";

			this->output->print(std::string("# "));

			std::getline(*this->input, line, '\n');
			if (this->input->eof()) {
				this->output->print(std::string("\n"));
				return 0;
			}
			if(line == "") return 1;

			std::stringstream *s = new std::stringstream(std::stringstream::in | std::stringstream::out);

			(*s) << line;
			Inter4ql::Parser *p = new Inter4ql::Parser(this->output);
			p->init(s);

			Inter4ql::parser_line *l = 0;
			l = p->parse();

			delete s;
			delete p;

			/* check if parsing went well */
			if (!l) return 1;

			/* here we line value */
			switch (l->type) {
				case line_type::IMPORT:
				{
					/* import file type of line */
					std::string s = l->value.substr(1, l->value.length() - 2);
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					int ret = process_import(s);
					if (ret) return ret;
					break;
				}
				case line_type::QUIT:
				{
					if(l->program) delete l->program;
					if(l->rule) delete l->rule;
					delete l;
					return 0;
				}
				case line_type::PRINT:
				{
					ModuleGeneric *m = this->get_any(l->value);
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					process_print(m);

					break;
				}
				case line_type::EXECUTE:
				{
					ModuleGeneric *m = this->get_problem(l->value);
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					process_execute(m);

					break;
				}
				case line_type::MODULES:
				{
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					process_modules();
					break;
				}
				case line_type::ACTIONS:
				{
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					process_actions();
					break;
				}
				case line_type::PROBLEMS:
				{
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					process_problems();
					break;
				}
				case line_type::RULE:
				{
					Expression *e = l->expression;
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					process_rule(e);

					break;
				}
				case line_type::CLEAR:
				{
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					process_clear();
					break;
				}
				case line_type::DEBUG:
				{
					std::string module = l->value;
					bool value = l->bool_value;
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					process_debug(module, value);
					break;
				}
				case line_type::HELP:
				{
					if(l->rule) delete l->rule;
					if(l->program) delete l->program;
					delete l;
					process_help();
					break;
				}
				default:
				{
					break;
				}
			}
		} catch (Inter4ql::Exception *s) {
			this->output->print(s);
			delete s;
			return 1;
		}
		return 1;
	}

	int Application::process_import(std::string filename)
	{
		std::ifstream *file = new std::ifstream(filename.c_str());
		if (!file->is_open()) {
			delete file;

			throw new Exception(std::string("Cannot open file '") + filename.c_str() + "'.\n");
		}

		Inter4ql::Parser *p = new Inter4ql::Parser(this->output, file);
		Inter4ql::parser_line *line = 0;
		line = p->parse();
		delete p;

		file->close();
		delete file;

		if (line) {
			if (line->type != line_type::PROGRAM) {
				if(line->rule) delete line->rule;
				if(line->program) delete line->program;
				delete line;

				throw new Exception("Imported file must be a valid 4ql source code\n");
			}
		} else {
			return 1;
		}

		/* we have line->program - a valid program */
		Program *program = line->program;
		line->program = 0;

		if(line->rule) delete line->rule;
		if(line->program) delete line->program;
		delete line;

		try {
			this->checker->check(program);
		} catch (Exception *ex) {
			VariableSpace* temp = new VariableSpace();
			this->add_to_variable_space(program, temp);

			for(std::vector<ModuleGeneric*>::iterator it = program->get_modules()->begin(); it != program->get_modules()->end(); it++) {
				delete (*it);
			}

			delete temp;
			delete program->get_modules();
			delete program;

			throw;
		}

		this->add_to_variable_space(program, this->variable_space);

		// reasoning part
		if (program->get_order()) {
			for (const auto& module : *program->get_order()) {
				module->set_disable_constraints(true);
				//printf("Reasoning about module %s\n", module->get_name().c_str());
				module->reason();
				if (module->get_module_type() == module_type::BELIEF )
					module->set_modules(this->modules);
				if (module->get_module_type() != module_type::ACTION && module->get_module_type() != module_type::PROBLEM)
					this->modules->push_back(module);
				module->set_disable_constraints(false);
			}

			delete program->get_order();

			for(std::vector<ModuleGeneric*>::iterator it = program->get_modules()->begin(); it != program->get_modules()->end(); it++) {
				if ((*it)->get_module_type() == module_type::ACTION) {
					this->actions->push_back(*it);
				}
				if ((*it)->get_module_type() == module_type::PROBLEM) {
					this->problems->push_back(*it);
				}
			}
			for (auto problem : *this->problems) {
				ModuleProblem* pm = dynamic_cast<ModuleProblem*>(problem);
				for (auto action : *this->actions) {
					if (pm->contains_action(action->get_name()))
						pm->add_action_module(action);
				}
			}
		} else {
				for (int i = 0; i < program->get_modules()->size(); i++) {
					delete program->get_modules()->at(i);
				}
			delete program->get_modules();
			delete program;

			throw new Exception(std::string("No reasoning order specified!\n"));
		}

		delete program->get_modules();
		delete program;
		this->output->print("Program loaded!\n");
		return 0;
	}

	void Application::process_print(ModuleGeneric* module)
	{
		this->output->print(module->print());
	}

	void Application::process_help()
	{
		this->output->print(std::string("Available commands:\n"));
		this->output->print(std::string("--------------------------\n"));
		this->output->print(std::string("import \"FILE PATH\". - imports modules from a specified file\n"));
		this->output->print(std::string("RULE. - executes specified rule\n"));
		this->output->print(std::string("modules. - shows a list of available modules\n"));
		this->output->print(std::string("actions. - shows a list of available actions\n"));
		this->output->print(std::string("problems. - shows a list of available problems\n"));
		this->output->print(std::string("print NAME. - prints a declaration of specified module/action/problem\n"));
		this->output->print(std::string("execute PROBLEM_NAME. - executes planning on specified problem\n"));
		this->output->print(std::string("clear. - restores interpreter to launch state\n"));
		this->output->print(std::string("debug. - shows status of all debug switches\n"));
		this->output->print(std::string("debug \"SWITCH\" enable/disable. - changes status of specified debug switch\n"));
		this->output->print(std::string("help. - shows this message\n"));
		this->output->print(std::string("exit. - exits the interpereter\n"));
	}

	void Application::process_debug(std::string module, bool value)
	{
		if (module == "") {
			this->output->print(std::string("Debug switches:\n"));
			this->output->print(std::string("--------------------------\n"));
			this->output->print(std::string("evaluation: "));
			this->output->print(std::string((debug_evaluation) ? "enabled" : "disabled"));
			this->output->print(std::string("\n"));

			this->output->print(std::string("planning: "));
			this->output->print(std::string((debug_planning) ? "enabled" : "disabled"));
			this->output->print(std::string("\n"));

			this->output->print(std::string("composite: "));
			this->output->print(std::string((debug_planning_composite) ? "enabled" : "disabled"));
			this->output->print(std::string("\n"));

			this->output->print(std::string("constraints: "));
			this->output->print(std::string((debug_constraints) ? "enabled" : "disabled"));
			this->output->print(std::string("\n"));

			this->output->print(std::string("disposable: "));
			this->output->print(std::string((debug_disposable) ? "enabled" : "disabled"));
			this->output->print(std::string("\n"));

			this->output->print(std::string("database: "));
			this->output->print(std::string((debug_database) ? "enabled" : "disabled"));
			this->output->print(std::string("\n"));

			this->output->print(std::string("probing: "));
			this->output->print(std::string((debug_probing) ? "enabled" : "disabled"));
			this->output->print(std::string("\n"));

			this->output->print(std::string("warning: "));
			this->output->print(std::string((debug_warning) ? "enabled" : "disabled"));
			this->output->print(std::string("\n"));
		} else {
			if (module == "\"evaluation\"") {
				debug_evaluation = value;
			} else if (module == "\"planning\"") {
				debug_planning = value;
			} else if (module == "\"composite\"") {
				debug_planning_composite = value;
			} else if (module == "\"constraints\"") {
				debug_constraints = value;
			} else if (module == "\"disposable\"") {
				debug_disposable = value;
			} else if (module == "\"database\"") {
				debug_database = value;
			} else if (module == "\"probing\"") {
				debug_probing = value;
			} else if (module == "\"warning\"") {
				debug_warning = value;
			} else {
				throw new Exception("Unknown debug switch provided!\n");
			}

			this->output->print(std::string("Debugging for switch "));
			this->output->print(module);
			this->output->print(std::string((value) ? " enabled!\n" : " disabled!\n"));
		}
	}


	void Application::process_modules()
	{
		this->output->print(std::string("List of available modules:\n"));
		this->output->print(std::string("--------------------------\n"));
		for(int i = 0; i < this->modules->size(); i++) {
			if (this->modules->at(i)->get_name() == "BEL") continue;
			this->output->print("- " + this->modules->at(i)->get_name() + "\n");
		}
	}

	void Application::process_actions()
	{
		this->output->print(std::string("List of available actions:\n"));
		this->output->print(std::string("--------------------------\n"));
		for(int i = 0; i < this->actions->size(); i++) {
			this->output->print("- " + this->actions->at(i)->get_name() + "\n");
		}
	}

	void Application::process_problems()
	{
		this->output->print(std::string("List of available problems:\n"));
		this->output->print(std::string("--------------------------\n"));
		for(int i = 0; i < this->problems->size(); i++) {
			this->output->print("- " + this->problems->at(i)->get_name() + "\n");
		}
	}

	void Application::process_execute(ModuleGeneric* problem)
	{
		this->output->print(std::string("Executing planning for problem ") + problem->get_name() + "...\n");

		problem->reason();

	}

	void Application::process_rule(Expression * e)
	{
		try {
			this->checker->check_expression(e);
		} catch (Inter4ql::Exception *ex) {
			delete e;
			throw;
		}

		Result result;
		int rating; // not used
		try {
			result = evaluate_expression_global(e, nullptr, this->modules, nullptr, rating);
		} catch (Exception* ex) {
			delete e;
			throw;
		}
		std::cout << "== Results ==\n";
		std::cout << result.print();
		delete e;
	}

	void Application::process_clear()
	{
		/* clear modules */
		std::vector<ModuleGeneric*>* new_m = new std::vector<ModuleGeneric*>();
		for(int mm = 0; mm < this->modules->size(); mm++){
			if(this->modules->at(mm)->get_module_type() == module_type::LOCAL
			|| this->modules->at(mm)->get_module_type() == module_type::GROUP
			|| this->modules->at(mm)->get_module_type() == module_type::BELIEF) {
				delete  this->modules->at(mm);
			} else {
				if (this->modules->at(mm)->get_name() == "BEL")
					this->modules->at(mm)->set_modules(new_m);
				new_m->push_back(this->modules->at(mm));
			}
		}
		delete this->modules;

		this->modules = new_m;

		for(int aa = 0; aa < this->actions->size(); aa++){
			delete this->actions->at(aa);
		}
		delete this->actions;
		this->actions = new std::vector<ModuleGeneric *>();

		for(int pp = 0; pp < this->problems->size(); pp++){
			delete this->problems->at(pp);
		}
		delete this->problems;
		this->problems = new std::vector<ModuleGeneric *>();

		/* clear variable space */
		this->variable_space->clear();

		debug_evaluation = false;
		debug_planning = false;
		debug_planning_composite = false;
		debug_constraints = false;
		debug_disposable = false;
		debug_database = false;
		debug_probing = false;

		this->checker->set_modules(this->modules);
		this->output->print("Modules cleared!\n");
	}

	ModuleGeneric *Application::get_module(std::string v) {
		for (int i = 0; i < this->modules->size(); i++) {
			if (this->modules->at(i)->get_name() == v) {
				return static_cast<ModuleGeneric *>(this->modules->at(i));
			}
		}
		throw new Exception(std::string("Module ") + v + " does not exist in program!\n");
	}

	ModuleGeneric *Application::get_problem(std::string v) {
		for (int i = 0; i < this->problems->size(); i++) {
			if (this->problems->at(i)->get_name() == v) {
				return static_cast<ModuleGeneric *>(this->problems->at(i));
			}
		}
		throw new Exception(std::string("Problem ") + v + " does not exist in program!\n");
	}

	ModuleGeneric *Application::get_action(std::string v) {
		for (int i = 0; i < this->actions->size(); i++) {
			if (this->actions->at(i)->get_name() == v) {
				return static_cast<ModuleGeneric *>(this->actions->at(i));
			}
		}
		throw new Exception(std::string("Action ") + v + " does not exist in program!\n");
	}

	ModuleGeneric *Application::get_any(std::string v) {
		for (int i = 0; i < this->modules->size(); i++) {
			if (this->modules->at(i)->get_name() == v) {
				return static_cast<ModuleGeneric *>(this->modules->at(i));
			}
		}
		for (int i = 0; i < this->actions->size(); i++) {
			if (this->actions->at(i)->get_name() == v) {
				return static_cast<ModuleGeneric *>(this->actions->at(i));
			}
		}
		for (int i = 0; i < this->problems->size(); i++) {
			if (this->problems->at(i)->get_name() == v) {
				return static_cast<ModuleGeneric *>(this->problems->at(i));
			}
		}
		throw new Exception(std::string("Module/action/problem ") + v + " does not exist in program!\n");
	}

	Relation *Application::get_relation_by_module(std::string module, std::string name) {
		for (int i = 0; i < this->modules->size(); i++)
			if (this->modules->at(i)->get_name() == module)
				return this->modules->at(i)->get_relation(name);

		return 0;
	}

	void Application::set_input(std::istream *i) {
		this->input = i;

		return;
	}

	Inter4ql::Output *Application::get_output() {
		return this->output;
	}

}

#endif
