#ifndef __APPLICATION_H__
#define __APPLICATION_H__

#include <set>

#include "VariableSpace.h"
#include "Parser.h"
#include "Output.h"
#include "Checker.h"
#include "Value.h"

namespace Inter4ql {

	class ModuleGeneric;
	class ExternalDeclaration;

	/** @brief Application class - class that merges all modules and creates interpreter
	 *
	 */
	class Application {
	public:
		/** @brief constructor used to build program and set input and outputs
		 *
		 * @param i default interpreter input (default is standard input)
		 * @param o default interpreter output (default is standard output)
		 * @param e default interpreter error pipe (default is standard error)
		 */
		Application(std::istream *i = 0, std::ostream *o = 0, std::ostream *e = 0);
		/** @brief destructor
		 */
		~Application();
		/** @brief main interpreter loop
		 */
		void start();
		/** @brief function that parses a line
		 *
		 * @return zero if it is the end of parsing, positive otherwise
		 */
		int parse_line();
		/** @brief getter for output module
		 *
		 * @return output module
		 */
		Output *get_output();
		/** @brief setter for default input
		 *
		 * @param i input stream
		 */
		void set_input(std::istream *i);
		/** @brief getter for global variable space
		 *
		 * @return variable space
		 */
		VariableSpace *public_get_vs() { return variable_space; };
		/** @brief getter for single module from global modules list
		 *
		 * @return module
		 */
		ModuleGeneric *public_get_module(std::string v) { return get_module(v); };
		/** @brief getter for all modules from global modules list
		 *
		 * @return module list
		 */
		std::vector<ModuleGeneric *> *public_get_modules() { return this->modules; };
		/** @brief getter for all actions from global action list
		 *
		 * @return actions list
		 */
		std::vector<ModuleGeneric *> *public_get_actions() { return this->actions; };

		static Application* app_instance; // this could be replaced by passing Application pointer to modules

		bool debug_evaluation = false;
		bool debug_planning = false;
		bool debug_planning_composite = false;
		bool debug_constraints = false;
		bool debug_disposable = false;
		bool debug_database = false;
		bool debug_probing = false;
		bool debug_warning = false;
	private:
		void add_to_variable_space(Program *p, VariableSpace* vs);
		bool add_to_variable_space(Fact *r, VariableSpace* vs, std::vector<Relation*>* relations);

		int process_import(std::string filename);
		void process_print(ModuleGeneric* module);
		void process_modules();
		void process_actions();
		void process_problems();
		void process_execute(ModuleGeneric* problem);
		void process_rule(Expression* e);
		void process_clear();
		void process_help();
		void process_debug(std::string module, bool value);

		ModuleGeneric *get_module(std::string v);
		ModuleGeneric *get_problem(std::string v);
		ModuleGeneric *get_action(std::string v);
		ModuleGeneric *get_any(std::string v);
		Relation *get_relation_by_module(std::string module, std::string name);
		void print_module(ModuleGeneric *m);

		std::vector<ModuleGeneric *> *modules;
		std::vector<ModuleGeneric *> *actions;
		std::vector<ModuleGeneric *> *problems;
		std::istream *input;
		Output *output;
		Checker *checker;
		VariableSpace *variable_space;
	};
}

#endif /* __APPLICATION_H__ */
