#include "Program.h"
#include "Checker.h"
#include "Exception.h"
#include "Domain.h"
#include "Relation.h"
#include "Rule.h"
#include "Functions.h"
#include "ModuleLocal.h"
#include "ModuleGroup.h"
#include "ModuleBelief.h"
#include "ModuleAction.h"
#include "ModuleProblem.h"
#include "Constraints.h"

#include <sstream>
#include <iostream>
#include <map>
#include <set>
#include <queue>

#define DEBUG_GRAPH false

namespace Inter4ql {

	Checker::Checker(std::vector<ModuleGeneric *> *_modules) {
		this->modules = _modules;
		this->program = 0;
	}

	void Checker::set_modules(std::vector<ModuleGeneric *> *_modules){
		this->modules = _modules;
	}

	void Checker::check(Program *p) {
		this->program = p;
		std::vector<std::string> s;
		std::vector<std::string> actions;
		for (int i = 0; i < p->get_modules()->size(); i++) {
			if ( ModuleBelief* mb = dynamic_cast<ModuleBelief*>(p->get_modules()->at(i))) {
				mb->set_modules(p->get_modules());
			}
		}
		for (int i = 0; i < p->get_modules()->size(); i++) {
			for (int j = 0; j < s.size(); j++)
				if (s[j] == p->get_modules()->at(i)->get_name())
					throw (new Exception(std::string("Module ") + s[j] + " declared more than once in program!\n"));
			s.push_back(p->get_modules()->at(i)->get_name());

			if (p->get_modules()->at(i)->get_module_type() == module_type::ACTION)
				actions.push_back(p->get_modules()->at(i)->get_name());
		}
		for (int i = 0; i < p->get_modules()->size(); i++) {
			if ( ModuleLocal* ml = dynamic_cast<ModuleLocal*>(p->get_modules()->at(i)) ) {
				this->check_module(ml);
			}

			if ( ModuleGroup* mg = dynamic_cast<ModuleGroup*>(p->get_modules()->at(i)) ) {
				this->check_module(mg);
			}

			if ( ModuleBelief* mb = dynamic_cast<ModuleBelief*>(p->get_modules()->at(i)) ) {
				this->check_module(mb);
			}

			if ( ModuleAction* ma = dynamic_cast<ModuleAction*>(p->get_modules()->at(i)) ) {
				this->check_module(ma);
			}

			if ( ModuleProblem* mp = dynamic_cast<ModuleProblem*>(p->get_modules()->at(i)) ) {
				if (mp->get_actions()->empty()) {
					mp->replace_actions(new std::vector<std::string>(actions.begin(), actions.end()));
				}
				this->check_module(mp);
			}
		}
		this->check_cycles();
		this->program = 0;
		return;
	}

	void Checker::check_cycles() {
		/* we need to check if module calls are good and create module reasoning order */
		std::vector<ModuleGeneric *> *order = new std::vector<ModuleGeneric *>();
		std::map<ModuleGeneric*, std::set<ModuleGeneric*>*> *graph = new std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>();
		std::queue<std::string> q;
		/* lets create module calls graph */
		for (ModuleGeneric* current_module : *this->program->get_modules()) {
			// actions and problems are not taken into account when checking cycles (as are not query-able)
			if (current_module->get_module_type() == module_type::ACTION ||
				current_module->get_module_type() == module_type::PROBLEM)
				continue;
			graph->insert(std::pair<ModuleGeneric*, std::set<ModuleGeneric*>*>(current_module, new std::set<ModuleGeneric*>()));

			if (current_module->get_module_type() == module_type::GROUP) {
				ModuleGroup *mg = static_cast<ModuleGroup*>(current_module);
				for (const auto& current_child : *mg->get_members()) {
					ModuleGeneric *m = 0;
					try {
						m = this->get_pointer_to_module(current_child, true);
					} catch (Exception *ex) {delete ex; continue;}
					if (m && (m->get_module_type() == module_type::LOCAL || m->get_module_type() == module_type::GROUP || m->get_module_type() == module_type::BELIEF)) {
						std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->find(current_module);
						it->second->insert(m);
					}
				}
			}

			if (current_module->get_module_type() == module_type::BELIEF) {
				ModuleBelief *mg = static_cast<ModuleBelief*>(current_module);
				for (const auto& current_child : *mg->get_members()) {
					ModuleGeneric *m = 0;
					try {
						m = this->get_pointer_to_module(current_child, true);
					} catch (Exception *ex) {
						delete ex;
						throw (new Exception(std::string("Module ") + current_child + " from " + mg->get_name() + "'s members list does not exist!\n"));
					}
					if (m && (m->get_module_type() == module_type::LOCAL || m->get_module_type() == module_type::GROUP || m->get_module_type() == module_type::BELIEF)) {
						std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->find(current_module);
						it->second->insert(m);
					}
				}

				Constraints* constr = mg->get_constraints();
				std::vector<Expression*>* soft_constraints = constr->get_soft_constraints();
				std::vector<Expression*>* hard_constraints = constr->get_hard_constraints();

				add_constraints_to_graph(current_module, "soft", soft_constraints, graph);
				add_constraints_to_graph(current_module, "hard", hard_constraints, graph);
			}

			if (current_module->get_module_type() == module_type::LOCAL) {
				ModuleLocal *ml = static_cast<ModuleLocal*>(current_module);

				Constraints* constr = ml->get_constraints();
				std::vector<Expression*>* soft_constraints = constr->get_soft_constraints();
				std::vector<Expression*>* hard_constraints = constr->get_hard_constraints();

				add_constraints_to_graph(current_module, "soft", soft_constraints, graph);
				add_constraints_to_graph(current_module, "hard", hard_constraints, graph);

				for (Rule *r : *ml->get_rules()) {
					for (int k = 0; k < r->get_body()->size(); k++)
						for (int l = 0; l < r->get_body()->at(k)->size(); l++) {
							ModuleGeneric *m = 0;
							try {
								m = this->get_pointer_to_module(r->get_body()->at(k)->at(l)->get_module(), true);
							} catch (Exception *ex) {delete ex; continue;}
							if (m->get_name() == "BEL") {
								std::vector<std::string> references = m->get_references(r->get_body()->at(k)->at(l));
								for (std::vector<std::string>::iterator it = references.begin(); it != references.end(); it++) {
									//printf("Reference of BEL: %s\n", (*it).c_str());
									m = this->get_pointer_to_module(*it, true);
									if (m && (m->get_module_type() == module_type::LOCAL || m->get_module_type() == module_type::GROUP || m->get_module_type() == module_type::BELIEF))
										if (m != ml) {
											std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->find(current_module);
											it->second->insert(m);
										}
								}
							} else {

								if (m && (m->get_module_type() == module_type::LOCAL || m->get_module_type() == module_type::GROUP || m->get_module_type() == module_type::BELIEF))
									if (m != ml) {
										std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->find(current_module);
										it->second->insert(m);
								}
							}
						}
				}
			}
		}

		/* lets print the graph */
		for (std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->begin(); it != graph->end(); it++) {
			if (DEBUG_GRAPH) {
				std::cout << it->first->get_name() << ": ";
				for (std::set<ModuleGeneric *>::iterator its = it->second->begin(); its != it->second->end(); its++) {
					std::cout << (*its)->get_name() << " ";
				}
				std::cout << "\n";
			}
			if (it->second->size() == 0)
				q.push(it->first->get_name());
		}

		while (q.size() > 0) {
			std::string name = q.front();
			if (DEBUG_GRAPH)
				std::cout << "pushing into reasoning queue: " << name << "\n";
			q.pop();
			ModuleGeneric *m = this->get_pointer_to_module(name, true);
			if (m->get_module_type() != module_type::LOCAL
			 && m->get_module_type() != module_type::GROUP
			 && m->get_module_type() != module_type::BELIEF)
				continue;
			// we need to erase all occurrences of m in graph
			for (std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->begin(); it != graph->end(); it++) {
				std::set<ModuleGeneric*>::iterator s = it->second->find(m);
				if (s != it->second->end()) {
					it->second->erase(s);
					if (it->second->size() == 0)
						q.push(it->first->get_name());
				}
			}
			order->push_back(m);
			std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->find(m);
			if (it != graph->end()) {
				delete it->second;
				graph->erase(it);
			}
		}

		if (graph->size() > 0) {
			std::stringstream ss;
			for (std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->begin(); it != graph->end(); it++)
				ss << it->first->get_name() << " ";
			for(std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->begin(); it != graph->end(); it++) {
				delete it->second;
			}
			delete graph;
			delete order;

			throw new Exception(std::string("Cycle in module calls (") + ss.str() + "are in cycles)!\n");
		}

		for(std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->begin(); it != graph->end(); it++) {
				delete it->second;
		}
		delete graph;

		this->program->set_order(order);
		return;
	}

	void Checker::add_constraints_to_graph_recursive(
			Expression* expr,
			ModuleGeneric* current_module,
			std::string type,
			std::map<ModuleGeneric*, std::set<ModuleGeneric*>*> *graph) {
		switch(expr->get_type())
		{
			case expression_type::AND:
			case expression_type::OR:
			case expression_type::IMPLIES:
				add_constraints_to_graph_recursive(expr->get_left_expression(), current_module, type, graph);
				add_constraints_to_graph_recursive(expr->get_right_expression(), current_module, type, graph);
				break;
			case expression_type::NOT:
				add_constraints_to_graph_recursive(expr->get_left_expression(), current_module, type, graph);
				break;
			case expression_type::QUANTIFIED:
				add_constraints_to_graph_recursive(expr->get_quantified_entry()->get_body(), current_module, type, graph);
				break;
			case expression_type::TERM:
				Fact* current = expr->get_term();
				ModuleGeneric *m = nullptr;
				try {
					m = this->get_pointer_to_module(current->get_module(), true);
				} catch (Exception *ex) {
					delete ex;
					throw new Exception("Module '" + current->get_module() + "' used in " + type + " of module '" + current_module->get_name() + "' does not exist!\n");
				}
				if (m && (m->get_module_type() == module_type::LOCAL || m->get_module_type() == module_type::GROUP || m->get_module_type() == module_type::BELIEF)) {
					if (m != current_module) {
						std::map<ModuleGeneric*, std::set<ModuleGeneric*>*>::iterator it = graph->find(current_module);
						it->second->insert(m);
					}
				}
				break;
		}
	}

	void Checker::add_constraints_to_graph(
			ModuleGeneric* current_module,
			std::string type,
			std::vector<Expression*>* constraints,
			std::map<ModuleGeneric*,
			std::set<ModuleGeneric*>*> *graph) {
		if (constraints && constraints->size()) {
			for (int index = 0; index < constraints->size(); index++) {
				add_constraints_to_graph_recursive(constraints->at(index), current_module, type, graph);
			}
		}
	}

	void Checker::check_constraints_recursive(
			Expression* expr,
			std::map<std::string, variable_type::type> *vm,
			std::string module,
			std::string type,
			std::vector<Domain*>* domains,
			bool check_quantify) {
		switch(expr->get_type())
		{
			case expression_type::AND:
			case expression_type::IMPLIES:
				check_constraints_recursive(expr->get_left_expression(), vm, module, type, domains);
				check_constraints_recursive(expr->get_right_expression(), vm, module, type, domains);
				break;
			case expression_type::OR: {
				std::map<std::string, variable_type::type> *vmr = new std::map<std::string, variable_type::type > ();
				vmr->insert(vm->begin(), vm->end());
				try {
					check_constraints_recursive(expr->get_left_expression(), vmr, module, type, domains);
				} catch (Exception* ex) {
					delete vmr;
					throw;
				}
				vmr->clear();
				vmr->insert(vm->begin(), vm->end());
				try {
					check_constraints_recursive(expr->get_right_expression(), vmr, module, type, domains);
				} catch (Exception* ex) {
					delete vmr;
					throw;
				}
				delete vmr;
			}
				break;
			case expression_type::NOT:
				check_constraints_recursive(expr->get_left_expression(), vm, module, type, domains);
				break;
			case expression_type::QUANTIFIED: {
				ConstraintEntry* entry = expr->get_quantified_entry();

				std::map<std::string, variable_type::type> *vmr = new std::map<std::string, variable_type::type > ();
				vmr->insert(vm->begin(), vm->end());

				std::vector<QuantifierEntry*>* q = entry->get_quantifiers();
				for (int i = 0; i < q->size(); i++) {
					bool declared = false;
					for (int j = 0; j < domains->size(); j++)
						if (q->at(i)->get_domain() == domains->at(j)->get_name()) {
							if (vmr->find(q->at(i)->get_id()) != vmr->end()) {
								throw new Exception("Variable '" + q->at(i)->get_id() + "' used multiple times (possibly redeclared) in " + type + " of module '" + module + "'!\n");
							}

							vmr->insert(std::make_pair(q->at(i)->get_id(), domains->at(j)->get_type()));
							declared = true;
						}
						if (!declared) {
							throw new Exception("Domain '" + q->at(i)->get_domain() + "' used in " + type + " of module '" + module + "' does not exist in that module!\n");
						}
				}

				check_constraints_recursive(entry->get_body(), vmr, module, type, domains, true);
			}
				break;
			case expression_type::TERM: {
				Fact *_fact = expr->get_term();
//				if (check_quantify) {
//					std::vector<std::string> vv = _fact->get_variable_names();
//					for (auto variable : vv) {
//						if (vm->find(variable) == vm->end()) {
//							throw new Exception("Fact '" + _fact->get_relation() + "' used in " + type + " of module '" + module + "' contains unquantified variable " + variable  + "!\n");
//						}
//					}
//				}

				if (_fact->get_module() == "BEL") {
					Expression* bel_body = _fact->get_bel_params();
					std::string relation_or_as = _fact->get_relation();
					AsExpr* as_expr;
					if (relation_or_as.find("<") != -1) {
						int pos = 0;
						as_expr = parse_asExpr(relation_or_as, pos);
					} else {
						as_expr = new AsExpr(relation_or_as);
					}
					try {
						check_bel(module, bel_body, nullptr, vm, as_expr);
					} catch (Exception* ex) {
						delete as_expr;
						throw;
					}

					delete as_expr;

					return;
				}
				/* we need to get pointer to a relation, so we need to check not local modules and modules declared in this program */
				ModuleGeneric *m = nullptr;
				std::string module_to_query = (_fact->get_module() != "") ? _fact->get_module() : module;
				try {
					m = this->get_pointer_to_module(module_to_query);
				} catch (Exception *ex) {
					delete ex;
					throw new Exception("Module '" + _fact->get_module() + "' used in " + type + " of module '" + module + "' does not exist!\n");
				}
				Relation* relation = m->get_relation(_fact->get_relation());

				// We can allow for set queries as Well Supported Model is already computed...
				//
				//if (_fact->is_set_query() && _fact->get_module() == module) {
				//	throw new Exception(std::string("Fact ") + _fact->get_module() + "." + _fact->get_relation()+(" with a set query (in {...}) is asking for a relation within its module!\n"));
				//}

				this->check_parameters(type, _fact->get_relation(), m->get_name(), relation->get_params(), _fact->get_params(), vm);

				break;
			}
			default:
				throw new Exception("Unexpected operand used in " + type + "/n");
				break;
		}
	}

	void Checker::check_constraints(std::string module, std::string type, std::vector<Expression*>* constraints, std::vector<Domain*>* domains) {
		if (constraints && constraints->size()) {
			for (int index = 0; index < constraints->size(); index++) {
				std::map<std::string, variable_type::type> *vm = new std::map<std::string, variable_type::type>();

				try {
					check_constraints_recursive(constraints->at(index), vm, module, type, domains);
				} catch (Exception* ex) {
					delete vm;
					throw;
				}

				delete vm;
			}
		}
	}

	void Checker::check_module(ModuleLocal *m) {
		/* do we have same named module? */

		for (int j = 0; j < this->modules->size(); j++)
			if (this->modules->at(j)->get_name() == m->get_name())
				throw (new Exception(std::string("Module ") + m->get_name() + " already declared in interpreter!\n"));

		for (int j = 0; j < m->get_domains()->size(); j++)
			this->check_domain(m, m->get_domains()->at(j));

		for (int j = 0; j < m->get_relations()->size(); j++)
			this->check_relation(m, m->get_relations()->at(j));

		add_domain_facts(m);

		for (int j = 0; j < m->get_rules()->size(); j++)
			this->check_rule(m, m->get_rules()->at(j), m->get_name());

		for (int j = 0; j < m->get_facts()->size(); j++)
			this->check_fact(m, m->get_facts()->at(j));

		Constraints* constr = m->get_constraints();
		std::vector<Expression*>* soft_constraints = constr->get_soft_constraints();
		std::vector<Expression*>* hard_constraints = constr->get_hard_constraints();

		check_constraints(m->get_name(), "soft constraint",  soft_constraints, m->get_domains());
		check_constraints(m->get_name(), "hard constraint", hard_constraints, m->get_domains());

		return;
	}

	void Checker::check_module(ModuleGroup *m) {
		for (int j = 0; j < this->modules->size(); j++)
			if (this->modules->at(j)->get_name() == m->get_name())
				throw (new Exception(std::string("Module ") + m->get_name() + " already declared in interpreter!\n"));

		for (int j = 0; j < m->get_relations()->size(); j++)
			this->check_relation(m, m->get_relations()->at(j));

		for (int j = 0; j < m->get_members()->size(); j++)
			this->check_members(m, m->get_members()->at(j));

		return;
	}

	void Checker::check_module(ModuleBelief *m) {
		m->set_modules(this->program->get_modules());
		for (int j = 0; j < this->modules->size(); j++)
			if (this->modules->at(j)->get_name() == m->get_name())
				throw (new Exception(std::string("Module ") + m->get_name() + " already declared in interpreter!\n"));

		for (int j = 0; j < m->get_domains()->size(); j++)
			this->check_domain(m, m->get_domains()->at(j));

		auto module_list = m->get_member_modules();
		for (auto module_it = module_list->begin(); module_it != module_list->end(); module_it++) {
			if ((*module_it)->get_module_type() != Inter4ql::module_type::LOCAL) {
				throw (new Exception(std::string("Member module ") + (*module_it)->get_name() + " of belief " + m->get_name() + " is not a pure 4QL module (a local module)!\n"));
			}
		}

		Constraints* constr = m->get_constraints();
		std::vector<Expression*>* soft_constraints = constr->get_soft_constraints();
		std::vector<Expression*>* hard_constraints = constr->get_hard_constraints();

		check_constraints(m->get_name(), "soft constraint", soft_constraints, m->get_domains());
		check_constraints(m->get_name(), "hard constraint", hard_constraints, m->get_domains());

		return;
	}

	void Checker::check_module(ModuleAction *m) {
		return;
	}

	void Checker::check_module(ModuleProblem *m) {
		for (auto module : *m->get_beliefs()) {
			ModuleGeneric* bm = nullptr;
			try {
				bm = get_pointer_to_module(module);
			} catch (Exception* e) {
				delete e;
				throw new Exception("Belief module '" + module + "' used in problem '" + m->get_name() + "' does not exist!\n");
			}
			if (bm->get_module_type() == module_type::ACTION || bm->get_module_type() == module_type::PROBLEM) {
				throw new Exception("Module '" + module + "' used as planning environment in problem '" + m->get_name() + "' is not a query-able module (actions and problems cannot be queried)!\n");
			}

			m->add_belief_module(bm);

			for (auto action : *m->get_actions()) {
				ModuleGeneric* am = nullptr;
				try {
					am = get_pointer_to_module(action);
				} catch (Exception* e) {
					delete e;
					throw new Exception("Action '" + action + "' used in problem '" + m->get_name() + "' does not exist!\n");
				}

				if (am->get_module_type() != module_type::ACTION) {
					throw new Exception("Module '" + action + "' used as action in problem '" + m->get_name() + "' is not an action-type module!\n");
				}

				ModuleAction* current_action = dynamic_cast<ModuleAction*>(am);

				std::map<std::string, variable_type::type> *vm = new std::map<std::string, variable_type::type>();
				check_constraints_recursive(current_action->get_preconditions(), vm, bm->get_name(), "'" + am->get_name() + "' (action's preconditions)", bm->get_domains());

				Expression* composite = current_action->get_composite_expr();
				if (composite)
					check_composite(composite, action, bm);

				std::vector<Rule*>* to_add_rules = current_action->get_add_rules();
				if (to_add_rules) {
					for(const auto rule : *to_add_rules) {
						check_rule(bm, rule, std::string(action + " (action)"), vm);
					}
				}

				std::vector<Rule*>* to_remove_rules = current_action->get_remove_rules();
				if (to_remove_rules) {
					for(const auto rule : *to_remove_rules) {
						check_rule(bm, rule, std::string(action + " (action)"), vm);
					}
				}

				delete vm;
			}
		}

		return;
	}

	void Checker::check_composite(Expression* expr, std::string action_name, ModuleGeneric* m)
	{
		switch(expr->get_type())
		{
			case expression_type::SEQUENTIAL: {
				check_composite(expr->get_left_expression(), action_name, m);
				check_composite(expr->get_right_expression(), action_name, m);
					}
				break;
			case expression_type::PARALLEL: {
				check_composite(expr->get_left_expression(), action_name, m);
				check_composite(expr->get_right_expression(), action_name, m);
					}
				break;
			case expression_type::IMPLIES: {
				std::map<std::string, variable_type::type> *vm = new std::map<std::string, variable_type::type>();
				try {
					check_constraints_recursive(expr->get_left_expression(), vm, m->get_name(), "'" + action_name + "' (action's preconditions)", m->get_domains());
				} catch (...) {
					delete vm;
					throw;
				}
				delete vm;

				check_composite(expr->get_right_expression(), action_name, m);
				check_composite(expr->get_else_expression(), action_name, m);
					}
				break;
			case expression_type::TERM: {
				Fact* term = expr->get_term();

				ModuleGeneric* bm  = get_pointer_to_module(term->get_relation());

				if (bm->get_module_type() != module_type::ACTION) {
					throw new Exception("Module '" + term->get_relation() + "' not being an action used in composite expression in action " + action_name + "\n");
				}

				ModuleAction* ma = dynamic_cast<ModuleAction*>(bm);
				std::vector<Value*>* args = ma->get_args();

				if (term->get_params()->size() != args->size()) {
					throw new Exception("Action '" + term->get_relation() + "' expects " + std::to_string(args->size()) +
							" arguments but " + std::to_string(term->get_params()->size()) + " given - in composite expression of action " + action_name + "\n");
				}
					}
				break;
			default:
				throw new Exception("Allowed relations between actions in composite expression are '||', ';' and '=>'. Unexpected operator found in composite action " + action_name);
		}
	}

	void Checker::check_domain(ModuleLocal *_m, Domain *_d) {
		/* do we have duplicated domains? */
		int how_many = 0;
		std::vector<Domain*> *_v = _m->get_domains();
		for (int i = 0; i < _v->size(); i++)
			if (_d->get_name() == _v->at(i)->get_name())
				how_many++;
		if (how_many > 1)
			throw new Exception(std::string("Domain ") + _d->get_name() + " in module " + _m->get_name() + " is declared more than one time!\n");
		return;
	}

	void Checker::check_domain(ModuleGroup *_m, Domain *_d) {
		/* do we have duplicated domains? */
		int how_many = 0;
		std::vector<Domain*> *_v = _m->get_domains();
		for (int i = 0; i < _v->size(); i++)
			if (_d->get_name() == _v->at(i)->get_name())
				how_many++;
		if (how_many > 1)
			throw new Exception(std::string("Domain ") + _d->get_name() + " in module " + _m->get_name() + " is declared more than one time!\n");
		return;
	}

	void Checker::check_domain(ModuleBelief *_m, Domain *_d) {
		/* do we have duplicated domains? */
		int how_many = 0;
		std::vector<Domain*> *_v = _m->get_domains();
		for (int i = 0; i < _v->size(); i++)
			if (_d->get_name() == _v->at(i)->get_name())
				how_many++;
		if (how_many > 1)
			throw new Exception(std::string("Domain ") + _d->get_name() + " in belief module " + _m->get_name() + " is declared more than one time!\n");
		return;
	}

	void Checker::check_relation(ModuleGeneric *_m, Relation *_r) {
		/* check if we do not have duplicated names of relations */
		int how_many = 0;
		for (int j = 0; j < _m->get_relations()->size(); j++)
			if (_m->get_relations()->at(j)->get_name() == _r->get_name())
				how_many++;
		if (how_many > 1)
			throw new Exception(std::string("Relation ") + _r->get_name() + " in module " + _m->get_name() + " is declared more than one time!\n");
		/* check if every parameter is a valid domain */
		for (int i = 0; i < _r->get_params()->size(); i++)
			if (_r->get_params()->at(i)->get_type() == variable_type::INVALID)
				throw new Exception(std::string("Unknown domain ") + _r->get_params()->at(i)->get_name() + " in relation " + _r->get_name() + " in module " + _m->get_name() + "!\n");
		return;
	}

	void Checker::check_bel_recursively(
			std::string curent_module,
			AsExpr* e,
			Fact* term,
			Expression* quantified_entry,
			std::set<std::string> *head_variables,
			std::map<std::string, variable_type::type> *vm) {
		if (e->type_ == 0) {
			ModuleGeneric* m = this->get_pointer_to_module(e->id_);

			if (term) {
				Relation* r = m->get_relation(term->get_relation());

				this->check_parameters(curent_module, term->get_relation(), e->id_, r->get_params(), term->get_params(), vm, head_variables);
			}
			if (quantified_entry) {
				this->check_constraints_recursive(quantified_entry, vm, m->get_name(), "BEL expression", m->get_domains());
			}
		} else {
			check_bel_recursively(curent_module, e->left_, term, quantified_entry, head_variables, vm);
			check_bel_recursively(curent_module, e->right_, term, quantified_entry, head_variables, vm);
		}
	}

	void Checker::check_bel(
			std::string current_module,
			Expression* bel_body,
			std::set<std::string> *head_variables,
			std::map<std::string, variable_type::type> *vm,
			AsExpr* modules) {
		switch(bel_body->get_type()) {
		case expression_type::AND:
		case expression_type::IMPLIES:
			check_bel(current_module, bel_body->get_left_expression(), head_variables, vm, modules);
			check_bel(current_module, bel_body->get_right_expression(), head_variables, vm, modules);
			break;
		case expression_type::OR:
		{
			std::map<std::string, variable_type::type> *vmrL = new std::map<std::string, variable_type::type > ();
			vmrL->insert(vm->begin(), vm->end());

			std::set<std::string> *head_variables_left = new std::set<std::string>();
			try {
				check_bel(current_module, bel_body->get_left_expression(), head_variables_left, vmrL, modules);
			} catch (Exception* ex) {
				delete vmrL;
				delete head_variables_left;
				throw;
			}

			std::map<std::string, variable_type::type> *vmrR = new std::map<std::string, variable_type::type > ();
			vmrR->insert(vm->begin(), vm->end());

			std::set<std::string> *head_variables_right = new std::set<std::string>();
			try {
			check_bel(current_module, bel_body->get_right_expression(), head_variables_right, vmrR, modules);
			} catch (Exception* ex) {
				delete vmrL;
				delete vmrR;
				delete head_variables_left;
				delete head_variables_right;
				throw;
			}

			if(head_variables) {
				for(auto variable = head_variables_left->begin(); variable != head_variables_left->end(); variable++) {
					if (head_variables_right->find(*variable) != head_variables_right->end()) {
						head_variables->insert(*variable);
					}
				}
			}

			if(vm) {
				for(auto var = vmrL->begin(); var != vmrL->end(); var++) {
					if (vmrR->find(var->first) != vmrR->end()) {
						vm->insert(*var);
					}
				}
			}

			delete vmrL;
			delete vmrR;
			delete head_variables_left;
			delete head_variables_right;
			break;
		}
		case expression_type::NOT:

			check_bel(current_module, bel_body->get_left_expression(), head_variables, vm, modules);
			break;
		case expression_type::QUANTIFIED:
			check_bel_recursively(current_module, modules, nullptr, bel_body, head_variables, vm);
			break;
		case expression_type::TERM:
			Fact* term = bel_body->get_term();
			check_bel_recursively(current_module, modules, term, nullptr, head_variables, vm);
			break;
		}
 	}

	void Checker::check_rule(ModuleGeneric *_m, Rule *_r, std::string msource, std::map<std::string, variable_type::type> *vm_init) {
		if (_m->get_module_type() != module_type::LOCAL && _m->get_module_type() != module_type::BELIEF) {
			throw new Exception(std::string("Module " + _m->get_name() + " does not provide information about relations!\n"));
		}

		Relation *relation = _m->get_relation(_r->get_head()->get_relation());
		/* variable map - to map variable types */
		std::map<std::string, variable_type::type> *vm = new std::map<std::string, variable_type::type > ();

		if (relation->get_params()->size() != _r->get_head()->get_params()->size()) {
			delete vm;
			throw new Exception(std::string("Wrong number of parameters passed for relation ") + _r->get_head()->get_relation() + " in module " + msource + "!\n");
		}

		try {
			this->check_parameters(msource, relation->get_name(), _m->get_name(), relation->get_params(), _r->get_head()->get_params(), vm);
		} catch (Exception* ex) {
			delete vm;
			throw;
		}

		for (int i = 0; i < _r->get_body()->size(); i++) {
			/* for each OR rules */
			std::map<std::string, variable_type::type> *vmr = new std::map<std::string, variable_type::type > ();
			// I have not idea who called it "head_valriables"... These are variables that were spotted in
			// body of a rule - why "head"!?
			std::set<std::string> *head_variables = new std::set<std::string>;
			/* copy vm to vmr (because each branch of rule can have each own variables */
			vmr->insert(vm->begin(), vm->end());
			if (vm_init) {
				vmr->insert(vm_init->begin(), vm_init->end());
				for(const auto var_name : *vm_init) {
					head_variables->insert(var_name.first);
				}
			}
			for (int j = 0; j < _r->get_body()->at(i)->size(); j++) {
				Fact *_fact = _r->get_body()->at(i)->at(j);
				if (_fact->get_module() == "BEL") {
					Expression* bel_body = _fact->get_bel_params();
					std::string relation_or_as = _fact->get_relation();
					AsExpr* as_expr;
					if (relation_or_as.find("<") != -1) {
						int pos = 0;
						as_expr = parse_asExpr(relation_or_as, pos);
					} else {
						as_expr = new AsExpr(relation_or_as);
					}
					try {
						check_bel(msource, bel_body, head_variables, vmr, as_expr);
					} catch (Exception* ex) {
						delete vm;
						delete vmr;
						delete head_variables;
						delete as_expr;
						throw;
					}
					delete as_expr;

					continue;
				}
				/* we need to get pointer to a relation, so we need to check not local modules and modules declared in this program */
				ModuleGeneric *m = 0;
				try {
					if (_fact->get_module() == "") {
						m = _m;
					} else {
						m = this->get_pointer_to_module(_fact->get_module());
					}
					relation = m->get_relation(_fact->get_relation());
				} catch (Exception* ex) {
					delete vm;
					delete vmr;
					delete head_variables;
					throw;
				}
				if (_fact->is_set_query() && _fact->get_module() == _m->get_name()) {
					delete vm;
					delete vmr;
					delete head_variables;
					throw new Exception(std::string("Fact ") + _fact->get_module() + "." + _fact->get_relation()+(" with a set query (in {...}) is asking for a relation within its module!\n"));
				}

				try {
					this->check_parameters(msource, _fact->get_relation(), m->get_name(), relation->get_params(), _fact->get_params(), vmr, head_variables);
				} catch(Exception *ex) {
					delete vm;
					delete vmr;
					delete head_variables;
					throw;
				}
			}
			for (std::map<std::string, variable_type::type>::iterator it = vmr->begin(); it != vmr->end(); it++) {
				if (head_variables->find(it->first) == head_variables->end()) {
					std::string varName = it->first;
					delete vmr;
					delete vm;
					delete head_variables;
					throw new Exception(std::string("Variable '") + varName + "' defined in rule's '" + _r->get_head()->get_relation() + "' head in module '" + msource + "' is not defined in one of the OR-branches in body!\n");
				}
			}
			
			delete vmr;
			delete head_variables;
		}
		delete vm;
		return;
	}

	void Checker::check_fact(ModuleLocal *_m, Fact *_r) {
		Relation *relation = _m->get_relation(_r->get_relation());
		std::map<std::string, variable_type::type> *vm = new std::map<std::string, variable_type::type > ();
		try {
			this->check_parameters(_m->get_name(), _r->get_relation(), _m->get_name(), relation->get_params(), _r->get_params(), vm);
		} catch (Inter4ql::Exception *ex) {
			delete vm;
			throw;
		}

		delete vm;
		return;
	}

	void Checker::add_domain_facts(ModuleLocal *m) {
		std::vector<Relation*> *_r = m->get_relations();
		for (int j = 0; j < m->get_domains()->size(); j++) {
			/* is the domain called like any of relations? */
			for (int i = 0; i < _r->size(); i++)
				if (_r->at(i)->get_name() == m->get_domains()->at(j)->get_name()) {
					throw new Exception(std::string("Domain ") + m->get_domains()->at(j)->get_name() + " in module " + m->get_name() + " has the same name as one of the relations in this module! Please use distinct names for relations and domains within one module.\n");
				}
			std::vector<Domain*>* params = new std::vector<Domain*>();
			params->push_back(new Domain(m->get_domains()->at(j)->get_name(), m->get_domains()->at(j)->get_type()));
			Relation* r = new Relation(m->get_domains()->at(j)->get_name(), params, m->get_name());
			m->get_relations()->push_back(r);
		}
	}

	void Checker::check_members(ModuleGroup *m, std::string name)
	{
		ModuleGeneric* member_module = this->get_pointer_to_module(name);

		for (int i = 0; i < m->get_relations()->size(); i++)
			// check if that module has relation i
			try {
				member_module->get_relation( m->get_relations()->at(i)->get_name() );
			} catch(Inter4ql::Exception * e)
			{
				e->set(e->get() + std::string("Required by group ") + m->get_name() + ".\n");
				throw;
			}

	}

	void Checker::check_parameters(
			std::string msource,
			std::string rname,
			std::string mname,
			std::vector<Domain*> *vd,
			std::vector<Value*> *_vv,
			std::map<std::string, variable_type::type> *map,
			std::set<std::string> *vars) {
		if (vd->size() != _vv->size()) {
			throw new Exception(std::string("Wrong number of parameters specified in module " + msource + " for relation ") + rname + " from module " + mname + "!\n");
		}
		for (int i = 0; i < vd->size(); i++) {
			if (_vv->at(i)->get_type() == variable_type::VARIABLE) {
				// does every variable holds only one type?
				if (vars)
					vars->insert(_vv->at(i)->get_value_string());
				if (map) {
					if (map->find(_vv->at(i)->get_value_string()) != map->end())
						if (map->at(_vv->at(i)->get_value_string()) != vd->at(i)->get_type())
							throw new Exception(std::string("Variable ") + _vv->at(i)->get_value_string() +
								" was previously declared as " + variable_type2string(map->at(_vv->at(i)->get_value_string())) +
								" (now in module " + msource + " is " + variable_type2string(vd->at(i)->get_type()) + ")" +
								" in relation " + rname + " from module " + mname + "!\n");
					map->insert(std::pair<std::string, variable_type::type > (_vv->at(i)->get_value_string(), vd->at(i)->get_type()));
				}
			} else if (_vv->at(i)->get_type() != vd->at(i)->get_type()) {
				/* are the types for constant values matched? */
				std::stringstream ss;
				ss << "Error in modue " + msource + ". Parameter number " << i + 1 << " should have type ";
				ss << variable_type2string(vd->at(i)->get_type()) << " not ";
				ss << variable_type2string(_vv->at(i)->get_type()) << " in relation ";
				ss << rname << " from module " << mname << "!\n";
				throw new Inter4ql::Exception(ss.str());
			}
		}
		return;
	}

	ModuleGeneric *Checker::get_pointer_to_module(std::string _module_name, bool outer) {
		if (program) {
			for (int i = 0; i < this->program->get_modules()->size(); i++) {
				ModuleGeneric *m = this->program->get_modules()->at(i);
				if (m->get_name() == _module_name)
					return m;
			}
		}
		/* if we did not find module in imported module, we want to give module from already defined */
		for (int i = 0; i < this->modules->size(); i++) {
			ModuleGeneric *m = this->modules->at(i);
			if (m->get_name() == _module_name)
				if(outer) {
					if(m->get_module_type() != module_type::LOCAL) return m;
				} else {
					return m;
				}
		}
		throw new Inter4ql::Exception(std::string("Cannot find module ") + _module_name + "!\n");
	}

	void Checker::check_query(Fact *_fact) {
		if (_fact->get_module() == "")
			throw new Exception("Empty module reference found in the query! This is not permitted.\n");
		ModuleGeneric *m = this->get_pointer_to_module(_fact->get_module());
		std::map<std::string, variable_type::type> vm;
		if(m->get_name() == "BEL") {
			Expression* bel_body = _fact->get_bel_params();
			std::string relation_or_as = _fact->get_relation();
			AsExpr* as_expr;
			if (relation_or_as.find("<") != -1) {
				int pos = 0;
				as_expr = parse_asExpr(relation_or_as, pos);
			} else {
				as_expr = new AsExpr(relation_or_as);
			}
			try {
			check_bel(m->get_name(), bel_body, nullptr, &vm, as_expr);
			} catch(Exception* ex) {
				delete as_expr;
				throw;
			}
			delete as_expr;

			return;
		}
		Relation *r = m->get_relation(_fact->get_relation());
		this->check_parameters(m->get_name(), _fact->get_relation(), m->get_name(), r->get_params(), _fact->get_params(), &vm);
	}

	void Checker::check_expression(Expression* expr)
	{
		switch(expr->get_type())
		{
			case expression_type::AND:
			case expression_type::IMPLIES: {
						expr->get_left_expression()->set_module(expr->get_module());
						check_expression(expr->get_left_expression());
						expr->get_right_expression()->set_module(expr->get_module());
						check_expression(expr->get_right_expression());
					}
				break;
			case expression_type::OR: {
						expr->get_left_expression()->set_module(expr->get_module());
						check_expression(expr->get_left_expression());
						expr->get_right_expression()->set_module(expr->get_module());
						check_expression(expr->get_right_expression());
					}
				break;
			case expression_type::NOT: {
						expr->get_left_expression()->set_module(expr->get_module());
						check_expression(expr->get_left_expression());
					}
				break;
			case expression_type::QUANTIFIED: {
						throw new Exception("Quantified expression found without module context. This is not allowed due to a lack of possibility of domain validation.\n");
					}
				break;
			case expression_type::TERM: {
						//expr->get_term()->set_module(expr->get_module());
						check_query(expr->get_term());
					}
				break;
			default:
				return;
		}
	}
}
