#ifndef __CHECKER_H__
#define __CHECKER_H__

#include <vector>
#include <string>
#include <map>
#include <set>

#include "Functions.h"
#include "Types.h"

namespace Inter4ql {

	class VariableSpace;
	class ModuleGeneric;
	class ModuleLocal;
	class ModuleGroup;
	class ModuleBelief;
	class Program;
	class Value;
	class Fact;
	class Domain;
	class Rule;
	class Relation;
	class ConstraintEntry;
	/** @brief Checker class - class that validates 4QL programs
	 *
	 */
	class Checker {
	public:
		/** @brief constructor
		 *
		 * @param _modules pointer to modules in interpreters enviroment
		 */
		explicit Checker(std::vector<ModuleGeneric *> *_modules);
		/** @brief function to set the modules pointer
		 *
		 * @param _modules pointer to modules
		 */
		void set_modules(std::vector<ModuleGeneric *> *_modules);
		/** @brief function that validates 4QL program, throws exception on failure
		 *
		 * @param p pointer to Program class with a 4QL program to validate
		 */
		void check(Program *p);
		/** @brief function that validates query
		 *
		 * @param r pointer to a Fact class to validate
		 */
		void check_query(Fact *r);

		/** @brief function that validates expression
		 *
		 * @param expr pointer to a Expression class to validate
		 */
		void check_expression(Expression* expr);
	private:
		void check_cycles();
		void check_module(ModuleLocal *m);
		void check_module(ModuleGroup *m);
		void check_module(ModuleBelief *m);
		void check_module(ModuleAction *m);
		void check_module(ModuleProblem *m);
		void check_domain(ModuleLocal *m, Domain *_d);
		void check_domain(ModuleGroup *m, Domain *_d);
		void check_domain(ModuleBelief *m, Domain *_d);
		void check_relation(ModuleGeneric *_m, Relation *_r);
		void check_rule(ModuleGeneric *_m, Rule *_r, std::string msource,
				std::map<std::string, variable_type::type> *vm_init = nullptr);
		void check_fact(ModuleLocal *_m, Fact *_r);
		void add_domain_facts(ModuleLocal *m);
		void check_members(ModuleGroup *m, std::string name);
		void check_composite(Expression* expr, std::string action_name, ModuleGeneric* m);

		void add_constraints_to_graph(
				ModuleGeneric* current_module,
				std::string type,
				std::vector<Expression*>* constraints,
				std::map<ModuleGeneric*,
				std::set<ModuleGeneric*>*> *graph);
		void check_parameters(
				std::string msource,
				std::string rname,
				std::string mname,
				std::vector<Domain*> *vd,
				std::vector<Value*> *_vv,
				std::map<std::string,variable_type::type> *map = 0,
				std::set<std::string> *vars = 0);
		void check_constraints(
				std::string module,
				std::string type,
				std::vector<Expression*>* constraints,
				std::vector<Domain*>* domains);
		void check_bel(
				std::string current_module,
				Expression* bel_body,
				std::set<std::string> *head_variables,
				std::map<std::string, variable_type::type> *vm,
				AsExpr* modules);
		void check_bel_recursively(
				std::string curent_module,
				AsExpr* e,
				Fact* term,
				Expression* quantified_entry,
				std::set<std::string> *head_variables,
				std::map<std::string, variable_type::type> *vm);
		void add_constraints_to_graph_recursive(
				Expression* expr,
				ModuleGeneric* current_module,
				std::string type,
				std::map<ModuleGeneric*, std::set<ModuleGeneric*>*> *graph);
		void check_constraints_recursive(
				Expression* expr,
				std::map<std::string, variable_type::type> *vm,
				std::string module,
				std::string type,
				std::vector<Domain*>* domains,
				bool check_quantify = false);

		ModuleGeneric *get_pointer_to_module(std::string _module_name, bool outer = false);
		std::vector<ModuleGeneric *> *modules;
		/* this variable is set during checking */
		Program *program;
	};
}

#endif
