
#include "Fact.h"
#include "Value.h"
#include "Logic.h"
#include "Disposable.h"

#include <set>
#include <string>

namespace Inter4ql {

	Fact::Fact(std::string _module, std::string a_relation, std::vector<Value *> *a_params, bool _t, bool _f)
	: module(_module)
	, relation(a_relation)
	, params(a_params)
	, bel_params(nullptr)
	, bel_free_query(true)
	, delete_bel_params(true)
	{
		this->set = 0;
		if (_t && _f)
			this->value = Inter4ql::logic_type::INCONSISTENT;
		else if (!_t && _f)
			this->value = Inter4ql::logic_type::FALSE;
		else if (_t && !_f)
			this->value = Inter4ql::logic_type::TRUE;
		else if (!_t && !_f)
			this->value = Inter4ql::logic_type::UNKNOWN;
	}

	Fact::Fact(std::string _module, std::string a_relation, Expression *a_params, bool _t, bool _f)
	: Fact(_module, a_relation, new std::vector<Value*>(), _t, _f)
	{
		this->bel_params = a_params;
	}

	Fact::Fact(Fact* f)
	: module(f->module)
	, relation(f->relation)
	, params(new std::vector<Value*>())
	, bel_params(f->bel_params)
	, bel_free_query(f->bel_free_query)
	, delete_bel_params(f->delete_bel_params)
	{
		this->set = 0; // this is a copy only for Results... SET is not needed...
		for(std::vector<Value*>::iterator it = f->params->begin(); it != f->params->end(); it++) {
			this->params->push_back(*it);
		}
		this->value = f->value;
	}

	Fact::~Fact() {
		if(!Disposable::errorMode) {
			delete this->params;
			if (delete_bel_params)
				delete this->bel_params;

			if(this->set) {
				for(std::vector<Value*>::iterator it = this->set->begin(); it != this->set->end(); it++) {
					delete (*it);
				}
				delete this->set;
			}
		} else {
			if (module == "BEL") {
				delete this->params;
			}
			//std::cout << "DELETE FACT*" << std::endl;
		}
	}

	bool Fact::operator==(Fact &r) {
		if ((this->module != r.module) || (this->relation != r.relation) || (this->params->size() != r.params->size()))
			return false;
		for (int i = 0; i < this->params->size(); i++) {
			if (*this->params->at(i) != *r.params->at(i))
				return false;
		}
		return true;
	}

	void printBelParams(std::string& s, Expression* expr) {
		switch(expr->get_type())
		{
			case expression_type::AND: {
						printBelParams(s, expr->get_left_expression());
						s += ", ";
						printBelParams(s, expr->get_right_expression());
					  }
				break;
			case expression_type::IMPLIES: {
						printBelParams(s, expr->get_left_expression());
						s += "-> ";
						printBelParams(s, expr->get_right_expression());
					  }
				break;
			case expression_type::OR: {
						printBelParams(s, expr->get_left_expression());
						s += "| ";
						printBelParams(s, expr->get_right_expression());
					  }
				break;
			case expression_type::NOT: {
						s += "-";
						printBelParams(s, expr->get_left_expression());
					  }
				break;
			case expression_type::TERM: {
						s += expr->get_term()->print();;
			 		  }
				break;
		};	
	}

	std::string Fact::print(bool type) const {
		std::string s("");
		if (this->value == Inter4ql::logic_type::FALSE)
			s += "-";
		if (bel_params != nullptr) {
			s += "Bel";
			if (this->bel_free_query)
				s += "*";
			s += "[" + this->relation + "](";
			printBelParams(s, bel_params);
			s += ") ";
		} else {
			if (this->module != std::string(""))
				s += this->module + ".";
			s += this->relation + "(" + this->print_params(type) + ") ";
		}
		if(!this->set)
			s += ": " + Logic::get_logic_name(this->value);
		else{
			s += " in {";
			for(int i = 0; i < this->set->size(); i++){
				s += this->set->at(i)->print(false);
				if(i+1 < this->set->size())
					s += ", ";
			}
			s += "}";
		}
		return s;
	}

	std::string Fact::print_params(bool type) const {
		std::string s("");
		for (int i = 0; i < this->params->size(); i++) {
			if (i != 0)
				s += ", ";
			s += this->params->at(i)->print(type);
		}
		return s;
	}

	void Fact::set_value(logic_type::type _v) {
		this->value = _v;
	}

	Inter4ql::logic_type::type Fact::get_value() const {
		return this->value;
	}

	Expression *Fact::get_bel_params() const {
		return this->bel_params;
	}

	std::vector<Value *> *Fact::get_params() const {
		return this->params;
	}

	std::string Fact::get_module() const {
		return this->module;
	}

	void Fact::set_module(std::string _m) {
		this->module = _m;
	}

	std::string Fact::get_relation() const {
		return this->relation;
	}

	void Fact::set_set(std::vector<Value *> *_v) {
		this->set = _v;
	}

	std::vector<Value *> *Fact::get_set() {
		return this->set;
	}

	std::string Fact::get_set_string() {
		std::string s = "0000";
		for (int i = 0; i < this->set->size(); i++) {
			if (this->set->at(i)->get_type() == variable_type::LOGIC) {
				switch (this->set->at(i)->get_value_int()) {
					case logic_type::UNKNOWN:
						// NOT does not change UNKNOWN - therfore if we want it, we want UNKNOWN facts.
						s[0] = '1';
						break;
					case logic_type::FALSE:
						// Do we want FALSE facts? But what about negated facts?
						if (this->value == logic_type::FALSE) {
							// The facts are negated - so we want TRUE facts here as NOT is applied before IN.
							s[2] = '1';
						}
						else if(this->value == logic_type::TRUE) {
							// The facts are not negated - we want FALSE facts here.
							s[1] = '1';
						}
						break;
					case logic_type::TRUE:
						// Do we want TRUE facts? But what about negated facts?
						if (this->value == logic_type::FALSE) {
							// The facts are negated - so we want FALSE facts here as NOT is applied before IN.
							s[1] = '1';
						}
						else if(this->value == logic_type::TRUE) {
							// The facts are not negated - we want TRUE facts here.
							s[2] = '1';
						}
						break;
					case logic_type::INCONSISTENT:
						// NOT does not change INCONSISTENT - therfore if we want it, we want INCONSISTENT facts.
						s[3] = '1';
						break;
				}
			}
		}
		return s;
	}

	bool Fact::is_set_query() {
		return (this->set != 0);
	}

	bool operator<(const Fact &a, const Fact &b) {
		if (a.get_module() != b.get_module())
			return a.get_module() < b.get_module();
		else if (a.get_relation() != b.get_relation())
			return a.get_relation() < b.get_relation();
		if (a.get_params()->size() != b.get_params()->size())
			return a.get_params()->size() < b.get_params()->size();
		for (int i = 0; i < a.get_params()->size(); i++)
			if (a.get_params()->at(i) != b.get_params()->at(i))
				return a.get_params()->at(i) < b.get_params()->at(i);
		return false;
	}

	std::vector<std::string> get_variables(Expression* e) {
		if (e->get_type() == expression_type::TERM) {
			return e->get_term()->get_variable_names();
		}
		if (e->get_type() == expression_type::QUANTIFIED) {
			return get_variables(e->get_quantified_entry()->get_body());
		}
		if (e->get_type() == expression_type::NOT) {
			return get_variables(e->get_left_expression());
		}
		std::vector<std::string> to_return = get_variables(e->get_left_expression());
		std::vector<std::string> to_add = get_variables(e->get_right_expression());
		to_return.insert(to_return.end(), to_add.begin(), to_add.end());
		return to_return;
	}

	std::vector<std::string> Fact::get_variable_names() const
	{
		if (this->get_module() == "BEL") {
			return get_variables(this->bel_params);
		}

		std::set<std::string> variable_names;
		for (const auto& param : *get_params()) {
			if (param->get_type() == Inter4ql::variable_type::VARIABLE)
				variable_names.insert(param->get_value_string());
		}

		return std::vector<std::string>(variable_names.begin(), variable_names.end());

	}

	void Fact::valuate_fact(std::map<std::string, Value*> binding)
	{
		*this->params = apply_bind(this->params, binding);
	}

	void Fact::should_delete_bel_params(bool value)
	{
		this->delete_bel_params = value;
	}

	void Fact::set_bel_free_query(bool value)
	{
		this->bel_free_query = value;
	}

	bool Fact::is_bel_free_query()
	{
		return this->bel_free_query;
	}
}
