#ifndef __FUNCTIONS_H__
#define __FUNCTIONS_H__

#include <string>

#include "LineParser.tab.h"
#include "Types.h"
#include "VariableSpace.h"

namespace Inter4ql {
	class Valuation;

	/** @brief AsExpr class - class that stores as-expressions
 	 *
 	 */
	class Result;
	class AsExpr {
	public:
		/** @brief constructor that creates a new AsExpr without an ID
		 *
		 */
		AsExpr() {
			this->left_ = nullptr;
			this->right_ = nullptr;
			this->type_ = 1;
		};
		/** @brief constructor that creates a new AsExpr
		 *
		 * @param id ID of module stored in that object
		 */
		AsExpr(std::string id) {
			this->id_ = id;
			this->left_ = nullptr;
			this->right_ = nullptr;
			this->type_ = 0;
		};
		/** @brief destructor
		 *
		 */
		~AsExpr() {
			if (this->left_) delete this->left_;
			if (this->right_) delete this->right_;
		}

		AsExpr* left_;
		AsExpr* right_;
		std::string id_;
		bool type_; // 0 - id, 1 - connector (left/right)
	};

	/** @brief function that converts location_type to string
	 *
	 * @param _location location
	 * @return string with location
	 */
	std::string location2string(LineParser::location_type _location);
	/** @brief function that converts variable_type to string
	 *
	 * @param _variable variable type
	 * @return string with variable type
	 */
	std::string variable_type2string(Inter4ql::variable_type::type _variable);
	/** @brief function that converts string to variable_type
	 *
	 * @param s string with variable type
	 * @return variable_type converted from s
	 */
	variable_type::type string2variable_type(std::string s);

	/** @brief tries to match values from query and value
	 *
	 * @param query vector of values
	 * @param value vector of values
	 * @param binding mapping of variable names into values
	 * @return whether query can be matched witch value
	 */
	bool bind_variables(const std::vector<Value*>& query, const std::vector<Value*>& value, std::map<std::string, Value*>& binding);
	/** @brief applies valuation on values
	 *
	 * @param query vector of values
	 * @param binding mapping of variable names into values
	 * @return new vector of values with variables changed to their proper values
	 */
	std::vector<Value*> apply_bind(const std::vector<Value*>* query, const std::map<std::string, Value*>& binding);

	/** @brief parses string representation of as-expression into an object
	 *
	 * @param toParse string to be parsed
	 * @param pos current position of parsing
	 * @return an AsExpr object representing the as-expression
	 */
	AsExpr* parse_asExpr(std::string toParse, int& pos);

	/** @brief global printer for expressions
	 *
	 * @param expr expression to be printed
	 * @return formatted output string ready to be printed
	 */
	std::string print_expression(Expression* expr);
	/** @brief global printer for constraints
	 *
	 * @param s string to append to
	 * @param constraints list to be formatted
	 */
	void print_constraints(std::string& s, std::vector<Expression*>* constraints);
	/** @brief global validator for constraints
	 *
	 * @param to_check constraints to be checked
	 * @param disable_constraints variable allowing for disabling constraints (in case of nested call)
	 * @param domains domains to be checked against
	 * @param modules modules to be checked against
	 * @param current_modules module that called the check
	 * @return whether constraints are valid or not
	 */
	bool check_constraints(Ctraints to_check, bool& disable_constraints, std::vector<Domain*> *domains, std::vector<ModuleGeneric *> *modules, const ModuleGeneric* current_module);
	/** @brief global evaluator for expressions
	 *
	 * @param expr expression to be evaluated
	 * @param this_module module to be used when fact does not point to any module
	 * @param module list of module to evaluate against
	 * @param untill_now valuation obtained untill this moment to be used for internal and external modules (which do not accept variables)
	 * @param domains domain list which should be used for forall/forach evaluation
	 * @param rating rating stating how "true" the expression is
	 * @return result of evaluation
	 */
	Result evaluate_expression_global(Expression* expr, const ModuleGeneric* this_module, std::vector<ModuleGeneric*>* modules, std::vector<Domain*> *domains, int& rating);
	/** @brief internal evaluator for expressions
	 *
	 * @param expr expression to be evaluated
	 * @param this_module module to be used when fact does not point to any module
	 * @param module list of module to evaluate against
	 * @param untill_now valuation obtained untill this moment to be used for internal and external modules (which do not accept variables)
	 * @param domains domain list which should be used for forall/forach evaluation
	 * @param rating rating stating how "true" the expression is
	 * @return result of evaluation
	 */
	Result evaluate_expression_global_internal(Expression* expr, const ModuleGeneric* this_module, std::vector<ModuleGeneric*>* modules,
			Result untill_now, std::vector<Domain*> *domains, int& rating);
	/** @brief make snapshot of a database
	 *
	 * @param db database to be copied
	 * @return new database, equal to the input
	 */
	database* make_snapshot(database* db);
	/** @brief adds fact to database
	 *
	 * @param database database to be modified
	 * @param fact fact to be added
	 * @param value logic value to be used while adding
	 */
	void add_to_database(database* database, Fact* fact, Inter4ql::logic_type::type value);
	/** @brief removes fact from database
	 *
	 * @param database database to be modified
	 * @param fact fact to be removed
	 */
	void remove_from_database(database* database, Fact* fact);
	/** @brief checks whether database contains specific relation with arguments
	 *
	 * @param database database to be checked
	 * @param relation relation name to be checked
	 * @param vect argument list of the relation
	 * @return logic value which is assigned to that relation in databse (UNKNOWN if not found)
	 */
	logic_type::type value_in_database(database* database, std::string relation, const std::vector<Value*>& vect);
	/** @brief evaluates given term (possibly with variables) in given database
	 *
	 * @param db database to be checked
	 * @param term Fact to be evaluated
	 * @param red output argument for storing a result
	 * @param db_to_rm list to forbidden relations (will be removed from result of evaluation)
	 * @param check_existance informs whether a non-existing relation should cause exeption
	 * @return true if some items were added to the result, false otherwise
	 */
	bool ask_for_fact_in_database(const database* db, Fact* term, Result& res, const database* db_to_rm = nullptr, bool check_existance = true);
}

#endif
