#ifndef __LOGIC_H__
#define __LOGIC_H__

#include <string>
#include "Types.h"
#include "Functions.h"

namespace Inter4ql {


	/** @brief Static class for handling four-valued logic operations
	 */
	class Logic {
	public:
		/** @brief or method - when you have type a in model and put type b...
		 *
		 * @param a first logic value
		 * @param b second logic value
		 * @returns a or b
		 */
		static logic_type::type OR_RESULT(logic_type::type a, logic_type::type b) {
			if (a == b)
				return a;
			if (a == logic_type::TRUE && b == logic_type::FALSE
			 || b == logic_type::TRUE && a == logic_type::FALSE
			 || a == logic_type::INCONSISTENT
			 || b == logic_type::INCONSISTENT)
				return logic_type::INCONSISTENT;
			if (a == logic_type::UNKNOWN)
				return b;
			if (b == logic_type::UNKNOWN)
				return a;
		}

		/** @brief or logic method - in standard logic terms
		 *
		 * @param a first logic value
		 * @param b second logic value
		 * @returns a or b
		 */
		static logic_type::type OR(logic_type::type a, logic_type::type b) {
			if (a == b)
				return a;

			if (a == logic_type::FALSE) {
				return b;
			}

			if (a == logic_type::UNKNOWN) {
				if(b == logic_type::FALSE)        return logic_type::UNKNOWN;
				if(b == logic_type::UNKNOWN)      return logic_type::UNKNOWN;
				if(b == logic_type::INCONSISTENT) return logic_type::INCONSISTENT;
				if(b == logic_type::TRUE)         return logic_type::TRUE;
			}

			if (a == logic_type::INCONSISTENT) {
				if(b == logic_type::FALSE)        return logic_type::INCONSISTENT;
				if(b == logic_type::UNKNOWN)      return logic_type::INCONSISTENT;
				if(b == logic_type::INCONSISTENT) return logic_type::INCONSISTENT;
				if(b == logic_type::TRUE)         return logic_type::TRUE;
			}

			if (a == logic_type::TRUE) {
				return logic_type::TRUE;
			}
		}

		/** @brief and logic method
		 *
		 * @param a first logic value
		 * @param b second logic value
		 * @returns a and b
		 */
		static logic_type::type AND(logic_type::type a, logic_type::type b) {
			if (a == b)
				return a;

			if (a == logic_type::FALSE) {
				return logic_type::FALSE;
			}

			if (a == logic_type::UNKNOWN) {
				if(b == logic_type::FALSE)        return logic_type::FALSE;
				if(b == logic_type::UNKNOWN)      return logic_type::UNKNOWN;
				if(b == logic_type::INCONSISTENT) return logic_type::UNKNOWN;
				if(b == logic_type::TRUE)         return logic_type::UNKNOWN;
			}

			if (a == logic_type::INCONSISTENT) {
				if(b == logic_type::FALSE)        return logic_type::FALSE;
				if(b == logic_type::UNKNOWN)      return logic_type::UNKNOWN;
				if(b == logic_type::INCONSISTENT) return logic_type::INCONSISTENT;
				if(b == logic_type::TRUE)         return logic_type::INCONSISTENT;
			}

			if (a == logic_type::TRUE) {
				return b;
			}
		}

		/** @brief implication logic method
		 *
		 * @param a first logic value
		 * @param b second logic value
		 * @returns a -> b
		 */
		static logic_type::type IMPLICATION(logic_type::type a, logic_type::type b) {
			 // f < u < i < t
			
			if (a == logic_type::FALSE) {
				return logic_type::TRUE;
			}

			if (a == logic_type::UNKNOWN) {
				if(b == logic_type::FALSE)        return logic_type::UNKNOWN;
				if(b == logic_type::UNKNOWN)      return logic_type::UNKNOWN;
				if(b == logic_type::INCONSISTENT) return logic_type::INCONSISTENT;
				if(b == logic_type::TRUE)         return logic_type::TRUE;
			}

			if (a == logic_type::INCONSISTENT) {
				if(b == logic_type::FALSE)        return logic_type::INCONSISTENT;
				if(b == logic_type::UNKNOWN)      return logic_type::INCONSISTENT;
				if(b == logic_type::INCONSISTENT) return logic_type::INCONSISTENT;
				if(b == logic_type::TRUE)         return logic_type::TRUE;
			}

			if (a == logic_type::TRUE) {
				if(b == logic_type::FALSE)        return logic_type::FALSE;
				if(b == logic_type::UNKNOWN)      return logic_type::UNKNOWN;
				if(b == logic_type::INCONSISTENT) return logic_type::INCONSISTENT;
				if(b == logic_type::TRUE)         return logic_type::TRUE;
			}
		}

		/** @brief negates logic value
		 *
		 * @param a logic value
		 * @returns logic value that is negations of a
		 */
		static logic_type::type NOT(logic_type::type a) {
			if (a == logic_type::TRUE) {
				return logic_type::FALSE;
			}

			if (a == logic_type::FALSE) {
				return logic_type::TRUE;
			}
			return a;
		}

		/** @brief returns upper bound on truth ordering
		 *
		 * @param a first logic value
		 * @param b second logic value
		 * @returns upper bound of a and b with respect to truth ordering
		 */
		static logic_type::type TUP(logic_type::type a, logic_type::type b) {
			if (a == logic_type::TRUE || b == logic_type::TRUE) {
				return logic_type::TRUE;
			}

			if (a == logic_type::INCONSISTENT || b == logic_type::INCONSISTENT) {
				return logic_type::INCONSISTENT;
			}

			if (a == logic_type::UNKNOWN || b == logic_type::UNKNOWN) {
				return logic_type::UNKNOWN;
			}

			return logic_type::FALSE;
		}

		/** @brief returns MAX logic value wrt to f<u<i<t order
		 *
		 * @param a first logic value
		 * @param b second logic value
		 * @returns MAX logic value wrt to f<u<i<t order
		 */
		static logic_type::type MAX(logic_type::type a, logic_type::type b) {
			if (a == logic_type::FALSE) {
				return b;
			}

			if (a == logic_type::UNKNOWN) {
				if(b == logic_type::FALSE)        return logic_type::UNKNOWN;
				if(b == logic_type::UNKNOWN)      return logic_type::UNKNOWN;
				if(b == logic_type::INCONSISTENT) return logic_type::INCONSISTENT;
				if(b == logic_type::TRUE)         return logic_type::TRUE;
			}

			if (a == logic_type::INCONSISTENT) {
				if(b == logic_type::FALSE)        return logic_type::INCONSISTENT;
				if(b == logic_type::UNKNOWN)      return logic_type::INCONSISTENT;
				if(b == logic_type::INCONSISTENT) return logic_type::INCONSISTENT;
				if(b == logic_type::TRUE)         return logic_type::TRUE;
			}

			if (a == logic_type::TRUE)
				return a;
		}

		/** @brief returns MIN logic value wrt to f<u<i<t order
		 *
		 * @param a first logic value
		 * @param b second logic value
		 * @returns MIN logic value wrt to f<u<i<t order
		 */
		static logic_type::type MIN(logic_type::type a, logic_type::type b) {
			if (a == logic_type::TRUE) {
				return b;
			}

			if (a == logic_type::INCONSISTENT) {
				if(b == logic_type::FALSE)        return logic_type::FALSE;
				if(b == logic_type::UNKNOWN)      return logic_type::UNKNOWN;
				if(b == logic_type::INCONSISTENT) return logic_type::INCONSISTENT;
				if(b == logic_type::TRUE)         return logic_type::INCONSISTENT;
			}

			if (a == logic_type::UNKNOWN) {
				if(b == logic_type::FALSE)        return logic_type::FALSE;
				if(b == logic_type::UNKNOWN)      return logic_type::UNKNOWN;
				if(b == logic_type::INCONSISTENT) return logic_type::UNKNOWN;
				if(b == logic_type::TRUE)         return logic_type::UNKNOWN;
			}

			if (a == logic_type::FALSE)
				return a;
		}

		/** @brief converts logic value to string
		 *
		 * @param t logic value
		 * @returns string value of input logic value
		 */
		static std::string get_logic_name(Inter4ql::logic_type::type t) {
			switch (t) {
				case Inter4ql::logic_type::FALSE:
					return std::string("false");
				case Inter4ql::logic_type::TRUE:
					return std::string("true");
				case Inter4ql::logic_type::UNKNOWN:
					return std::string("unknown");
				case Inter4ql::logic_type::INCONSISTENT:
					return std::string("inconsistent");
				default:
					return "invalid";
			}
		}

		/** @brief craetes four-valued logic value from two booleans
		 *
		 * @param _t is true value
		 * @param _f is false value
		 * @returns four-valued logic value
		 */
		static Inter4ql::logic_type::type create(bool _t, bool _f) {
			if (_t && _f)
				return logic_type::INCONSISTENT;
			if (!_t && _f)
				return logic_type::FALSE;
			if (_t && !_f)
				return logic_type::TRUE;
			if (!_t && !_f)
				return logic_type::UNKNOWN;
		}

		/** @brief checks if input four-valued logic value is true
		 *
		 * @param i four-valued logic value
		 * @returns is value i true
		 */
		static bool is_true(logic_type::type i) {
			if (i == logic_type::TRUE || i == logic_type::INCONSISTENT)
				return true;
			return false;
		}

		/** @brief checks if input four-valued logic value is false
		 *
		 * @param i four-valued logic value
		 * @returns is value i false
		 */
		static bool is_false(logic_type::type i) {
			if (i == logic_type::FALSE || i == logic_type::INCONSISTENT)
				return true;
			return false;
		}

		/** @brief checks if first logic value contains more informations than the second
		 *
		 * @param a logic value to test
		 * @param b logic value to test
		 * @returns whether a contains more informations than b
		 */
		static bool is_more_knowledge(logic_type::type a, logic_type::type b)
		{
			if (a == b)
				return false;
			if (a == logic_type::INCONSISTENT)
				return true;
			if (b == logic_type::UNKNOWN)
				return true;
			return false;

		}
	private:

	};
}

#endif /* __LOGIC_H__ */

