#include "ModuleBelief.h"

#include "Application.h"
#include "Exception.h"
#include "Result.h"
#include "Fact.h"
#include "Domain.h"
#include "Relation.h"
#include "Disposable.h"
#include "Logic.h"
#include "Types.h"
#include "BooleanExpression.h"
#include "ModuleSum.h"

#include <memory>
#include <tuple>
#include <map>

namespace Inter4ql {

	ModuleBelief::ModuleBelief(std::string _name,
			Constraints* _constraints,
			std::vector<std::string>* _members)
	: members(_members),
	  name(_name),
	  constraints(_constraints),
	  disable_constraints(false)
	{
		this->modules = nullptr;
		this->domains = new std::vector<Domain*>();
		this->member_modules = new std::vector<ModuleGeneric *>();

		this->members->push_back("plugin");
		this->members->push_back("math");
		this->members->push_back("BEL");
	}

	ModuleBelief::~ModuleBelief() {
		if(!Disposable::errorMode) {
			delete this->constraints;
			delete this->members;
		}
		delete this->domains;
		delete this->member_modules;
	}

	variable_type::type ModuleBelief::get_domain_type(std::string name) {
		return variable_type::INVALID;
	}

	Relation *ModuleBelief::get_relation(std::string s) const {
		std::vector<Relation*> *relations = new std::vector<Relation*>();
		Relation* r;
		for (std::vector<ModuleGeneric *>::iterator it = this->member_modules->begin(); it != this->member_modules->end(); it++) {
			if ((*it)->get_name() == this->get_name())
				continue;
			try {
				r = (*it)->get_relation(s);
			} catch(Exception* ex) {
				delete ex;
				r = nullptr;
			};
			if (r) {
				relations->push_back(r);
			}
		}
		if(relations->size() == 0) {
			delete relations;
			throw new Exception(std::string("None of the worlds in belief '") + this->name + "' declare relation '" + s + "'! Referencing this relations cannot be therefore type-checked and executed.\n");
		}
		if(relations->size() == 1) {
			Relation* to_return = relations->at(0);
			delete relations;
			return to_return;
		}
		std::string rel = "";
		for (std::vector<Relation*>::iterator it = relations->begin(); it != relations->end(); it++) {
			if (rel == "") {
				rel = (*it)->print();
			} else {
				if ((*it)->print() != rel) {
					throw new Exception(std::string("Ambigous reference to relation '") + s + "' located in belief '" + this->name + "'! E.g. declarations from worlds " + relations->at(0)->get_module() + " and " + (*it)->get_module() + " differ!\n");
				}
			}
		}

		Relation* to_return = relations->at(0);
		delete relations;
		return to_return;
	}

	std::vector<std::string> *ModuleBelief::get_members() const {
		return this->members;
	}

	std::vector<Domain*> *ModuleBelief::get_domains() const {
		return this->domains;
	}

	std::string ModuleBelief::get_name() const {
		return this->name;
	}

	std::string ModuleBelief::print() {
		std::string s("");
		std::vector<Expression*>* softConstraints = this->constraints->get_soft_constraints();
		std::vector<Expression*>* hardConstraints = this->constraints->get_hard_constraints();
		s += "belief " + this->name + ":\n";
		if ((hardConstraints && hardConstraints->size()) ||
		    (softConstraints && softConstraints->size())) {
			s += "constraints:\n";	
		}
		if (hardConstraints && hardConstraints->size()) {
			s += "\thard:\n";	
			print_constraints(s, hardConstraints);
		}
		if (softConstraints && softConstraints->size()) {
			s += "\tsoft:\n";
			print_constraints(s, softConstraints);
		}
		if (this->members && this->members->size()) {
			s += "members:\n";
			for (int i = 0; i < this->members->size(); i++) {
				if (this->members->at(i) == "BEL")
					continue;
				s += "\t" + this->members->at(i) + "\n";
			}
		}
		s += "\n";
		return s;
	}

	module_type::type ModuleBelief::get_module_type() const {
		return module_type::BELIEF;
	}

	void ModuleBelief::add_domain(Domain *d) {
		return;
	}

	void ModuleBelief::add_relation(Relation *r) {
		return;
	}

	void ModuleBelief::reason()
	{

	}

	Result ModuleBelief::ask_for_result(Fact * term) const
	{
		//printf("ModuleBelief asking for result for: %s\n", term->print().c_str());

		Result to_return;
		if (!this->modules)
			return to_return;
		if (!check_constraints_local(true, true)) {
			return to_return;
		}

		bool skip_summing = false;
		for (std::vector<ModuleGeneric *>::iterator itMod = this->member_modules->begin(); itMod != this->member_modules->end(); itMod++) {
			bool skip_summing = false;
			Result r;
			// constraints already checked
			try {
				// this may fail as relation may not be in the module
				r = (*itMod)->ask_for_result(term);
			} catch(Exception* ex) {
				delete ex;
				skip_summing = true;
			}
			if (!skip_summing)
				to_return.sum_Results_OP(r, operation_type::SUM);
		}
		return to_return;
	}

	bool ModuleBelief::check_constraints_local(bool check_rigid, bool check_flexible) const
	{
		Ctraints to_check = new std::vector<Expression*>();
		if (check_rigid && this->get_constraints()->get_hard_constraints())
			to_check->insert(to_check->end(), this->get_constraints()->get_hard_constraints()->begin(), this->get_constraints()->get_hard_constraints()->end());
		if (check_flexible && this->get_constraints()->get_soft_constraints())
			to_check->insert(to_check->end(), this->get_constraints()->get_soft_constraints()->begin(), this->get_constraints()->get_soft_constraints()->end());
		bool is_valid = check_constraints(to_check, disable_constraints, get_domains(), this->member_modules, this);
		delete to_check;
		return is_valid;
	}

	std::vector<ModuleGeneric*>* ModuleBelief::get_member_modules()
	{
		return this->member_modules;
	}

	Result ModuleBelief::evaluate_expression_bf(Expression * expr) const
	{
		//printf("ModuleBelief::evaluate_expression_bf\n");

		Result to_return;
		if (!this->modules)
			return to_return;
		if (!check_constraints_local(true, true)) {
			return to_return;
		}

		// here we have to make a sum of sets...
		ModuleSum sum_of_worlds;

		// constraints already checked...
		sum_of_worlds.set_disable_constraints(true);

		for (std::vector<ModuleGeneric *>::iterator itMod = this->member_modules->begin(); itMod != this->member_modules->end(); itMod++) {
			sum_of_worlds.add_module(*itMod);
		}

		//printf("Sum of worlds: %s\n", sum_of_worlds.print().c_str());

		return sum_of_worlds.evaluate_expression(expr);
	}

	Result ModuleBelief::evaluate_expression(Expression * expr) const
	{
		Result to_return;
		if (!this->modules)
			return to_return;
		if (!check_constraints_local(true, true)) {
			return to_return;
		}

		bool skip_summing = false;
		for (std::vector<ModuleGeneric *>::iterator itMod = this->member_modules->begin(); itMod != this->member_modules->end(); itMod++) {
			bool skip_summing = false;
			Result r;
			// constraints already checked
			try {
				// this may fail as relation may not be in the module
				r = (*itMod)->evaluate_expression(expr);
			} catch(Exception* ex) {
				delete ex;
				skip_summing = true;
			}
			if (!skip_summing)
				to_return.sum_Results_OP(r, operation_type::SUM);
		}
		return to_return;
	}

	void ModuleBelief::set_modules(std::vector<ModuleGeneric *> *_modules) {
		this->modules = _modules;

		//printf("Set modules called in belief %s\n", name.c_str());

		delete this->domains;
		delete this->member_modules;
		this->domains = new std::vector<Domain*>();
		this->member_modules = new std::vector<ModuleGeneric *>();

		VariableSpace * vs = Application::app_instance->public_get_vs();
		//printf("Variable space: %s\n", vs->print().c_str());

		std::map<std::string, Inter4ql::variable_type::type> domain_map; //name -> type
		for (std::vector<std::string>::iterator it = this->members->begin(); it != this->members->end(); it++) {
			for (std::vector<ModuleGeneric *>::iterator itMod = this->modules->begin(); itMod != this->modules->end(); itMod++) {
				//printf("Checking member %s against module %s\n", it->c_str(), (*itMod)->get_name().c_str());
				if ((*itMod)->get_name() == (*it)) {
					//printf("Adding %s to member_modules of belief %s\n", (*itMod)->get_name().c_str(), name.c_str());
					member_modules->push_back(*itMod);
					std::vector<Domain*> *domains = (*itMod)->get_domains();
					if (!domains)
						break;
					for(std::vector<Domain*>::iterator itDom = domains->begin(); itDom != domains->end(); itDom++) {
						std::map<std::string, Inter4ql::variable_type::type>::iterator prev_dom = domain_map.find((*itDom)->get_name());
						if (prev_dom == domain_map.end()) {
							//there was not such domain on other modules
							domain_map.insert(std::pair<std::string, Inter4ql::variable_type::type>((*itDom)->get_name(), (*itDom)->get_type()));
							this->domains->push_back(*itDom);
						} else {
							if (prev_dom->second != (*itDom)->get_type()) {
								throw new Exception(std::string("Mismatch of types for domain '") + (*itDom)->get_name() + "' among worlds of belief '" + this->name + "'. Domains should have the same type among worlds.\n");
							}
						}
					}
					continue;
				}
			}
		}
		//printf("\n\n\n");
	}

	bool ModuleBelief::constraints_violated() {
		return !check_constraints_local(true, true);
	}

}
