#ifndef __MODULEGENERIC_H__
#define __MODULEGENERIC_H__

#include "Constraints.h"
#include "Types.h"

#include <string>
#include <vector>

namespace Inter4ql {

	class Fact;
	class Relation;
	class Rule;
	class Result;
	class Domain;
	class Expression;
	class Constraints;
	/** @brief Generic class of module (only virtual methods)
	 */
	class ModuleGeneric {
	public:
		/** @brief default constructor
		 */
		ModuleGeneric() {
			database_to_add = new database();
			database_to_remove = new database();
		}
		/** @brief default destructor
		 */
		virtual ~ModuleGeneric() {
			// TODO: memory leak!
			//delete database_to_add;
			//delete database_to_remove;
		};
		/** @brief getter for a relation
		 *
		 * @param s name of a relation
		 * @returns relation with name s inside module
		 */
		virtual Relation *get_relation(std::string s) const {}
		/** @brief getter for a name of module
		 *
		 * @returns module name
		 */
		virtual std::string get_name() const {}
		/** @brief function that returns string with module information
		 *
		 * @return module information packed in a std::string
		 */
		virtual std::string print() {}
		/** @brief getter for module relations
		 *
		 * @return module relations
		 */
		virtual std::vector<Relation*> *get_relations() const {}
		/** @brief getter for module type
		 *
		 * @return module type
		 */
		virtual module_type::type get_module_type() const {}
		/** @brief performs reasoning
		 *
		 */
		virtual void reason() = 0;
		/** @brief performs reasing with additional Rules added
		 *
		 * @param rules list of Rules to be added before reasoning
		 */
		virtual void reason_with_rules(std::vector<Rule*>* rules) {};
		/** @brief provides valuations matching given Fact
		 *
		 * @param term query to find in database
		 * @return possible valuations of term
		 */
		virtual Result ask_for_result(Fact * term) const = 0;
		/** @brief provides valuations matching given Expression
		 *
		 * @param expr query to find in database
		 * @return possible valuations of expression
		 */
		virtual Result evaluate_expression(Expression* expr) const;
		/** @brief provides bel-free valuations matching given Expression
		 *
		 * @param expr query to find in database
		 * @return possible valuations of expression
		 */
		virtual Result evaluate_expression_bf(Expression* expr) const;
		/** @brief provides a list of Module references is term (useful when e.g. term is a shadowed Bel)
		 *
		 * @param term Fact to be checked
		 * @return list of referenced Modules
		 */
		virtual std::vector<std::string> get_references(Fact * term);
		/** @brief setter for modules needed in this module
		 *
		 * @param _modules list of modules
		 */
		virtual void set_modules(std::vector<ModuleGeneric *> *_modules) {}
		/** @brief getter for constraints
		 *
		 * @return list of constraints
		 */
		virtual Constraints* get_constraints() const { return nullptr; }
		/** @brief getter for domains
		 *
		 * @return list of domains
		 */
		virtual std::vector<Domain*> *get_domains() const { return nullptr; };
		/** @brief check if module contains domain
		 *
		 * @return whether module contains the domain
		 */
		virtual bool contains_domain(Domain* d) const;
		/** @brief check if module contains relation (ignoring module name)
		 *
		 * @return whether module contains the relation
		 */
		virtual bool contains_relation(Relation* d) const;
		/** @brief allows for disabling constraints check
		 *
		 * @param value value to be set
		 */
		virtual void set_disable_constraints(bool value) const {};
		/** @brief evaluates constraints locally, in current module
		 *
		 * @param check_rigid whether rigid constrains should be checked
		 * @param check_flexible whether flexible constraints should be checked
		 * @return whether constraints are valid
		 */
		virtual bool check_constraints_local(bool check_rigid, bool check_flexible) const { return true; };
		/** @brief fetches a database of facts from well-supported model (for summing purposes)
		 *
		 * @return database of wsm facts
		 */
		virtual database* get_wsm_database() { return nullptr; }
		/** @brief fetches a database of facts which were added by actions to this module
		 *
		 * @return database of added facts
		 */
		virtual database* get_add_database() { return database_to_add; };
		/** @brief fetches a database of facts which were removed by actions from this module
		 *
		 * @return database of removed facts
		 */
		virtual database* get_remove_database() { return database_to_remove; };
		/** @brief overrides a database of facts which were added by actions to this module
		 *
		 * @param db database to be set
		 */
		virtual void set_add_database(database* db) { database_to_add = db; };
		/** @brief overrides a database of facts which were removed by actions from this module
		 *
		 * @param db database to be set
		 */
		virtual void set_remove_database(database* db) { database_to_remove = db; };
		/** @brief checks whether constraints are violated
		 *
		 * @return true if constraints are vioilated, false otherwise
		 */
		virtual bool constraints_violated() { return false; };

	protected:
		// below databases are used for planning only
		Inter4ql::database* database_to_add;
		Inter4ql::database* database_to_remove;
	};
}

#endif
