#include "ModuleGroup.h"
#include "Application.h"
#include "Exception.h"
#include "Result.h"
#include "Fact.h"
#include "Domain.h"
#include "Relation.h"
#include "Disposable.h"
#include "Logic.h"
#include "BooleanExpression.h"

#include <cassert>
#include <memory>
#include <tuple>
#include <map>

namespace Inter4ql {

	ModuleGroup::ModuleGroup(std::string _name,
			std::vector<Domain*> *_domains,
			std::vector<Relation*> *_relations,
			std::vector<std::string> *_members)
	: name(_name)
	, domains(_domains)
	{
		this->domainsToDelete = new std::vector<Domain*>();
		this->domainsToDelete->push_back(new Domain("integer", variable_type::INTEGER));
		this->domainsToDelete->push_back(new Domain("real", variable_type::REAL));
		this->domainsToDelete->push_back(new Domain("string", variable_type::STRING));
		this->domainsToDelete->push_back(new Domain("literal", variable_type::ID));
		this->domainsToDelete->push_back(new Domain("date", variable_type::DATE));
		this->domainsToDelete->push_back(new Domain("date_time", variable_type::DATE_TIME));
		this->domainsToDelete->push_back(new Domain("logic", variable_type::LOGIC));
		for(std::vector<Domain*>::iterator it = this->domainsToDelete->begin(); it != this->domainsToDelete->end(); it++) {
			this->domains->push_back(*it);
		}
		this->relations = _relations;
		this->members = _members;
		this->true_when = 0;
		this->false_when = 0;
		this->inconsistent_when = 0;
	}

	ModuleGroup::~ModuleGroup() {
		if(!Disposable::errorMode) {
			for(std::vector<Domain*>::iterator it = this->domains->begin(); it != this->domains->end(); it++) {
				delete (*it);
			}
			delete this->domains;
			delete this->domainsToDelete;

			for(std::vector<Relation*>::iterator it = this->relations->begin(); it != this->relations->end(); it++) {
				delete (*it);
			}
			delete this->relations;
		} else {
			for(std::vector<Domain*>::iterator it = this->domainsToDelete->begin(); it != this->domainsToDelete->end(); it++) {
				delete (*it);
			}
			delete this->domainsToDelete;
		}
		if (true_when != 0) delete true_when;
		true_when = 0;
		if (false_when != 0) delete false_when;
		false_when = 0;
		if (inconsistent_when != 0) delete inconsistent_when;
		inconsistent_when = 0;
	}

	variable_type::type ModuleGroup::get_domain_type(std::string name) {
		for (int i = 0; i < this->domains->size(); i++) {
			if (this->domains->at(i)->get_name() == name)
				return this->domains->at(i)->get_type();
		}
		return variable_type::INVALID;
	}

	Relation *ModuleGroup::get_relation(std::string s) const {
		for (int i = 0; i < this->relations->size(); i++)
			if (this->relations->at(i)->get_name() == s)
				return this->relations->at(i);
		throw new Exception(std::string("Relation ") + s + " not found in module " + this->name + "!\n");
	}

	std::vector<Relation*> *ModuleGroup::get_relations() const {
		return this->relations;
	}

	std::vector<std::string> *ModuleGroup::get_members() const {
		return this->members;
	}

	std::vector<Domain*> *ModuleGroup::get_domains() const {
		return this->domains;
	}

	std::string ModuleGroup::get_name() const {
		return this->name;
	}

	std::string ModuleGroup::print() {
		std::string s("");
		s += "group " + this->name + ":\n";
		if (this->relations->size())
			s += "relations:\n";
		for (int i = 0; i < this->relations->size(); i++)
			s += "\t" + this->relations->at(i)->print() + "\n";
		s += "\n";
		return s;
	}

	module_type::type ModuleGroup::get_module_type() const {
		return module_type::GROUP;
	}

	void ModuleGroup::add_domain(Domain *d) {
		this->domains->push_back(d);
		return;
	}

	void ModuleGroup::add_relation(Relation *r) {
		this->relations->push_back(r);
		return;
	}

	void ModuleGroup::reason()
	{

	}

	Result ModuleGroup::ask_for_result_generic(Fact * term, Expression * expr) const {
		std::map <std::vector <Value*>, std::tuple<int, int, int, int>> results_distribution;
		std::vector <std::string> variable_names;

		for (const auto& name : *members) {
			ModuleGeneric* mg = Application::app_instance->public_get_module(name);
			Result partial_result;
			if (expr) {
				assert(term == nullptr);
				partial_result = mg->evaluate_expression(expr);
			} else {
				assert(term != nullptr);
				// "ask_for_result" does not apply set query!
				partial_result = mg->ask_for_result(term);
			}

			variable_names = partial_result.get_variable_names();
			for (const auto& val : partial_result.get_valuation())
			{
				auto it = results_distribution.begin();
				for (; it != results_distribution.end(); it++)
				{
					//auto it = results_distribution.find(val.first); // manual search required
					// it->first - vector <Value*>
					// it->second - <t,i,f,u>
					if (it->first.size() != val.first.size()) continue;
					bool found = true;
					for (int i = 0; i < val.first.size(); i++)
					{
						if ( *it->first[i] != *val.first[i] )
						{
							found = false;
							break;
						}
					}
					if (found)
						break;
				}
				if (it == results_distribution.end())
					std::tie(it, std::ignore) = results_distribution.insert(std::make_pair(val.first, std::make_tuple(0,0,0,0)));

				if (val.second == logic_type::UNKNOWN)
					std::get<0>(it->second) ++ ;
				else if (val.second == logic_type::TRUE)
					std::get<1>(it->second) ++ ;
				else if (val.second == logic_type::FALSE)
					std::get<2>(it->second) ++ ;
				else if (val.second == logic_type::INCONSISTENT)
					std::get<3>(it->second) ++ ;
			}
		}

		Result ret(variable_names);

		for (const auto& it : results_distribution)
		{
			int u, t, f, i;
			std::tie(u,t,f,i) = it.second;
			u = this->members->size() - t - f - i;

			logic_type::type result_value = logic_type::UNKNOWN;

			if (inconsistent_when == 0 && true_when == 0 && false_when == 0) // we will use default values
			{
				if (i > 0) result_value = Logic::OR_RESULT(result_value, logic_type::INCONSISTENT);
				if (t > 0) result_value = Logic::OR_RESULT(result_value, logic_type::TRUE);
				if (f > 0) result_value = Logic::OR_RESULT(result_value, logic_type::FALSE);

			}
			else {
				if (inconsistent_when != 0 && inconsistent_when->calculate(i, t, f, u) > 0)
					result_value = Logic::OR_RESULT(result_value, logic_type::INCONSISTENT);

				if (true_when != 0 && true_when->calculate(i, t, f, u) > 0)
					result_value = Logic::OR_RESULT(result_value, logic_type::TRUE);

				if (false_when != 0 && false_when->calculate(i, t, f, u) > 0)
					result_value = Logic::OR_RESULT(result_value, logic_type::FALSE);
			}

			ret.add_element(it.first, result_value);
		}

		return ret;
	}

	Result ModuleGroup::ask_for_result(Fact * term) const
	{
		return ask_for_result_generic(term, nullptr);
	}

	Result ModuleGroup::evaluate_expression(Expression * expr) const
	{
		return ask_for_result_generic(nullptr, expr);
	}
	void ModuleGroup::add_whens(BooleanExpression * t, BooleanExpression * f, BooleanExpression * i)
	{
		if (true_when != 0) delete true_when;
		true_when = t;
		if (false_when != 0) delete false_when;
		false_when = f;
		if (inconsistent_when != 0) delete inconsistent_when;
		inconsistent_when = i;

	}
}
