#include "ModuleLocal.h"

#include "Exception.h"
#include "Fact.h"
#include "Functions.h"
#include "Rule.h"
#include "Result.h"
#include "Domain.h"
#include "Relation.h"
#include "Disposable.h"
#include "Application.h"

#include <memory>
#include <utility>
#include <set>

#include <cstdio>

namespace Inter4ql {

	ModuleLocal::ModuleLocal(std::string _name,
			Constraints* _constraints,
			std::vector<Domain*> *_domains,
			std::vector<Relation*> *_relations,
			std::vector<Rule*> *_rules,
			std::vector<Fact*> *_facts)
	: name(_name)
	, constraints(_constraints)
	, domains(_domains)
	, relations(_relations)
	, facts(_facts)
	, rules(_rules)
	, disable_constraints(false)
	{
		this->domainsToDelete = new std::vector<Domain*>();
		this->domainsToDelete->push_back(new Domain("integer", variable_type::INTEGER));
		this->domainsToDelete->push_back(new Domain("real", variable_type::REAL));
		this->domainsToDelete->push_back(new Domain("string", variable_type::STRING));
		this->domainsToDelete->push_back(new Domain("literal", variable_type::ID));
		this->domainsToDelete->push_back(new Domain("date", variable_type::DATE));
		this->domainsToDelete->push_back(new Domain("date_time", variable_type::DATE_TIME));
		this->domainsToDelete->push_back(new Domain("logic", variable_type::LOGIC));
		for(std::vector<Domain*>::iterator it = this->domainsToDelete->begin(); it != this->domainsToDelete->end(); it++) {
			this->domains->push_back(*it);
		}
	}

	ModuleLocal::~ModuleLocal() {
		if(!Disposable::errorMode) {
			delete this->constraints;

			for(std::vector<Domain*>::iterator it = this->domains->begin(); it != this->domains->end(); it++) {
				delete (*it);
			}
			delete this->domains;
			delete this->domainsToDelete;

			for(std::vector<Relation*>::iterator it = this->relations->begin(); it != this->relations->end(); it++) {
				delete (*it);
			}
			delete this->relations;

			for(std::vector<Rule*>::iterator it = this->rules->begin(); it != this->rules->end(); it++) {
				delete (*it);
			}
			delete this->rules;

			for(std::vector<Fact*>::iterator it = this->facts->begin(); it != this->facts->end(); it++) {
				for(std::vector<Value*>::iterator it1 = (*it)->get_params()->begin(); it1 != (*it)->get_params()->end(); it1++) {
					if((*it1)->get_type() == variable_type::VARIABLE)
						delete (*it1);
				}

					delete (*it);
			}
			delete this->facts;
		} else {
			for(std::vector<Domain*>::iterator it = this->domainsToDelete->begin(); it != this->domainsToDelete->end(); it++) {
				delete (*it);
			}
			delete this->domainsToDelete;
		}
	}

	variable_type::type ModuleLocal::get_domain_type(std::string name) {
		for (int i = 0; i < this->domains->size(); i++) {
			if (this->domains->at(i)->get_name() == name)
				return this->domains->at(i)->get_type();
		}
		return variable_type::INVALID;
	}

	void ModuleLocal::generate_facts() {
		for (int i = 0; i < this->facts->size(); i++) {
			this->add_fact(this->facts->at(i));
		}
		return;
	}

	Relation *ModuleLocal::get_relation(std::string s) const {
		for (int i = 0; i < this->relations->size(); i++)
			if (this->relations->at(i)->get_name() == s)
				return this->relations->at(i);
		throw new Exception(std::string("Relation ") + s + " not found in module " + this->name + "!\n");
	}

	std::vector<Relation*> *ModuleLocal::get_relations() const {
		return this->relations;
	}

	std::vector<Fact*> *ModuleLocal::get_facts() const {
		return this->facts;
	}

	std::vector<Domain*> *ModuleLocal::get_domains() const {
		return this->domains;
	}

	std::string ModuleLocal::get_name() const {
		return this->name;
	}

	std::string ModuleLocal::print() {
		std::string s("");
		s += "module " + this->name + ":\n";
		std::vector<Expression*>* softConstraints = this->constraints->get_soft_constraints();
		std::vector<Expression*>* hardConstraints = this->constraints->get_hard_constraints();
		if ((hardConstraints && hardConstraints->size()) ||
		    (softConstraints && softConstraints->size())) {
			s += "constraints:\n";
		}
		if (hardConstraints && hardConstraints->size()) {
			s += "\thard:\n";
			print_constraints(s, hardConstraints);
		}
		if (softConstraints && softConstraints->size()) {
			s += "\tsoft:\n";
			print_constraints(s, softConstraints);
		}
		if (this->relations->size())
			s += "relations:\n";
		for (int i = 0; i < this->relations->size(); i++)
			s += "\t" + this->relations->at(i)->print() + "\n";
		if (this->facts->size())
			s += "beginning facts:\n";
		for (int i = 0; i < this->facts->size(); i++)
			s += "\t" + this->facts->at(i)->print() + "\n";
		if (this->rules->size())
			s += "rules:\n";
		for (int i = 0; i < this->rules->size(); i++)
			s += "\t" + this->rules->at(i)->print() + "\n";
		s += "\n";
		return s;
	}

	module_type::type ModuleLocal::get_module_type() const {
		return module_type::LOCAL;
	}

	std::vector<Rule*> *ModuleLocal::get_rules() const {
		return this->rules;
	}

	void ModuleLocal::add_domain(Domain *d) {
		this->domains->push_back(d);
		return;
	}

	void ModuleLocal::add_relation(Relation *r) {
		this->relations->push_back(r);
		return;
	}

	void ModuleLocal::add_fact(Fact *r) {
		this->facts->push_back(r);
		return;
	}

	void ModuleLocal::add_rule(Rule *r) {
		this->rules->push_back(r);
		return;
	}

	void ModuleLocal::reason_with_rules(std::vector<Rule*>* rules) {
		for (auto rule : *rules) {
			this->rules->push_back(rule);
		}
		reason();
		for (auto rule : *rules) {
			this->rules->pop_back();
		}
	}

	void ModuleLocal::reason() {
		initialize_database();

		Inter4ql::database initial_database;
		do {
		// deep copy
		initial_database.clear();
		for (const auto& it : this->database_)
		{
			initial_database.insert(std::make_pair(it.first, std::make_shared<Valuation>(*it.second) ));
		}

		bool new_facts_added = false;
		do {
			new_facts_added = false;
			for (const auto& rule : *rules) {
				new_facts_added |= reason_about_rule(rule, logic_type::TRUE);
			}
		} while (new_facts_added);

		auto first_database = this->database_;

		initialize_database();

		for (const auto& rule : *rules) {
			std::string relation = rule->get_head()->get_relation();
			auto place = this->database_.find(relation);
			auto first_place = first_database.find(relation);
			(*place->second).remove_inconsistent(*first_place->second);
		}

		new_facts_added = false;
		do {
			new_facts_added = false;
			for (const auto& rule : *rules) {
				std::string relation = rule->get_head()->get_relation();
				auto place = this->database_.find(relation);
				auto first_place = first_database.find(relation);
				int counter = 0;
				for (const auto& element : *place->second) {
					if (element.second == logic_type::INCONSISTENT)
						counter += 2;
					else {
						counter += 1;
					}
				}
				/*new_facts_added |=*/ reason_about_rule(rule, logic_type::TRUE);
				(*place->second).remove_inconsistent(*first_place->second);

				int new_counter = 0; // ideally this would be done some elegant way
				for (const auto& element : *place->second) {
					if (element.second == logic_type::INCONSISTENT)
						new_counter += 2;
					else {
						new_counter += 1;
					}
				}
				new_facts_added |= counter != new_counter;
			}
		} while (new_facts_added);

		for (const auto& position : first_database)
		{
			for (const auto& value : *position.second)
			{
				if (value.second == logic_type::INCONSISTENT)
				{
					auto place = this->database_.find(position.first);
					//assert(place != this->database.end());
					(*place->second).add_element(value.first, value.second);
				}
			}
		}

		new_facts_added = false;
		do {
			new_facts_added = false;
			for (const auto& rule : *rules) {
				new_facts_added |= reason_about_rule(rule, logic_type::INCONSISTENT);
			}
		} while (new_facts_added);

		} while (database_fingerprint(initial_database) != database_fingerprint(this->database_));

		for (const auto& relation : *relations)
		{
			generate_unknown_facts(relation);
		}

		bool debug_database = Application::app_instance->debug_database;
		if (debug_database) print_database();
	}

	bool ModuleLocal::check_constraints_local(bool check_rigid, bool check_flexible) const
	{
		Ctraints to_check = new std::vector<Expression*>();
		if (check_rigid && this->get_constraints()->get_hard_constraints())
			to_check->insert(to_check->end(), this->get_constraints()->get_hard_constraints()->begin(), this->get_constraints()->get_hard_constraints()->end());
		if (check_flexible && this->get_constraints()->get_soft_constraints())
			to_check->insert(to_check->end(), this->get_constraints()->get_soft_constraints()->begin(), this->get_constraints()->get_soft_constraints()->end());;
		//std::vector<ModuleGeneric*>* module_list = new std::vector<ModuleGeneric*>();
		std::vector<ModuleGeneric*>* global_modules = Application::app_instance->public_get_modules();
		//module_list->push_back((ModuleGeneric*)(this));
		//module_list->push_back(global_modules->at(0)); // Plugin
		//module_list->push_back(global_modules->at(1)); // Math
		//module_list->push_back(global_modules->at(2)); // BEL
		bool is_valid = check_constraints(to_check, disable_constraints, get_domains(), global_modules, this);
		//delete module_list;

		delete to_check;
		return is_valid;
	}

	Result ModuleLocal::ask_for_result(Fact * term) const
	{
		//printf("ModuleLocal asking for result for: %s\n", term->print().c_str());
		auto variable_names = term->get_variable_names();
		Result res(variable_names);

		if (!Application::app_instance->public_get_modules())
			return Result();
		if (!check_constraints_local(true, true))
			return Result();

		bool valuation_added = false;

		valuation_added |= ask_for_fact_in_database(&database_, term, res, database_to_remove);
		valuation_added |= ask_for_fact_in_database(database_to_add, term, res, database_to_remove, false);

		if (!valuation_added)
			return Result();

		//printf("ModuleLocal responding with: %s\n", res.print(true, true).c_str());
		return res;
	}

	bool ModuleLocal::add_to_variable_space(std::string module, std::string relation, std::vector<Value*> val)
	{
		VariableSpace * vs = Application::app_instance->public_get_vs();
		bool did_vs_change = false;

		for (int i = 0; i < val.size(); i++)
		{
			if (val.at(i)->get_type() != variable_type::VARIABLE)
			{
				std::string domain = "";
				for(const auto& it : *this->relations)
				{
					if(it->get_name() == relation)
					{
						domain = it->get_params()->at(i)->get_name();
						break;
					}
				}
				if(domain == "")
					throw new Exception(std::string("Relation " + relation + " could not be found!\n"));

				Value *v = val.at(i);
				vs->get_pointer(v, module, domain, did_vs_change);
			}
		}

		return did_vs_change;
	}

	bool ModuleLocal::reason_about_rule(Rule * rule, logic_type::type target_logic_value)
	{
		//printf("Reasoning about rule %s\n", rule->print().c_str());
		std::shared_ptr<Result> dysjunction_result = std::make_shared<Result>();
		for (const auto& conjunction : *rule->get_body())
		{
			std::shared_ptr<Result> conjunction_result = std::make_shared<Result>();
			conjunction_result->add_element( {}, logic_type::TRUE );
			for (const auto& term : *conjunction)
			{
				Result term_result;

				if (term->get_module() != get_name()) {
					ModuleGeneric* mg = Application::app_instance->public_get_module(term->get_module());
					if (mg->get_module_type() == module_type::INTERNAL || mg->get_module_type() == module_type::EXTERNAL)
					{
						Result* till_now = conjunction_result->trim_variables(term->get_variable_names());

						for (const auto& elem : till_now->get_valuation())
						{
							if (elem.second == logic_type::UNKNOWN || elem.second == logic_type::FALSE)
								continue;
							Fact copy_of_term (term);
							// make map from elem & till_now variables
							std::map<std::string, Value*> binding;
							for (int i = 0; i < till_now->get_variable_names().size(); i++)
							{
								binding.insert(make_pair(till_now->get_variable_names()[i], elem.first[i]));
							}
							copy_of_term.valuate_fact(binding);
							Result temp_res = mg->ask_for_result(&copy_of_term);

							// we have to pack it into proper result with proper variable assign
							Result toCheck(term->get_variable_names());
							std::vector<Value*> assigned_values;
							for(auto var : term->get_variable_names()) {
								assigned_values.push_back(binding[var]);
							}
							toCheck.add_element(assigned_values, temp_res.get_valuation().get_first_logic_value());

							term_result.sum_Results_OP(toCheck, operation_type::SUM);
							//printf("Got result INTERNAL %s\n", term_result.print(true, false).c_str());
						}
					} else {
						term_result = mg->ask_for_result(term);
						//printf("Got result %s\n", term_result.print(true, false).c_str());
					}
				} else {
					term_result = ask_for_result(term);
					//printf("Got result no module name: %s\n", term_result.print(true, false).c_str());
				}

				if (term->get_value() == logic_type::FALSE)
				{
					for (auto& t : term_result.get_valuation_destructive())
					{
						t.second = Logic::NOT(t.second);
					}
				}

				if (term->is_set_query())
				{
					std::string s = term->get_set_string();
					std::map<logic_type::type, bool> in;
					in.insert(std::make_pair(logic_type::UNKNOWN      , s[0] == '1'));
					in.insert(std::make_pair(logic_type::FALSE        , s[1] == '1'));
					in.insert(std::make_pair(logic_type::TRUE         , s[2] == '1'));
					in.insert(std::make_pair(logic_type::INCONSISTENT , s[3] == '1'));
					for (auto& t : term_result.get_valuation_destructive())
					{
						if (in.find(t.second)->second)
							t.second = logic_type::TRUE;
						else
							t.second = logic_type::FALSE;
					}
				}

				//printf("Result after negation/set modifications: %s\n", term_result.print(true, false).c_str());

				// so at this moment we have all data about current term in term_result
				// now we need to AND this with data from previous terms from this conjunction

				conjunction_result->sum_Results_OP(term_result, operation_type::AND);
				//std::cout << "Sum up till now is: " << conjunction_result->print(true, true) << std::endl;
				//std::cout << "\n---------------------------------------\n";
			}
			std::set<std::string> variables_from_head;
			auto params = rule->get_head()->get_params();
			for (const auto& param : *params) {
				if ( (param)->get_type() == Inter4ql::variable_type::VARIABLE)
				{
					variables_from_head.insert((param)->get_value_string());
				}
			}

			conjunction_result = std::shared_ptr<Result>(conjunction_result->trim_variables(rule->get_head()->get_variable_names()));
			//printf("Trimmed conjunction result: %s\n", conjunction_result->print(true, false).c_str());

			dysjunction_result->sum_Results_OP(*conjunction_result, operation_type::OR);
			//printf("Dysjunction result: %s\n", dysjunction_result->print(true, false).c_str());
		}

		// lets find correct slot in DB
		auto slot = database_.find(rule->get_head()->get_relation());
		if (slot == database_.end()) {
			throw new Exception("Relation " + rule->get_head()->get_relation() + " not found in database.\n");
		}

		bool return_val = false;

		for (const auto& val : dysjunction_result->get_valuation())
		{
			//if (val.second == logic_type::TRUE || val.second == logic_type::INCONSISTENT) {
			if (val.second == target_logic_value) {

				//let's create binding
				std::map<std::string, Value*> binding;
				for (int i = 0; i < dysjunction_result->get_variable_names().size(); i++)
				{
					binding.insert(make_pair(dysjunction_result->get_variable_names()[i], val.first[i]));
				}

				auto target_valuation = apply_bind(rule->get_head()->get_params(), binding);
				if (add_to_variable_space(rule->get_head()->get_module(), rule->get_head()->get_relation(), target_valuation))
					generate_domain_facts(false);
				logic_type::type value = val.second;
				if (rule->get_head()->get_value() == logic_type::FALSE)
					value = Logic::NOT(value);

				return_val |= slot->second->add_element(target_valuation, value);

				Fact* temp = new Fact(rule->get_head());
				temp->valuate_fact(binding);
				remove_from_database(database_to_remove, temp);
				delete temp;
			}
		}
		return return_val;
	}

	void ModuleLocal::generate_unknown_facts(Relation * relation) // TODO UGLY, fix when variable_space is fixed
	{
		Valuation to_add;
		to_add.add_element( {}, logic_type::UNKNOWN);
		VariableSpace * vs = Application::app_instance->public_get_vs();
		for (const auto& param : *relation->get_params())
		{
			std::vector<Value*> attrs;
			if (param->get_type() == Inter4ql::variable_type::INTEGER) {
				std::map<int, Value*>* temp = vs->get_integer_space(relation->get_module(), param->get_name());
				for (std::map<int, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					attrs.push_back(i->second);
				}
			} else if (param->get_type() == Inter4ql::variable_type::REAL) {
				std::map<double, Value*>* temp = vs->get_real_space(relation->get_module(), param->get_name());
				for (std::map<double, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					attrs.push_back(i->second);
				}
			} else if (param->get_type() == Inter4ql::variable_type::STRING) {
				std::map<std::string, Value*>* temp = vs->get_string_space(relation->get_module(), param->get_name());
				for (std::map<std::string, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					attrs.push_back(i->second);
				}
			} else if (param->get_type() == Inter4ql::variable_type::ID) {
				std::map<std::string, Value*>* temp = vs->get_const_space(relation->get_module(), param->get_name());
				for (std::map<std::string, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					attrs.push_back(i->second);
				}
			} else if (param->get_type() == Inter4ql::variable_type::DATE) {
				std::map<std::string, Value*>* temp = vs->get_date_space(relation->get_module(), param->get_name());
				for (std::map<std::string, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					attrs.push_back(i->second);
				}
			} else if (param->get_type() == Inter4ql::variable_type::DATE_TIME) {
				std::map<std::string, Value*>* temp = vs->get_datetime_space(relation->get_module(), param->get_name());
				for (std::map<std::string, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					attrs.push_back(i->second);
				}
			} else if (param->get_type() == Inter4ql::variable_type::LOGIC) {
				std::map<int, Value*>* temp = vs->get_logic_space(relation->get_module(), param->get_name());
				for (std::map<int, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					attrs.push_back(i->second);
				}
			} else if (param->get_type() == Inter4ql::variable_type::VARIABLE)
				throw new Exception(std::string("Trying to generate unknown fact with VARIABLE parameter type!\n"));

			to_add.expand_by(attrs);
		}

		auto it = database_.find(relation->get_name());
		if (it == database_.end()) {
			throw new Exception("Relation " + relation->get_name() + " not found in database.\n");
		}
		std::shared_ptr<Valuation> target_relation = it->second;

		for (const auto& elem : to_add)
		{
			target_relation->add_element(elem.first, elem.second);
		}
	}

	void ModuleLocal::generate_domain_facts(bool should_initialize)
	{
		VariableSpace * vs = Application::app_instance->public_get_vs();
		// Adding domains as single argument relations
		for (const auto& it : *domains) {
			if (should_initialize) {
				std::vector<Domain *>* domain_vector = new std::vector<Domain*>();
				domain_vector->push_back(new Domain(*it));
				Relation * r = new Relation(it->get_name(), domain_vector, get_name());
				add_relation(r);

				// insert into database
				if (this->database_.find(it->get_name()) == this->database_.end())
					this->database_.insert(std::make_pair(it->get_name(), std::make_shared<Valuation>()));
			}

			auto val = this->database_.find(it->get_name())->second;
			// select correct type in vs
			// for each element in selected vector add one <this value, true> to valuation
			if (it->get_type() == Inter4ql::variable_type::INTEGER) {
				std::map<int, Value*>* temp = vs->get_integer_space(get_name(), it->get_name());
				for (std::map<int, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					std::vector<Value *> v;
					v.push_back(i->second);
					val->add_element(v, logic_type::TRUE);
				}
			} else if (it->get_type() == Inter4ql::variable_type::REAL) {
				std::map<double, Value*>* temp = vs->get_real_space(get_name(), it->get_name());
				for (std::map<double, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					std::vector<Value *> v;
					v.push_back(i->second);
					val->add_element(v, logic_type::TRUE);
				}
			} else if (it->get_type() == Inter4ql::variable_type::STRING) {
				std::map<std::string, Value*>* temp = vs->get_string_space(get_name(), it->get_name());
				for (std::map<std::string, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					std::vector<Value *> v;
					v.push_back(i->second);
					val->add_element(v, logic_type::TRUE);
				}
			} else if (it->get_type() == Inter4ql::variable_type::ID) {
				std::map<std::string, Value*>* temp = vs->get_const_space(get_name(), it->get_name());
				for (std::map<std::string, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					std::vector<Value *> v;
					v.push_back(i->second);
					val->add_element(v, logic_type::TRUE);
				}
			} else if (it->get_type() == Inter4ql::variable_type::DATE) {
				std::map<std::string, Value*>* temp = vs->get_date_space(get_name(), it->get_name());
				for (std::map<std::string, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					std::vector<Value *> v;
					v.push_back(i->second);
					val->add_element(v, logic_type::TRUE);
				}
			} else if (it->get_type() == Inter4ql::variable_type::DATE_TIME) {
				std::map<std::string, Value*>* temp = vs->get_datetime_space(get_name(), it->get_name());
				for (std::map<std::string, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					std::vector<Value *> v;
					v.push_back(i->second);
					val->add_element(v, logic_type::TRUE);
				}
			} else if (it->get_type() == Inter4ql::variable_type::LOGIC) {
				std::map<int, Value*>* temp = vs->get_logic_space(get_name(), it->get_name());
				for (std::map<int, Value*>::iterator i = temp->begin(); i != temp->end(); i++) {
					std::vector<Value *> v;
					v.push_back(i->second);
					val->add_element(v, logic_type::TRUE);
				}
			} else if (it->get_type() == Inter4ql::variable_type::VARIABLE)
				throw new Exception(std::string("Attempt to generate domain relation from variable\n"));
		}
	}

	void ModuleLocal::print_database() {
		// typedef std::unordered_multimap<std::string, std::shared_ptr<Valuation> > database;
		printf("Database for module %s\n\n", this->get_name().c_str());

		for(auto entry : this->database_) {
			printf("Key %s:\n", entry.first.c_str());
			entry.second->print();
		}
	}

	void ModuleLocal::initialize_database()
	{
		this->database_.clear();

		for (const auto& relation : *(this->relations))
			this->database_.insert(std::make_pair(relation->get_name(), std::make_shared<Valuation>()));

		for (const auto& it : *facts) {
			auto place = database_.find(it->get_relation());
			if (place == database_.end()) {
				throw new Exception("Relation " + it->get_relation() + " not found in database.\n");
			}

			(*place->second).add_element(*it->get_params(), it->get_value());
		}

		generate_domain_facts(false);
	}

	int ModuleLocal::database_fingerprint(Inter4ql::database db)
	{
		int counter = 0;
		for (const auto& valuation : db)
		{
			for (const auto& val : *valuation.second)
			{
				if (val.second == logic_type::INCONSISTENT)
				{
					counter += 2;
				}
				else if (val.second == logic_type::TRUE || val.second == logic_type::FALSE)
				{
					counter += 1;
				}
			}
		}
		return counter;
	}

	bool ModuleLocal::constraints_violated() {
		return !check_constraints_local(true, true);
	}
}
