#ifndef __MODULELOCAL_H__
#define __MODULELOCAL_H__

#include "ModuleGeneric.h"

#include "Constraints.h"
#include "Types.h"

#include <map>
#include <memory>
#include <string>
#include <unordered_map>
#include <vector>

namespace Inter4ql {

	class Rule;
	class Domain;
	class Relation;
	class Result;
	class Fact;
	/** @brief Implementation of ModuleGeneric for modules with reasoning
	 */
	class ModuleLocal : public ModuleGeneric {
	public:
		/** @brief constructor
		 *
		 * @param a_name name of a module
		 * @param _constraints module constraints
		 * @param _domains initial vector of domains
		 * @param _relations initial vector of relations
		 * @param _rules initial vector of rules
		 * @param _facts initial vector of facts
		 */
		ModuleLocal(std::string a_name,
			Constraints* _constraints,
			std::vector<Domain*> *_domains,
			std::vector<Relation*> *_relations,
			std::vector<Rule*> *_rules,
			std::vector<Fact*> *_facts);
		/** @brief destructor
		 *
		 */
		~ModuleLocal();
		logic_type::type ask_for_ground_fact(Fact *instance) const;
		Relation *get_relation(std::string s) const;
		virtual std::vector<Domain*> *get_domains() const override;
		std::vector<Relation*> *get_relations() const;
		std::vector<Rule*> *get_rules() const;
		std::vector<Fact*> *get_facts() const;
		std::string get_name() const;
		std::string print();
		module_type::type get_module_type() const;
		void generate_facts();
		void reason();
		void reason_with_rules(std::vector<Rule*>* rules);
		Result ask_for_result(Fact * term) const;
		/** @brief method gives type of a domain inside module
		 *
		 * @param name name of a domain
		 * @return domain type
		 */
		variable_type::type get_domain_type(std::string name);
		/** @brief method adds Domain into Domain vector
		 *
		 * @param d pointer to a Domain instance
		 */
		void add_domain(Domain *d);
		/** @brief method adds Relation into Relation vector
		 *
		 * @param r pointer to a Relation instance
		 */
		void add_relation(Relation *r);
		/** @brief method adds Fact into Fact vector
		 *
		 * @param r pointer to a Fact instance
		 */
		void add_fact(Fact *r);
		/** @brief method adds Rule into Rule vector
		 *
		 * @param r pointer to a Rule instance
		 */
		void add_rule(Rule *r);
		/** @brief getter for constraints
		 *
		 * @return list of constraints
		 */
		Constraints* get_constraints() const override { return constraints; };
		/** @brief fetches a database of facts from well-supported model (for summing purposes)
		 *
		 * @return database of wsm facts
		 */
		virtual database* get_wsm_database() { return &database_; }
		/** @brief allows for disabling constraints check
		 *
		 * @param value value to be set
		 */
		bool check_constraints_local(bool check_rigid, bool check_flexible) const override;
		/** @brief allows for disabling constraints check
		 *
		 * @param value value to be set
		 */
		void set_disable_constraints(bool value) const override { disable_constraints = value; };
		/** @brief checks whether constraints are violated
		 *
		 * @return true if constraints are vioilated, false otherwise
		 */
		bool constraints_violated() override;
	private:
		/* returns true when actually adds something */
		bool reason_about_rule(Rule * rule, logic_type::type target_logic_value);
		void generate_unknown_facts(Relation * relation);
		void initialize_database();
		int database_fingerprint(Inter4ql::database db);

		bool add_to_variable_space(std::string module, std::string relation, std::vector<Value*> val);
		void generate_domain_facts(bool should_initialize = true);

		void print_database();

		std::string name;
		Constraints* constraints;
		std::vector<Domain *> *domains;
		std::vector<Relation *> *relations;
		std::vector<Fact *> *facts;
		std::vector<Rule*> *rules;

		mutable bool disable_constraints;

		std::vector<Domain *> *domainsToDelete;

	protected:
		Inter4ql::database database_;
	};
}

#endif  /* __MODULELOCAL_H__ */

