#ifndef __MODULE_PROBLEM_H__
#define __MODULE_PROBLEM_H__

#include <vector>
#include <string>
#include <stack>

#include "ModuleAction.h"
#include "ModuleGeneric.h"

namespace Inter4ql {

	class Expression;
	class Value;

	enum TagType {
		START_PARALLEL, RESET_ENV_PARALLEL, END_PARALLEL, EXPRESSION
	};

	class ActionEntry {
	public:
		ActionEntry(ModuleGeneric* belief, ModuleGeneric* world,
				ModuleAction* action, std::vector<Value *>* valuation,
				std::vector<std::string>* vn, int rating) {
			this->belief_ = belief;
			this->world_ = world;
			this->action_ = action;
			this->valuation_ = valuation;
			this->variable_names_ = vn;

			this->rating_ = rating;
		}

		ModuleGeneric* get_belief() const {
			return this->belief_;
		}

		ModuleGeneric* get_world() const {
			return this->world_;
		}

		ModuleAction* get_action() const {
			return this->action_;
		}

		std::vector<Value *>* get_valuation() const {
			return this->valuation_;
		}

		std::vector<std::string>* get_variable_names() const {
			return this->variable_names_;
		}

		int get_rating() const {
			return this->rating_;
		}

	private:
		ModuleGeneric* belief_;
		ModuleGeneric* world_;
		ModuleAction* action_;
		std::vector<Value *>* valuation_;
		std::vector<std::string>* variable_names_;
		int rating_;
	};

	struct action_entry_compare {
	    bool operator() (const ActionEntry* lhs, const ActionEntry* rhs) const {
	        return lhs->get_rating() > rhs->get_rating();
	    }
	};

	class StackElem {
	public:
		StackElem(TagType type, std::string action_name, int level) {
			this->elem_type = type;
			this->action_name = action_name;
			this->expr = nullptr;
			this->parallel_end = nullptr;

			this->stored_to_add_partial = nullptr;
			this->stored_to_remove_partial = nullptr;

			this->stored_to_add = nullptr;
			this->stored_to_remove = nullptr;

			this->required_binding = nullptr;

			this->level = level;
			this->is_root_in_parallel = false;
		}

		StackElem(TagType type, Expression* expr, std::map<std::string, Value*>* required_binding,
				std::string action_name, int level) {
			this->elem_type = type;
			this->action_name = action_name;
			this->expr = expr;
			this->parallel_end = nullptr;

			this->stored_to_add_partial = nullptr;
			this->stored_to_remove_partial = nullptr;

			this->stored_to_add = nullptr;
			this->stored_to_remove = nullptr;

			this->required_binding = required_binding;

			this->level = level;
			this->is_root_in_parallel = false;
		}

		StackElem(TagType type, StackElem* parallel_end, std::string action_name, int level) {
			this->elem_type = type;
			this->action_name = action_name;
			this->expr = nullptr;
			this->parallel_end = parallel_end;

			this->stored_to_add_partial = nullptr;
			this->stored_to_remove_partial = nullptr;

			this->stored_to_add = nullptr;
			this->stored_to_remove = nullptr;

			this->required_binding = nullptr;

			this->level = level;
			this->is_root_in_parallel = false;
		}

		std::string print() {
			if (elem_type == START_PARALLEL)
				return "START_PARALLEL";
			if (elem_type == RESET_ENV_PARALLEL)
				return "RESET_ENV_PARALLEL";
			if (elem_type == END_PARALLEL)
				return "END_PARALLEL";
			if (elem_type == EXPRESSION)
				return print_expression(expr);
		}

		Expression* expr;
		TagType elem_type;
		StackElem* parallel_end;

		database* stored_to_add_partial;
		database* stored_to_remove_partial;

		database* stored_to_add;
		database* stored_to_remove;

		int level;
		bool is_root_in_parallel;

		std::string action_name;

		std::map<std::string, Value*>* required_binding;
	};

	/** @brief Implementation of ModuleAction for modules with reasoning
	 */
	class ModuleProblem : public ModuleGeneric {
	public:
		/** @brief constructor
		 *
		 * @param _name name of a module
		 * @param _constraints initial list of constraints
		 * @param _members list of members
		 */
		ModuleProblem(std::string _name,
			std::vector<std::string>* _beliefs,
			std::vector<std::string>* _actions,
			Expression* _goal,
			int _max_depth,
			std::vector<std::string>* _heuristics);
		/** @brief destructor
		 *
		 */
		~ModuleProblem();

		std::string get_name() const;
		std::string print();
		module_type::type get_module_type() const;
		void reason();
		Result ask_for_result(Fact * term) const;

		/** @brief getter for a beliefs in this Problem
		 *
		 * @returns belief list (string list)
		 */
		std::vector<std::string>* get_beliefs();
		/** @brief getter for a actions in this Problem
		 *
		 * @returns action list (string list)
		 */
		std::vector<std::string>* get_actions();
		/** @brief getter for a goal in this Problem
		 *
		 * @returns goal Expression
		 */
		Expression* get_goal();
		/** @brief getter for a actions in this Problem
		 *
		 * @returns action list
		 */
		int get_max_depth();
		/** @brief adds Belief's Module object to the Problem
		 *
		 * @param module object to be added
		 */
		void add_belief_module(ModuleGeneric* module);
		/** @brief getter for a beliefs in this Problem
		 *
		 * @returns belief list (Module object list)
		 */
		std::vector<ModuleGeneric*>* get_belief_modules();
		/** @brief adds Action's Module object to the Problem
		 *
		 * @param module object to be added
		 */
		void add_action_module(ModuleGeneric* module);
		/** @brief getter for a actions in this Problem
		 *
		 * @returns action list (Module object list)
		 */
		std::vector<ModuleGeneric*>* get_action_modules();
		/** @brief setter for whole action list in the Problem
		 *
		 * @param actions a new list of actions (list of strings)
		 */
		void replace_actions(std::vector<std::string>* actions);
		/** @brief checks whether give actions is used in the Problem
		 *
		 * @returns true if action is used, false otherwise
		 */
		bool contains_action(std::string action_name);

	private:
		ModuleGeneric* select_action(std::vector<ModuleGeneric*>* actions, std::vector<ModuleGeneric*>* composite);
		std::vector<Value *>* select_valuation(std::vector<std::vector<Value *>*>* valuations);
		ModuleGeneric* select_generic_module(std::vector<ModuleGeneric*>* modules);

		bool find_and_apply_action(int step, std::vector<std::string>& plan, int& rating);

		bool traverse_composite_expr(std::stack<StackElem>* composite_stack, std::stack<database*>* add_database_stack,
				std::stack<database*>* remove_database_stack, ModuleAction* current_action, ModuleGeneric* member, ModuleGeneric* belief,
				std::vector<ModuleGeneric*>* action_modules, int step, int steps_from_start, std::vector<std::string>& plan, int& rating,
				int best_rating, std::string previous_id);

		std::vector<std::string>* beliefs;
		std::vector<ModuleGeneric*>* belief_modules;
		std::vector<std::string>* actions;
		std::vector<ModuleGeneric*>* action_modules;
		Expression* goal;
		int max_depth;
		std::string name;
		std::vector<std::string>* heuristics;

		bool goal_probing;
		bool increase_rating_only;
		bool strict_increase_rating_only;
		bool composite_first;
		bool random;
		bool disallow_failed_preconditions;

		std::stack<StackElem>* composite_stack;
		std::stack<database*>* remove_database_stack;
		std::stack<database*>* add_database_stack;

		std::stack<int>* selected_rating;

		std::map<int, std::map<std::string, int>> composite_ratings;
	};
}

#endif  /* __MODULE_PROBLEM_H__ */

