#include <sstream>
#include <string>
#include "Value.h"
#include "Functions.h"
#include "Types.h"
#include "Exception.h"
#include "Logic.h"
#include "Disposable.h"

namespace Inter4ql {

	Value::Value(Value* v)
	: value_string(v->value_string)
	, value_double(v->value_double)
	, value_int(v->value_int)
	, type(v->type)
	{ }

	Value::Value(std::string s) {
		this->set_value(s);
	}

	Value::Value(int i, Inter4ql::variable_type::type _type) {
		this->set_value(i);
		this->set_type(_type);
	}

	Value::Value(double d) {
		this->set_value(d);
	}

	Value::~Value() {
		if(!Disposable::errorMode) {
				// do nothing!
		} else {
				//std::cout << "DELETE VALUE*" << std::endl;
		}
	}

	void Value::set_type(Inter4ql::variable_type::type _type) {
		if (_type != Inter4ql::variable_type::INVALID)
			this->type = _type;
	}

	bool Value::is_numeric(){
		return (this->type == Inter4ql::variable_type::INTEGER || this->type == Inter4ql::variable_type::REAL);
	}

	double Value::get_value_numeric(){
		if(!this->is_numeric())
			throw new Exception("Trying to get numeric value from nonnumeric variable(" + variable_type2string(this->type)+ ")\n");
		if(this->type == Inter4ql::variable_type::INTEGER)
			return (double) this->value_int;
		if(this->type == Inter4ql::variable_type::REAL)
			return this->value_double;
	}

	std::string Value::get_value_string() const {
		if (this->type != Inter4ql::variable_type::STRING &&
				this->type != Inter4ql::variable_type::VARIABLE &&
				this->type != Inter4ql::variable_type::ID &&
				this->type != Inter4ql::variable_type::DATE &&
				this->type != Inter4ql::variable_type::DATE_TIME)
			throw (new Exception("trying to get character value from noncharacter variable(" + variable_type2string(this->type) + ")\n"));
		return this->value_string;
	}

	int Value::get_value_int() const {
		if (this->type != Inter4ql::variable_type::INTEGER &&
				this->type != Inter4ql::variable_type::LOGIC)
			throw (new Exception("trying to get integer value from noninteger variable(" + variable_type2string(this->type)+ ")\n"));
		return this->value_int;
	}

	double Value::get_value_double() const {
		if (this->type != Inter4ql::variable_type::REAL)
			throw (new Exception("trying to get real value from nonreal variable(" + variable_type2string(this->type)+ ")\n"));
		return this->value_double;
	}

	void Value::set_value(std::string s) {

		if ( s == std::string("incons") || s == std::string("inconsistent")){
			return this->set_value(logic_type::INCONSISTENT, variable_type::LOGIC);
		}
		else if ( s == std::string("true")){
			return this->set_value(logic_type::TRUE, variable_type::LOGIC);
		}
		else if ( s == std::string("false")){
			return this->set_value(logic_type::FALSE, variable_type::LOGIC);
		}
		else if ( s == std::string("unknown")){
			return this->set_value(logic_type::UNKNOWN, variable_type::LOGIC);
		}
		this->value_string = s;
		if ((s[0] >= 'a' && s[0] <= 'z') || s[0] == '_' || s[0] == '<')
			this->type = Inter4ql::variable_type::ID;
		else if (s[0] >= 'A' && s[0] <= 'Z')
			this->type = Inter4ql::variable_type::VARIABLE;
		else if (s[0] == '"')
			this->type = Inter4ql::variable_type::STRING;
		else if (s[0] >= '0' && s[0] <= '9' && s.length() == 10)
			this->type = Inter4ql::variable_type::DATE;
		else if (s[0] >= '0' && s[0] <= '9' && s.length() == 16)
			this->type = Inter4ql::variable_type::DATE_TIME;
		else
			this->type = Inter4ql::variable_type::INVALID;
		return;
	}

	void Value::set_value(int i, Inter4ql::variable_type::type _type) {
		this->value_int = i;
		this->set_type(_type);
	}

	void Value::set_value(double d) {
		this->value_double = d;
		this->type = Inter4ql::variable_type::REAL;
	}

	Inter4ql::variable_type::type Value::get_type() const {
		return this->type;
	}

	bool Value::operator==(std::string &s) {
		return this->value_string == s;
	}

	bool Value::operator==(int &i) {
		return this->value_int == i;
	}

	bool Value::operator==(double &d) {
		return this->value_double == d;
	}

	bool Value::operator==(Value &_v) {
		if (this->get_type() != _v.get_type())
			return false;
		if (this->get_type() == Inter4ql::variable_type::INTEGER ||
			this->get_type() == Inter4ql::variable_type::LOGIC)
			return this->get_value_int() == _v.get_value_int();
		if (this->get_type() == Inter4ql::variable_type::REAL)
			return this->get_value_double() == _v.get_value_double();
		else
			return this->get_value_string() == _v.get_value_string();
	}

	bool Value::operator !=(Value& _v) {
		if (this->get_type() != _v.get_type())
			return true;
		if (this->get_type() == Inter4ql::variable_type::INTEGER ||
			this->get_type() == Inter4ql::variable_type::LOGIC)
			return this->get_value_int() != _v.get_value_int();
		if (this->get_type() == Inter4ql::variable_type::REAL)
			return this->get_value_double() != _v.get_value_double();
		else
			return this->get_value_string() != _v.get_value_string();
	}

	bool operator<(const Value &a, const Value &b) {
		if (a.get_type() == b.get_type()) {
			if (a.get_type() == Inter4ql::variable_type::INTEGER ||
				a.get_type() == Inter4ql::variable_type::LOGIC)
				return a.get_value_int() < b.get_value_int();
			else if (a.get_type() == Inter4ql::variable_type::REAL)
				return a.get_value_double() < b.get_value_double();
			else
				return a.get_value_string() < b.get_value_string();
		} else
			return a.get_type() < b.get_type();
	}

	bool operator==(const Value &a, const Value &b) {
		if (a.get_type() == b.get_type()) {
			if (a.get_type() == Inter4ql::variable_type::INTEGER ||
				a.get_type() == Inter4ql::variable_type::LOGIC)
				return a.get_value_int() == b.get_value_int();
			else if (a.get_type() == Inter4ql::variable_type::REAL)
				return a.get_value_double() == b.get_value_double();
			else
				return a.get_value_string() == b.get_value_string();
		} else
			return false;
	}

	std::string Value::print(bool type) {
		std::string s("");
		switch (this->type) {
			case Inter4ql::variable_type::INTEGER:
			{
				std::stringstream ss;
				ss << this->value_int;
				s += ss.str();
				if (type)
					s += " (integer)";
				break;
			}
			case Inter4ql::variable_type::REAL:
			{
				std::stringstream ss;
				ss << this->value_double;
				s += ss.str();
				if (type)
					s += " (real)";
				break;
			}
			case Inter4ql::variable_type::LOGIC:
				s += Inter4ql::Logic::get_logic_name((Inter4ql::logic_type::type) this->value_int);
				if (type)
					s += " (logic)";
				break;
			case Inter4ql::variable_type::ID:
				s += this->value_string;
				if (type)
					s += " (literal)";
				break;
			case Inter4ql::variable_type::STRING:
				s += this->value_string;
				if (type)
					s += " (string)";
				break;
			case Inter4ql::variable_type::DATE:
				s += this->value_string;
				if (type)
					s += " (date)";
				break;
			case Inter4ql::variable_type::DATE_TIME:
				s += this->value_string;
				if (type)
					s += " (date_time)";
				break;
			case Inter4ql::variable_type::VARIABLE:
				s += this->value_string;
				if (type)
					s += " (variable)";
				break;
		}
		return s;
	}

}
