#ifndef VARIABLESPACE_H
#define VARIABLESPACE_H

#include <vector>
#include <map>
#include <string>

namespace Inter4ql {

	class Value;
	/** @brief VariableSpace class for handling all values in interpreter environment
	 */
	class VariableSpace {
	public:
		/** @brief constructor
		 */
		VariableSpace();
		/** @brief destructor
		 */
		~VariableSpace();
		/** @brief clears the variable space
		 */
		void clear();
		/** @brief returns string with whole space of variables information
		 *
		 * @return variable space information packed in a std::string
		 */
		std::string print();
		/** @brief getter for a pointer to a value from variable space (not to create many instances of the same value)
		 *
		 * @param _v pointer to Value class
		 * @param _module name of a module
		 * @param _domain name of a domain in the module
		 * @returns pointer to a Value
		 */
		Value *get_pointer(Value* _v, std::string _module, std::string _domain, bool& _changed);
		/** @brief getter of a variable type of a distinct type
		 *
		 * @param _module name of a module
		 * @param _domain name of a domain in the module
		 * @param _changed did variable space change by the call
		 * @returns variable space of type
		 */
		std::map<int, Value *> *get_integer_space(std::string _module, std::string _domain);
		/** @brief getter of a variable type of a distinct type
		 *
		 * @param _module name of a module
		 * @param _domain name of a domain in the module
		 * @returns variable space of type real
		 */
		std::map<double, Value *> *get_real_space(std::string _module, std::string _domain);
		/** @brief getter of a variable type of a distinct type
		 *
		 * @param _module name of a module
		 * @param _domain name of a domain in the module
		 * @returns variable space of type string
		 */
		std::map<std::string, Value *> *get_string_space(std::string _module, std::string _domain);
		/** @brief getter of a variable type of a distinct type
		 *
		 * @param _module name of a module
		 * @param _domain name of a domain in the module
		 * @returns variable space of type literal
		 */
		std::map<std::string, Value *> *get_const_space(std::string _module, std::string _domain);
		/** @brief getter of a variable type of a distinct type
		 *
		 * @param _module name of a module
		 * @param _domain name of a domain in the module
		 * @returns variable space of type date
		 */
		std::map<std::string, Value *> *get_date_space(std::string _module, std::string _domain);
		/** @brief getter of a variable type of a distinct type
		 *
		 * @param _module name of a module
		 * @param _domain name of a domain in the module
		 * @returns variable space of type date with time
		 */
		std::map<std::string, Value *> *get_datetime_space(std::string _module, std::string _domain);
		/** @brief getter of a variable type of a distinct type
		 *
		 * @param _module name of a module
		 * @param _domain name of a domain in the module
		 * @returns variable space of type logic
		 */
		std::map<int, Value *> *get_logic_space(std::string _module, std::string _domain);
	private:
		void initialize(std::string _module);

		Value *get_integer(Value *_v, std::string _module, std::string _domain, bool& _changed);
		Value *get_real(Value *v, std::string _module, std::string _domain, bool& _changed);
		Value *get_string(Value *v, std::string _module, std::string _domain, bool& _changed);
		Value *get_const(Value *v, std::string _module, std::string _domain, bool& _changed);
		Value *get_date(Value *v, std::string _module, std::string _domain, bool& _changed);
		Value *get_datetime(Value *v, std::string _module, std::string _domain, bool& _changed);
		Value *get_logic(Value *v, std::string _module, std::string _domain, bool& _changed);

		std::map<std::string, std::map<int, Value *>*> *integer_map;
		std::map<std::string, std::map<double, Value *>*> *real_map;
		std::map<std::string, std::map<std::string, Value *>*> *string_map;
		std::map<std::string, std::map<std::string, Value *>*> *const_map;
		std::map<std::string, std::map<std::string, Value *>*> *date_map;
		std::map<std::string, std::map<std::string, Value *>*> *datetime_map;
		std::map<std::string, std::map<int, Value *>*> *logic_map;


	};
}

#endif  /* VARIABLESPACE_H */

