// DISCLAIMER: this module is still under development, the implementation is
// still in a Proof-of-concept state. It will be cleaned after all tests are done
// (including proper memory management - it is non-trivial due to dependencies in
// other modules of 4ql as it was not meant to be that dynamic).

#include "ModuleProblem.h"

#include "Application.h"
#include "Exception.h"
#include "ModuleAction.h"
#include "ModuleBelief.h"
#include "Result.h"

#include <cassert>
#include <chrono>
#include <sstream>
#include <stdlib.h>
#include <time.h>

namespace Inter4ql {

	ModuleProblem::ModuleProblem(std::string _name,
			std::vector<std::string>* _beliefs,
			std::vector<std::string>* _actions,
			Expression* _goal,
			int _max_depth,
			std::vector<std::string>* _heuristics)
	: name(_name),
	  beliefs(_beliefs),
	  actions(_actions),
	  goal(_goal),
	  max_depth(_max_depth),
	  composite_first(false),
	  random(false),
	  heuristics(_heuristics) {
		belief_modules = new std::vector<ModuleGeneric*>();
		action_modules = new std::vector<ModuleGeneric*>();

		composite_stack = new std::stack<StackElem>();
		add_database_stack = new std::stack<database*>();
		remove_database_stack = new std::stack<database*>();

		selected_rating = new std::stack<int>();

		for(auto str : *_heuristics) {
			if(str == "random")
				random = true;

			if(str == "composite_first")
				composite_first = true;

			if(str == "goal_probing")
				goal_probing = true;

			if(str == "increase_rating_only")
				increase_rating_only = true;

			if(str == "strict_increase_rating_only")
				strict_increase_rating_only = true;

			if(str == "disallow_failed_preconditions")
				disallow_failed_preconditions = true;

			if (increase_rating_only || strict_increase_rating_only)
				goal_probing = true;
		}
	}

	ModuleProblem::~ModuleProblem() {
		delete belief_modules;
		delete action_modules;

		delete composite_stack;
		delete add_database_stack;
		delete remove_database_stack;

		delete selected_rating;
	}

	std::string ModuleProblem::get_name() const {
		return this->name;
	}

	std::string ModuleProblem::print() {
		std::string s("");
		s += "problem " + this->name + ":\n";	
		s += "beliefs:\n";
		for(auto belief : *beliefs) {
			s += "\t" + belief + "\n";
		}
		s += "actions:\n";
		for(auto action : *actions) {
			s += "\t" + action + "\n";
		}
		s += "goal:\n";
		s += "\t" + print_expression(goal);
		s += "\nmax_depth:\n";
		std::stringstream ss;
		ss << max_depth;
		s += "\t" + ss.str();
		s += "\n";
		s += "heuristics:\n";
		for(auto heuriscic : *heuristics) {
			s += "\t" + heuriscic + "\n";
		}
		return s;
	}

	module_type::type ModuleProblem::get_module_type() const {
		return module_type::PROBLEM;
	}

	void process_non_composite(ModuleGeneric* member, database* to_add_db, database* to_remove_db, ModuleAction* am,
			std::map<std::string, Value*> binding, bool debug_planning) {
		std::vector<Rule*>* to_add = am->get_add_rules();
		std::vector<Rule*>* to_remove = am->get_remove_rules();

		// TODO: memory leak
		std::vector<Rule*>* valuated_to_add = new std::vector<Rule*>();
		for(Rule* r : *to_add) {
			Rule* r_to_valuate = new Rule(r);
			r_to_valuate->valuate_rule(binding);
			if (debug_planning) printf("(planning) Adding valuated to-add rule %s\n", r_to_valuate->print().c_str());
			valuated_to_add->push_back(r_to_valuate);
		}

		// TODO: memory leak
		std::vector<Rule*>* valuated_to_remove = new std::vector<Rule*>();
		for(Rule* r : *to_remove) {
			Rule* r_to_valuate = new Rule(r);
			r_to_valuate->valuate_rule(binding);
			if (debug_planning) printf("(planning) Adding valuated to-remove rule %s\n", r_to_valuate->print().c_str());
			valuated_to_remove->push_back(r_to_valuate);
		}

		database* to_add_snapshot_final = make_snapshot(member->get_add_database());
		database* to_remove_snapshot_final = make_snapshot(member->get_remove_database());

		// "reason" may change this...
		database* to_remove_snapshot = make_snapshot(member->get_remove_database());

		member->reason_with_rules(valuated_to_add);
		for(Rule* r : *valuated_to_add) {
			Result res = member->ask_for_result(r->get_head());

			if (debug_planning) printf("To add result: %s (rule: %s)\n", res.print(true, true).c_str(), r->print().c_str());

			Valuation val_temp = res.get_valuation_destructive();
			for (auto& valuation : val_temp) {
				if (valuation.second != Inter4ql::logic_type::TRUE && valuation.second != Inter4ql::logic_type::INCONSISTENT)
					continue;

				binding.clear();
				int corresponding_index = 0;

				std::vector<std::string> var_names = res.get_variable_names();
				std::vector<Value *> arguments = valuation.first;

				for (auto name : var_names) {
					binding.insert({name, arguments.at(corresponding_index++)});
				}

				// TODO: memory leak
				Fact* to_process = new Fact(r->get_head());
				to_process->valuate_fact(binding);

				if (to_add_db && to_remove_db) {
					if (debug_planning) printf("(planning) Adding to add fact %s\n", to_process->print().c_str());
					add_to_database(to_add_db, to_process, Inter4ql::logic_type::TRUE);
					remove_from_database(to_remove_db, to_process);
				} else {
					if (debug_planning) printf("(planning) Adding to (main) add fact %s\n", to_process->print().c_str());
					add_to_database(to_add_snapshot_final, to_process, Inter4ql::logic_type::TRUE);
					remove_from_database(to_remove_snapshot_final, to_process);
				}
			}
		}

		// TODO: memory leak!!!
		member->set_remove_database(to_remove_snapshot);

		member->reason_with_rules(valuated_to_remove);
		for(Rule* r : *valuated_to_remove) {
			Result res = member->ask_for_result(r->get_head());

			if (debug_planning) printf("To remove result: %s (rule: %s)\n", res.print(true, true).c_str(), r->print().c_str());

			Valuation val_temp = res.get_valuation_destructive();
			for (auto& valuation : val_temp) {
				if (valuation.second != Inter4ql::logic_type::TRUE && valuation.second != Inter4ql::logic_type::INCONSISTENT)
					continue;

				binding.clear();
				int corresponding_index = 0;

				std::vector<std::string> var_names = res.get_variable_names();
				std::vector<Value *> arguments = valuation.first;

				for (auto name : var_names) {
					binding.insert({name, arguments.at(corresponding_index++)});
				}

				// TODO: memory leak
				Fact* to_process = new Fact(r->get_head());
				to_process->valuate_fact(binding);

				if (to_add_db && to_remove_db) {
					if (debug_planning) printf("(planning) Adding to remove fact %s\n", to_process->print().c_str());
					add_to_database(to_remove_db, to_process, Inter4ql::logic_type::TRUE);
					remove_from_database(to_add_db, to_process);
				} else {
					if (debug_planning) printf("(planning) Adding to (main) remove fact %s\n", to_process->print().c_str());
					add_to_database(to_remove_snapshot_final, to_process, Inter4ql::logic_type::TRUE);
					remove_from_database(to_add_snapshot_final, to_process);
				}
			}
		}

		// TODO: memory leak!!!
		member->set_add_database(to_add_snapshot_final);
		member->set_remove_database(to_remove_snapshot_final);

		member->reason();
	}

	void process_non_composite(ModuleGeneric* member, ModuleAction* am, std::map<std::string, Value*> binding, bool debug_planning) {
		process_non_composite(member, nullptr, nullptr, am, binding, debug_planning);
	}

	void extract_composite_to_stack(std::stack<StackElem>* composite_stack, Expression* composite_expr,
			std::stack<database*>* database_add_stack, std::stack<database*>* database_remove_stack,
			std::map<std::string, Value*>* binding, ModuleGeneric* member, ModuleAction* current_action,
			int level, std::string action_name, bool debug_planning) {
		std::vector<ModuleGeneric*>* global_modules = Application::app_instance->public_get_modules();

		switch(composite_expr->get_type()) {
		case expression_type::PARALLEL: {
			if (debug_planning) printf("[composite] Composite parallel!\n");

			database* to_add_snapshot = make_snapshot(member->get_add_database());
			database* to_remove_snapshot = make_snapshot(member->get_remove_database());
			database* to_add_snapshot_midway = make_snapshot(member->get_add_database());
			database* to_remove_snapshot_midway = make_snapshot(member->get_remove_database());
			database* to_add_snapshot_final = make_snapshot(member->get_add_database());
			database* to_remove_snapshot_final = make_snapshot(member->get_remove_database());

			composite_stack->push(StackElem(END_PARALLEL, action_name, level));
			composite_stack->top().stored_to_add = to_add_snapshot_final;
			composite_stack->top().stored_to_remove = to_remove_snapshot_final;

			if (debug_planning) printf("[composite] \"%s\" put on stack!\n", composite_stack->top().print().c_str());
			StackElem* end_tag = &composite_stack->top();
			composite_stack->push(StackElem(EXPRESSION, composite_expr->get_right_expression(), binding, action_name, level));
			composite_stack->top().is_root_in_parallel = true;
			if (debug_planning) printf("[composite] \"%s\" put on stack!\n", composite_stack->top().print().c_str());
			composite_stack->push(StackElem(RESET_ENV_PARALLEL, end_tag, action_name, level));
			composite_stack->top().stored_to_add = to_add_snapshot_midway;
			composite_stack->top().stored_to_remove = to_remove_snapshot_midway;
			if (debug_planning) printf("[composite] \"%s\" put on stack!\n", composite_stack->top().print().c_str());


			composite_stack->push(StackElem(EXPRESSION, composite_expr->get_left_expression(), binding, action_name, level));
			composite_stack->top().is_root_in_parallel = true;
			if (debug_planning) printf("[composite] \"%s\" put on stack!\n", composite_stack->top().print().c_str());
			composite_stack->push(StackElem(START_PARALLEL, action_name, level));
			composite_stack->top().stored_to_add = to_add_snapshot;
			composite_stack->top().stored_to_remove = to_remove_snapshot;
			if (debug_planning) printf("[composite] \"%s\" put on stack!\n", composite_stack->top().print().c_str());
			break;
		}
		case expression_type::SEQUENTIAL:
			if (debug_planning) printf("[composite] Composite sequential!\n");

			composite_stack->push(StackElem(EXPRESSION, composite_expr->get_right_expression(), binding, action_name, level));
			if (debug_planning) printf("[composite] \"%s\" put on stack!\n", composite_stack->top().print().c_str());
			composite_stack->push(StackElem(EXPRESSION, composite_expr->get_left_expression(), binding, action_name, level));
			if (debug_planning) printf("[composite] \"%s\" put on stack!\n", composite_stack->top().print().c_str());
			break;
		case expression_type::IMPLIES: {
			if (debug_planning) printf("[composite] Composite if!\n");
			int rating; // not used...
			Result if_result = evaluate_expression_global(composite_expr->get_left_expression(), member, global_modules, member->get_domains(), rating);

			if (!if_result.get_variable_names().empty()) {
				throw new Exception("Condition of composite if-action have to evaluate to single logic value. Problematic expression: " + print_expression(composite_expr->get_left_expression()));
			}

			const Valuation& v = if_result.get_valuation();

			if (v.get_first_logic_value() == Inter4ql::logic_type::TRUE) {
				composite_stack->push(StackElem(EXPRESSION, composite_expr->get_right_expression(), binding, action_name, level));
			} else {
				composite_stack->push(StackElem(EXPRESSION, composite_expr->get_else_expression(), binding, action_name, level));
			}
			if (debug_planning) printf("[composite] \"%s\" put on stack!\n", composite_stack->top().print().c_str());

			break;
		}
		case expression_type::TERM: {
			if (debug_planning) printf("[composite] Composite term!\n");

			// TODO: memory leak
			Expression* term_expr = new Expression(expression_type::TERM, nullptr, nullptr, composite_expr->get_term());
			term_expr->set_simple_relation_flag(true);
			composite_stack->push(StackElem(EXPRESSION, term_expr, binding, action_name, level));
			if (debug_planning) printf("[composite] \"%s\" put on stack!\n", composite_stack->top().print().c_str());

			break;
		}
		default:
			throw new Exception("Allowed relations between actions in composite expression are '||', ';' and '=>'. Unexpected operator found in composite action " + current_action->get_name());
		}
	}

	bool maybe_violated_constraints(ModuleGeneric* belief, int database_stack_size) {
		// if there is something in stack then there is no
		// point in checking constraints - literals are not put to
		// belief's world so nothing will change
		// (constraints will be checked upon END_PARALLEL)
		if (database_stack_size > 0)
			return false;

		return belief->constraints_violated();
	}

	void combine_parallel_parts(database* to_add_first, database*  to_remove_first, database* to_add_second, database* to_remove_second, bool debug_planning) {
		if (debug_planning) printf("[composite] Checking to_add_first.\n");
		for(const auto elem : *to_add_first) {
			if (debug_planning) printf("[composite] Checking single entry: %s\n", elem.first.c_str());
			if (debug_planning) elem.second->print();
			for(auto valuation : *elem.second) {
				if (debug_planning) printf("[composite] Checking elem with value %d\n", valuation.second);

				logic_type::type value_in_other = value_in_database(to_remove_second, elem.first, valuation.first);
				logic_type::type value_in_same = value_in_database(to_remove_first, elem.first, valuation.first);

				if (debug_planning) printf("[composite] Checking elem: value_in_other %d,  value_in_same %d\n", value_in_other, value_in_same);

				// if same fact found to be removed by second action then make it inconsistent
				if (value_in_other == logic_type::TRUE) {
					elem.second->add_element(valuation.first, logic_type::INCONSISTENT);
					continue;
				}

				if (value_in_same == logic_type::TRUE) {
					valuation.second = logic_type::UNKNOWN;
				}
			}
			if (debug_planning) printf("[composite] After changes entry: %s\n", elem.first.c_str());
			if (debug_planning) elem.second->print();
		}

		if (debug_planning) printf("[composite] Checking to_add_second.\n");
		for(const auto elem : *to_add_second) {
			if (debug_planning) printf("[composite] Checking single entry: %s\n", elem.first.c_str());
			if (debug_planning) elem.second->print();

			auto slot = to_add_first->find(elem.first);

			if (slot == to_add_first->end()) {
				to_add_first->insert(std::make_pair(elem.first, std::make_shared<Valuation>()));
				slot = to_add_first->find(elem.first);
			}

			for(auto valuation : *elem.second) {
				if (debug_planning) printf("[composite] Checking elem with value %d\n", valuation.second);

				logic_type::type value_in_other = value_in_database(to_remove_first, elem.first, valuation.first);
				logic_type::type value_in_same = value_in_database(to_remove_second, elem.first, valuation.first);

				if (debug_planning) printf("[composite] Checking elem: value_in_other %d,  value_in_same %d\n", value_in_other, value_in_same);

				if (value_in_other == logic_type::TRUE) {
					if (debug_planning) printf("[composite] Adding entry: %s with value INCONSISTENT\n", elem.first.c_str());
					slot->second->add_element(valuation.first, logic_type::INCONSISTENT);
					continue;
				}

				if (value_in_same != logic_type::TRUE) {
					if (debug_planning) printf("[composite] Adding entry: %s with value %d\n", elem.first.c_str(), valuation.second);
					slot->second->add_element(valuation.first,
							(valuation.second == logic_type::INCONSISTENT) ? logic_type::INCONSISTENT : logic_type::TRUE);
				}
			}
		}

		if (debug_planning) printf("[composite] Checking to_remove_second.\n");
		for(const auto elem : *to_remove_second) {
			if (debug_planning) printf("[composite] Checking single entry of to_remove_second: %s\n", elem.first.c_str());
			if (debug_planning) elem.second->print();
			auto slot = to_remove_first->find(elem.first);

			if (slot == to_remove_first->end()) {
				to_remove_first->insert(std::make_pair(elem.first, std::make_shared<Valuation>()));
				slot = to_remove_first->find(elem.first);
			}

			for(const auto valuation : *elem.second) {
				assert(valuation.second == logic_type::TRUE);
				slot->second->add_element(valuation.first, logic_type::TRUE);
			}
		}

		// in to_remove_first is a combination of remove databases
		// in to_add_first is a combination of add databases
	}

	std::string get_id(std::string action_name, std::string belief_name, std::string world_name,
			int step, int steps_from_start, const std::vector<Value*>* valuation) {
		std::stringstream ss;
		ss << step << "." << steps_from_start << "." << action_name << "." << belief_name << "." << world_name << ".";
		if (valuation) {
			for(auto value : *valuation)
				ss << value->print(true) << ".";
		} else {
			ss << "failed.";
		}
		return ss.str();

	}

	bool ModuleProblem::traverse_composite_expr(std::stack<StackElem>* composite_stack, std::stack<database*>* add_database_stack,
			std::stack<database*>* remove_database_stack, ModuleAction* current_action, ModuleGeneric* member, ModuleGeneric* belief,
			std::vector<ModuleGeneric*>* action_modules, int step, int steps_from_start, std::vector<std::string>& plan, int& rating,
			int best_rating, std::string previous_id) {
		bool debug_probing = Application::app_instance->debug_probing;
		bool debug_warning = Application::app_instance->debug_warning;
		bool debug_planning_composite = Application::app_instance->debug_planning_composite;

		std::vector<ModuleGeneric*>* global_modules = Application::app_instance->public_get_modules();
		std::vector<ModuleGeneric*>* global_actions = Application::app_instance->public_get_actions();

		database* to_add_snapshot_end_parallel = nullptr;

		if (debug_planning_composite | debug_probing) printf("[composite] ======================= traverse_composite_expr step %d (rating %d) =======================\n", step, rating);

		assert(!composite_stack->empty());

		if(!composite_stack->empty()) {
			StackElem elem = composite_stack->top();
			composite_stack->pop();

			int level = elem.level;
			int stack_size = composite_stack->size();

			std::string action_name = elem.action_name;

			if (debug_planning_composite) printf("[composite] (Step %d) Read \"%s\" from stack! Current size is: %d\n", step, elem.print().c_str(), stack_size);

			std::map<std::string, Value*>* required_binding = elem.required_binding;

			bool flow_element = false;

			// flow control elements
			if (elem.elem_type == START_PARALLEL) {
				if (debug_planning_composite) printf("[composite] (Step %d) Read START_PARALLEL!\n", step);

				add_database_stack->push(new database());
				remove_database_stack->push(new database());

				database* to_add_snapshot = make_snapshot(elem.stored_to_add);
				database* to_remove_snapshot = make_snapshot(elem.stored_to_remove);

				member->set_add_database(to_add_snapshot);
				member->set_remove_database(to_remove_snapshot);
				flow_element = true;
			}
			if (elem.elem_type == RESET_ENV_PARALLEL) {
				if (debug_planning_composite) printf("[composite] (Step %d) Read RESET_ENV_PARALLEL!\n", step);
				database* to_add = add_database_stack->top();
				add_database_stack->pop();

				database* to_remove = remove_database_stack->top();
				remove_database_stack->pop();

				elem.parallel_end->stored_to_add_partial = to_add;
				elem.parallel_end->stored_to_remove_partial = to_remove;

				add_database_stack->push(new database());
				remove_database_stack->push(new database());

				database* to_add_snapshot = make_snapshot(elem.stored_to_add);
				database* to_remove_snapshot = make_snapshot(elem.stored_to_remove);

				member->set_add_database(to_add_snapshot);
				member->set_remove_database(to_remove_snapshot);

				flow_element = true;
			}
			if (elem.elem_type == END_PARALLEL) {
				if (debug_planning_composite) printf("[composite] (Step %d) Read END_PARALLEL!\n", step);

				database* to_add_snapshot = make_snapshot(elem.stored_to_add);
				database* to_remove_snapshot = make_snapshot(elem.stored_to_remove);

				member->set_add_database(to_add_snapshot);
				member->set_remove_database(to_remove_snapshot);

				database* to_add_second = add_database_stack->top();
				add_database_stack->pop();
				database* to_remove_second = remove_database_stack->top();
				remove_database_stack->pop();

				database* to_add_first = make_snapshot(elem.stored_to_add_partial);
				database* to_remove_first = make_snapshot(elem.stored_to_remove_partial);

				// in to_remove_first is a combination of remove databases
				// in to_add_first is a combination of add databases
				combine_parallel_parts(to_add_first, to_remove_first, to_add_second, to_remove_second, debug_planning_composite);

				database* member_add_database = member->get_add_database();
				if(!add_database_stack->empty()) {
					member_add_database = add_database_stack->top();
				} else {
					member_add_database = member->get_add_database();
				}

				to_add_snapshot_end_parallel = make_snapshot(member_add_database);

				for(const auto elem : *to_add_first) {
					auto slot = member_add_database->find(elem.first);

					if (slot == member_add_database->end()) {
						member_add_database->insert(std::make_pair(elem.first, std::make_shared<Valuation>()));
						slot = member_add_database->find(elem.first);
					}

					if (debug_planning_composite) printf("[composite] (Step %d) Adding name %s to 'add_database' with valuation:\n", step, elem.first.c_str());
					if (debug_planning_composite) elem.second->print();

					for(const auto valuation : *elem.second) {
						slot->second->add_element(valuation.first, valuation.second);
					}
				}

				flow_element = true;
			}

			if (flow_element) {
				bool result = false;
				int selected_rating_size = -10;
				if (!maybe_violated_constraints(belief, add_database_stack->size())) {
					if (!composite_stack->empty()) {
						result = traverse_composite_expr(composite_stack, add_database_stack,
								remove_database_stack, current_action, member, belief,
								action_modules, step, steps_from_start + 1, plan, rating,
								best_rating, previous_id);
					} else {
						selected_rating->push(best_rating);
						selected_rating_size = selected_rating->size();
						result = find_and_apply_action(step + 1, plan, rating);
					}

					if (debug_probing) printf("[probing] (Step %d) Setting rating %d as output (flow elem - pass through)!\n", step, rating);
				} else {
					if (debug_probing) printf("[probing] (Step %d) Setting rating %d as output (flow elem - pass through)!\n", step, -50);
					rating = -50;
				}

				if (to_add_snapshot_end_parallel) {
					if (!add_database_stack->empty()) {
						add_database_stack->pop();
						add_database_stack->push(to_add_snapshot_end_parallel);
					} else {
						member->set_add_database(to_add_snapshot_end_parallel);
					}
				}

				if (selected_rating_size != -10) {
					assert(selected_rating->size() == selected_rating_size);
					selected_rating->pop();
				}

				if (result)
					return true;
			}

			if (elem.elem_type == EXPRESSION) {
				if (debug_planning_composite) printf("[composite] (Step %d) Read EXPRESSION \"%s\"!\n", step, print_expression(elem.expr).c_str());

				// actual expression (which may be a "simple" action or a composite one...)
				Expression* current = elem.expr;

				if (current->get_type() == expression_type::TERM) {
					if (debug_planning_composite) printf("[composite] (Step %d) Composite term!\n", step);

					if (!current->is_simple_relation()) {
						throw new Exception("[composite] Allowed terms in composite expression should be only action name with arguments. Unexpected term found in composite action " + current_action->get_name() + "\n");
					}
					// we have a single relation - which means action name + arguments

					int max_rating = -100;

					for (ModuleGeneric* action : *global_actions) {
						if (action->get_name() != current->get_term()->get_relation())
							continue;

						ModuleAction* am = dynamic_cast<ModuleAction*>(action);
						if (debug_planning_composite) printf("[composite] (Step %d) Checking action \"%s\":\n\n", step, am->get_name().c_str());

						int precondition_rating; // not used here...
						Result precondition_result;
						try {
							precondition_result = evaluate_expression_global(am->get_preconditions(), member, global_modules, member->get_domains(), precondition_rating);
						} catch (Inter4ql::Exception *s) {
							if (debug_warning) {
								printf("WARNING in step %d at preconditions check for action %s in world %s: %s", step,  am->get_name().c_str(), member->get_name().c_str(), s->get().c_str());
							}
							// do nothing, missing relation is fine here...
						}

						if (debug_planning_composite) printf("[composite] (Step %d) Plan execution - precondition result for action \"%s\":\n%s\n", step, am->get_name().c_str(), precondition_result.print().c_str());

						const auto& valuation = precondition_result.get_valuation();
						const std::vector<std::string>& vn =  precondition_result.get_variable_names();

						std::multiset<ActionEntry*, action_entry_compare> true_valuations;

						std::vector<std::vector<Value *>*>* valuations = new std::vector<std::vector<Value *>*>();

						for(const auto& val : valuation) {
							if (val.second != Inter4ql::logic_type::TRUE)
								continue;

							valuations->push_back(new std::vector<Value *>(val.first.begin(), val.first.end()));
						}

						while(std::vector<Value *>* val = select_valuation(valuations)) {
							int rating_local = 0;
							auto rating_it = composite_ratings.find(step);
							if (rating_it != composite_ratings.end()) {
								std::map<std::string, int> rating_map = rating_it->second;

								std::string id = get_id(am->get_name(), belief->get_name(), member->get_name(),
										step, steps_from_start, val);

								auto map_it = rating_map.find(previous_id + id);
								if (map_it != rating_map.end()) {
									rating_local = map_it->second;
								}
							}

							true_valuations.insert(new ActionEntry(belief, member, am, val, nullptr, rating_local));
						}

						if (!true_valuations.empty()) {
							std::vector<Value*>* action_arguments = am->get_args();
							for (auto name : *action_arguments) {
								if(std::find(vn.begin(), vn.end(), name->get_value_string()) == vn.end()) {
									std::string error_msg = "Precondition of action '" + am->get_name() + "' does not assign value to variable '"
											+ name->get_value_string() + "' which is required by action's argument\n";
									throw new Exception(error_msg);
								}
							}
						}

						// if we are in parallel execution and this action is not executable due to failed preconditions...
						if (!disallow_failed_preconditions && elem.is_root_in_parallel && true_valuations.empty()) {
							if (debug_planning_composite) {
								printf("[composite] (Step %d) Action \"%s\" is not executable due to failed preconditions! Letting it conditionally through! ", step, am->get_name().c_str());
							}

							bool result = false;
							int rating_local = rating;
							if (debug_probing) printf("[probing] (Step %d) Before recursive call; rating_local: %d\n", step, rating_local);
							std::string id = get_id(am->get_name(), belief->get_name(), member->get_name(),
									step, steps_from_start, nullptr);

							int rating_stored = 0;
							auto rating_it = composite_ratings.find(step);
							if (rating_it != composite_ratings.end()) {
								std::map<std::string, int> rating_map = rating_it->second;

								auto map_it = rating_map.find(previous_id + id);
								if (map_it != rating_map.end()) {
									rating_stored = map_it->second;
								}
							}

							int selected_rating_size = -10;
							if (!composite_stack->empty()) {
								result = traverse_composite_expr(composite_stack, add_database_stack,
										remove_database_stack, current_action, member, belief,
										action_modules, step, steps_from_start + 1, plan, rating_local,
										rating_stored, previous_id + id);
							} else {
								selected_rating->push(rating_stored);
								selected_rating_size = selected_rating->size();
								result = find_and_apply_action(step + 1, plan, rating_local);
							}

							// if rating was checked then store result for further use
							if (rating > 0) {
								int previous_value = rating_local;

								auto value = composite_ratings.find(step)->second.find(previous_id + id);
								if (value != composite_ratings.find(step)->second.end()) {
									previous_value = value->second;
								}

								if (previous_value < rating_local) {
									composite_ratings.find(step)->second.erase(previous_id + id);
								}

								if (debug_probing) printf("[probing] (Step %d) Adding key \"%s\" to database with rating: %d\n", step, (previous_id + id).c_str(), rating_local);
								composite_ratings.find(step)->second.insert(std::make_pair(previous_id + id, rating_local));
							}

							max_rating = std::max(max_rating, rating_local);

							if (selected_rating_size != -10) {
								assert(selected_rating->size() == selected_rating_size);
								selected_rating->pop();
							}

							if (debug_probing) printf("[probing] (Step %d) After recursive call; rating_local: %d, max_rating: %d\n", step, rating_local, max_rating);

							if (result) {
								if (debug_planning_composite) {
									printf("[composite] (Step %d) (planning) Planning success! %s() - failed preconditions!\n", step, am->get_name().c_str());
								}

								std::stringstream ss;
								ss << "  ";
								for(int i = 0; i < level; i++) {
									ss << "  ";
								}
								ss << am->get_name().c_str() << "() - failed preconditions";
								plan.push_back(ss.str());

								while (composite_stack->size() > stack_size) {
									if (debug_planning_composite) printf("[composite] (Step %d) Popping \"%s\" from stack! (stack size is %lu, going down to %d)\n",
											step, composite_stack->top().print().c_str(), composite_stack->size(), stack_size);
									StackElem elem = composite_stack->top();
									composite_stack->pop();

									if (elem.elem_type == END_PARALLEL) {
										if (!add_database_stack->empty())
											add_database_stack->pop();
										if (!remove_database_stack->empty())
											remove_database_stack->pop();
									}
								}

								return true;
							}
						} else {
							while (!true_valuations.empty()) {
								//while(std::vector<Value *>* val = select_valuation(true_valuations)) {

								ActionEntry* entry = *(true_valuations.begin());
								true_valuations.erase(true_valuations.begin());
								std::vector<Value *>* val = entry->get_valuation();

								// all actions in composite one have the same "best" rating - we need to allow for
								// equal ratings to be able to apply composite actions at all...
								if (rating < 0 && (increase_rating_only || strict_increase_rating_only)
										&& entry->get_rating() < selected_rating->top()) {
									continue;
								}

								std::map<std::string, Value*> bindingNew;

								for(int i = 0; i < vn.size(); i++) {
									bindingNew.insert({vn.at(i).c_str(), val->at(i)});
								}

								bool matches_required = true;
								for(auto iter = bindingNew.begin(); iter != bindingNew.end(); iter++) {
									auto found_elem = required_binding->find(iter->first);

									// required binding does not say anything about this literal - so lets skip it
									if (found_elem == required_binding->end()) {
										continue;
									}

									if (*found_elem->second != *iter->second) {
										matches_required = false;
										break;
									}
								}

								if (!matches_required) {
									if (debug_planning_composite) printf("Warning: TRUE binding skipped due to mismatch with binding from upper layer!\n");
									continue;
								}

								if (debug_planning_composite | debug_probing) {
									printf("[composite] (Step %d) Action \"%s\" is executable with rating %d and values: ", step, am->get_name().c_str(), entry->get_rating());
									for(int i = 0; i < vn.size(); i++) {
										printf("%s: %s, ", vn.at(i).c_str(), val->at(i)->print(true).c_str());
									}
									printf("\nmax_rating: %d\n", max_rating);
								}

								// We do not push current rating as it is may be the best possible
								// but the treshold is in fact lower
								//selected_rating->push(entry->get_rating());

								// We have TRUE valuation - let's apply it
								database* to_add_snapshot = make_snapshot(member->get_add_database());
								database* to_remove_snapshot = make_snapshot(member->get_remove_database());

								if (!am->is_composite()) {
									process_non_composite(member, add_database_stack->empty() ? nullptr : add_database_stack->top(),
											remove_database_stack->empty() ? nullptr : remove_database_stack->top(),
											am, bindingNew, debug_planning_composite);
								} else {
									if (debug_planning_composite) printf("[composite] (Step %d) Action is composite! \"%s\" will be put to stack!\n", step, print_expression(am->get_composite_expr()).c_str());
									// Optimization! We could just put the action to stack here...
									extract_composite_to_stack(composite_stack, am->get_composite_expr(), add_database_stack, remove_database_stack,
											&bindingNew, member, current_action, level + 1, am->get_name(), debug_planning_composite);
								}

								if (!maybe_violated_constraints(belief, add_database_stack->size())) {
									bool result = false;
									int selected_rating_size = -10;
									int rating_local = rating;
									if (debug_probing) printf("[probing] (Step %d) Before recursive call; rating_local: %d\n", step, rating_local);
									std::string id = get_id(am->get_name(), belief->get_name(), member->get_name(),
											step, steps_from_start, val);
									if (!composite_stack->empty()) {
										result = traverse_composite_expr(composite_stack, add_database_stack,
												remove_database_stack, current_action, member, belief,
												action_modules, step, steps_from_start + 1, plan, rating_local,
												entry->get_rating(), previous_id + id);
									} else {
										selected_rating->push(entry->get_rating());
										selected_rating_size = selected_rating->size();
										result = find_and_apply_action(step + 1, plan, rating_local);
									}

									// if rating was checked then store result for further use
									if (rating > 0) {
										int previous_value = rating_local;

										auto value = composite_ratings.find(step)->second.find(previous_id + id);
										if (value != composite_ratings.find(step)->second.end()) {
											previous_value = value->second;
										}

										if (previous_value < rating_local) {
											composite_ratings.find(step)->second.erase(previous_id + id);
										}

										if (debug_probing) printf("[probing] (Step %d) Adding key \"%s\" to database with rating: %d\n", step, (previous_id + id).c_str(), rating_local);
										composite_ratings.find(step)->second.insert(std::make_pair(previous_id + id, rating_local));
									}

									max_rating = std::max(max_rating, rating_local);

									if (selected_rating_size != -10) {
										assert(selected_rating->size() == selected_rating_size);
										selected_rating->pop();
									}

									if (debug_probing) printf("[probing] (Step %d) After recursive call; rating_local: %d, max_rating: %d\n", step, rating_local, max_rating);

									if (result) {
										std::vector<Value*>* action_arguments = am->get_args();

										if (debug_planning_composite) {
											printf("[composite] (Step %d) (planning) Planning success! %s(", step, am->get_name().c_str());
											for(int i = 0; i < vn.size(); i++) {
												bool is_argument = false;
												for (auto name : *action_arguments) {
													if (name->get_value_string() == vn.at(i).c_str())
														is_argument = true;
												}
												if (!is_argument)
													continue;
												printf("%s: %s, ", vn.at(i).c_str(), val->at(i)->print(true).c_str());
											}
											printf(")\n");
										}

										std::stringstream ss;
										ss << "  ";
										for(int i = 0; i < level; i++) {
											ss << "  ";
										}
										ss << am->get_name().c_str() << "(";
										for(int i = 0; i < vn.size(); i++) {
											bool is_argument = false;
											for (auto name : *action_arguments) {
												if (name->get_value_string() == vn.at(i).c_str())
													is_argument = true;
											}
											if (!is_argument)
												continue;

											ss << vn.at(i).c_str() << ": " << val->at(i)->print(true).c_str();
											if (i < vn.size() - 1)
												ss << ", ";
										}
										ss << ")";
										if (am->is_composite()) {
											ss << ":";
										}
										plan.push_back(ss.str());

										// TODO: memory leak!!!
										member->set_add_database(to_add_snapshot);
										member->set_remove_database(to_remove_snapshot);

										while (composite_stack->size() > stack_size) {
											if (debug_planning_composite) printf("[composite] (Step %d) Popping \"%s\" from stack! (stack size is %lu, going down to %d)\n",
													step, composite_stack->top().print().c_str(), composite_stack->size(), stack_size);
											StackElem elem = composite_stack->top();
											composite_stack->pop();

											if (elem.elem_type == END_PARALLEL) {
												if (!add_database_stack->empty())
													add_database_stack->pop();
												if (!remove_database_stack->empty())
													remove_database_stack->pop();
											}
										}

										delete val;

										return true;
									}
								} else {
									if (debug_planning_composite | debug_probing) {
										printf("[composite] (Step %d) Action \"%s\" violated constraints with values: ", step, am->get_name().c_str());
										for(int i = 0; i < vn.size(); i++) {
											printf("%s: %s, ", vn.at(i).c_str(), val->at(i)->print(true).c_str());
										}
										printf("\n");
									}
									// if rating was checked then store result for further use
									if (rating > 0) {
										std::string id = get_id(am->get_name(), belief->get_name(), member->get_name(),
												step, steps_from_start, val);

										int previous_value = -50;

										auto value = composite_ratings.find(step)->second.find(previous_id + id);
										if (value != composite_ratings.find(step)->second.end()) {
											previous_value = value->second;
										}

										if (previous_value < -50) {
											composite_ratings.find(step)->second.erase(previous_id + id);
										}

										if (debug_probing) printf("[probing] (Step %d) Adding key \"%s\" to database with rating: %d\n", step, (previous_id + id).c_str(), -50);
										composite_ratings.find(step)->second.insert(std::make_pair(previous_id + id, -50));
									}
									max_rating = std::max(max_rating, -50);
								}

								while (composite_stack->size() > stack_size) {
									if (debug_planning_composite) printf("[composite] (Step %d) Popping \"%s\" from stack! (stack size is %lu, going down to %d)\n",
											step, composite_stack->top().print().c_str(), composite_stack->size(), stack_size);
									StackElem elem = composite_stack->top();
									composite_stack->pop();

									if (elem.elem_type == END_PARALLEL) {
										if (!add_database_stack->empty())
											add_database_stack->pop();
										if (!remove_database_stack->empty())
											remove_database_stack->pop();
									}
								}

								// TODO: memory leak!!!
								member->set_add_database(to_add_snapshot);
								member->set_remove_database(to_remove_snapshot);

								delete val;
							}
						}

						if (debug_probing) printf("[probing] (Step %d) Setting rating %d as output (from max_rating)!\n", step, max_rating);
						rating = max_rating;
					}
				} else {
					if (debug_planning_composite) printf("[composite] (Step %dd) Expression read from stack is composite! Read expression is \"%s\" and will be put on stack!\n",
							step, print_expression(current).c_str());
					extract_composite_to_stack(composite_stack, current, add_database_stack, remove_database_stack,
							required_binding, member, current_action, level, action_name, debug_planning_composite);

					if (!maybe_violated_constraints(belief, add_database_stack->size())) {
						bool result = false;
						int selected_rating_size = -10;
						if (!composite_stack->empty()) {
							result = traverse_composite_expr(composite_stack, add_database_stack,
									remove_database_stack, current_action, member, belief,
									action_modules, step, steps_from_start + 1, plan, rating,
									best_rating, previous_id);
						} else {
							selected_rating->push(best_rating);
							selected_rating_size = selected_rating->size();
							result = find_and_apply_action(step + 1, plan, rating);
						}

						if (selected_rating_size != -10) {
							assert(selected_rating->size() == selected_rating_size);
							selected_rating->pop();
						}

						if (debug_probing) printf("[probing] (Step %d) Setting rating %d as output (cexpr - pass through)!\n", step, rating);

						if (result)
							return true;
					} else {
						if (debug_probing) printf("[probing] (Step %d) Setting rating %d as output (cexpr - failed precondition)!\n", step, -50);
						rating = -50;
					}

					while (composite_stack->size() > stack_size) {
						if (debug_planning_composite) printf("[composite] (Step %d) Popping \"%s\" from stack! (stack size is %lu, going down to %d)\n",
								step, composite_stack->top().print().c_str(), composite_stack->size(), stack_size);
						StackElem elem = composite_stack->top();
						composite_stack->pop();

						if (elem.elem_type == END_PARALLEL) {
							if (!add_database_stack->empty())
								add_database_stack->pop();
							if (!remove_database_stack->empty())
								remove_database_stack->pop();
						}
					}
				}
			}

			// no luck this time, let's put the elem back
			composite_stack->push(elem);
			if (elem.elem_type == START_PARALLEL) {
				if (debug_planning_composite) printf("[composite] (Step %d) Putting back START_PARALLEL!\n", step);

				add_database_stack->pop();
				remove_database_stack->pop();
			}
			if (elem.elem_type == RESET_ENV_PARALLEL) {
				if (debug_planning_composite) printf("[composite] (Step %d) Putting back RESET_ENV_PARALLEL!\n", step);

				add_database_stack->pop();
				remove_database_stack->pop();

				add_database_stack->push(new database());
				remove_database_stack->push(new database());
			}
			if (elem.elem_type == END_PARALLEL) {
				if (debug_planning_composite) printf("[composite] (Step %d) Putting back END_PARALLEL!\n", step);

				add_database_stack->push(new database());
				remove_database_stack->push(new database());
			}

			if (debug_planning_composite) printf("[composite] (Step %d) Finished execution without success - putting elem \"%s\" back!!\n", step, elem.print().c_str());
		}
		return false;
	}

	ModuleGeneric* ModuleProblem::select_action(std::vector<ModuleGeneric*>* actions, std::vector<ModuleGeneric*>* composite) {
		std::vector<ModuleGeneric*>* action_list = nullptr;

		if (composite_first && !composite->empty()) {
			action_list = composite;
		} else {
			action_list = actions;
		}

		if(action_list->empty())
			return nullptr;

		int toReturnIndex = 0; // first element
		if (random) {
			toReturnIndex = rand() % action_list->size();
		}

		ModuleGeneric* toReturn = action_list->at(toReturnIndex);
		action_list->erase(action_list->begin() + toReturnIndex);

		return toReturn;
	}

	std::vector<Value *>* ModuleProblem::select_valuation(std::vector<std::vector<Value *>*>* valuations) {
		if(valuations->empty())
			return nullptr;

		int toReturnIndex = 0; // first element
		if (random) {
			toReturnIndex = rand() % valuations->size();
		}

		std::vector<Value *>* toReturn = valuations->at(toReturnIndex);
		valuations->erase(valuations->begin() + toReturnIndex);

		return toReturn;
	}

	ModuleGeneric* ModuleProblem::select_generic_module(std::vector<ModuleGeneric*>* modules) {
		if(modules->empty())
			return nullptr;

		int toReturnIndex = 0; // first element
		if (random) {
			toReturnIndex = rand() % modules->size();
		}

		ModuleGeneric* toReturn = modules->at(toReturnIndex);
		modules->erase(modules->begin() + toReturnIndex);

		return toReturn;
	}

	bool ModuleProblem::find_and_apply_action(int step, std::vector<std::string>& plan, int& rating) {
		std::vector<ModuleGeneric*>* global_modules = Application::app_instance->public_get_modules();
		std::vector<ModuleGeneric*>* global_actions = Application::app_instance->public_get_actions();
		bool debug_planning = Application::app_instance->debug_planning;
		bool debug_probing = Application::app_instance->debug_probing;
		bool debug_warning = Application::app_instance->debug_warning;
		bool debug_planning_composite = Application::app_instance->debug_planning_composite;

		if (debug_planning) printf("======================= Planning step %d =======================\n", step);

		bool check_rating_only = false;
		if (rating >= 0)
			check_rating_only = true;

		int max_rating = -100;
		for (ModuleGeneric* belief: *belief_modules) {
			if (check_rating_only) {
				if (debug_probing) printf("(Step %d) Plan execution - goal rating in belief %s!\n", step, belief->get_name().c_str());
			} else  {
				if (debug_planning) printf("(Step %d) Plan execution - goal check start in belief %s!\n", step, belief->get_name().c_str());
			}

			int rating_local = rating;
			Result goal_result = evaluate_expression_global(this->goal, belief, global_modules, belief->get_domains(), rating_local);

			if (check_rating_only) {
				if (goal_result.get_variable_names().empty() && goal_result.get_valuation().get_first_logic_value() == logic_type::TRUE) {
					// this action fills the goal, it has to get huuuge rating!
					max_rating = std::max(max_rating, 999999);
				} else {
					max_rating = std::max(max_rating, rating_local);
				}

				if (debug_probing) printf("(Step %d) Plan execution - goal rating in belief %s returned %d (current max: %d)!\n",
						step, belief->get_name().c_str(), rating_local, max_rating);
			} else {
				if (debug_planning) printf("(Step %d) Plan execution - goal result: %s\n", step, goal_result.print().c_str());

				if (goal_result.get_variable_names().empty() && goal_result.get_valuation().get_first_logic_value() == logic_type::TRUE)
					return true;
			}
		}

		if (check_rating_only) {
			if (debug_probing) printf("(Step %d) Plan execution - goal rating returning %d!\n", step, max_rating);
			rating = max_rating;
			return false;
		}


		if (step >= max_depth) {
			if (debug_planning) printf("(Step %d) Warning: planning reached max allowed recursion depth %d\n", step, max_depth);
			return false;
		}

		std::multiset<ActionEntry*, action_entry_compare> action_sets;

		std::vector<ModuleGeneric*>* belief_module_list = new std::vector<ModuleGeneric*>();
		for(auto belief : *belief_modules) {
			belief_module_list->push_back(belief);
		}

		// let's check each belief module
		while (ModuleGeneric* belief = select_generic_module(belief_module_list)) {

			if (debug_planning) printf("(Step %d) Plan execution - checking belief module %s\n", step, belief->get_name().c_str());
			ModuleBelief* bm = dynamic_cast<ModuleBelief*>(belief);
			std::vector<ModuleGeneric*>* members = bm->get_member_modules();

			std::vector<ModuleGeneric*>* member_module_list = new std::vector<ModuleGeneric*>();
			for(auto member : *members) {
				member_module_list->push_back(member);
			}

			// check each member module...
			while (ModuleGeneric* member = select_generic_module(member_module_list)) {
				// First, lets check the planning goal!

				if (member->get_module_type() != Inter4ql::module_type::LOCAL && member->get_module_type() != Inter4ql::module_type::GROUP)
					continue;

				if (debug_planning) printf("(Step %d) Plan execution - checking world module %s\n", step, member->get_name().c_str());

				std::vector<ModuleGeneric*>* composite_actions = new std::vector<ModuleGeneric*>();
				std::vector<ModuleGeneric*>* actions = new std::vector<ModuleGeneric*>();

				// fills_goal_first
				// max of ORs, za "true" 1 pkt

				for(auto action : *action_modules) {
					ModuleAction* am = dynamic_cast<ModuleAction*>(action);

					if (am->is_composite()) {
						if (composite_first) {
							composite_actions->push_back(am);
						} else {
							actions->push_back(am);
						}
					} else {
						actions->push_back(am);
					}
				}

				while (ModuleGeneric* action = select_action(actions, composite_actions)) {
					ModuleAction* am = dynamic_cast<ModuleAction*>(action);
					if (debug_planning) printf("(Step %d) Checking action \"%s\":\n\n", step, am->get_name().c_str());

					int rating; // not used here...
					Result precondition_result;
					try {
						precondition_result = evaluate_expression_global(am->get_preconditions(), member, global_modules, member->get_domains(), rating);
					} catch (Inter4ql::Exception *s) {
						if (debug_warning) {
							printf("WARNING in step %d at preconditions check for action %s in world %s: %s", step,  am->get_name().c_str(), member->get_name().c_str(), s->get().c_str());
						}
						// do nothing, missing relation is fine here...
					}

					if (debug_planning) printf("(Step %d) Plan execution - precondition result for action \"%s\":\n%s\n", step, am->get_name().c_str(), precondition_result.print().c_str());

					const auto& valuation = precondition_result.get_valuation();
					const std::vector<std::string>& vn =  precondition_result.get_variable_names();

					std::vector<std::vector<Value *>*>* true_valuations = new std::vector<std::vector<Value *>*>();

					for(const auto& val : valuation) {
						if (val.second != Inter4ql::logic_type::TRUE)
							continue;

						true_valuations->push_back(new std::vector<Value *>(val.first.begin(), val.first.end()));
					}

					if (!true_valuations->empty()) {
						std::vector<Value*>* action_arguments = am->get_args();
						for (auto name : *action_arguments) {
							if(std::find(vn.begin(), vn.end(), name->get_value_string()) == vn.end()) {
								std::string error_msg = "Precondition of action '" + am->get_name() + "' does not assign value to variable '"
										+ name->get_value_string() + "' which is required by action's argument\n";
								throw new Exception(error_msg);
							}
						}
					}

					while(std::vector<Value *>* val = select_valuation(true_valuations)) {
						if (goal_probing) {
							std::map<std::string, Value*> binding;

							for(int i = 0; i < vn.size(); i++) {
								binding.insert({vn.at(i).c_str(), val->at(i)});
							}

							database* to_add_snapshot = make_snapshot(member->get_add_database());
							database* to_remove_snapshot = make_snapshot(member->get_remove_database());

							int rating_local = 50;
							if (!am->is_composite()) {
								process_non_composite(member, am, binding, debug_planning);

								// if applying action violated some constraints then do not probe
								if (!maybe_violated_constraints(belief, add_database_stack->size())) {
									find_and_apply_action(step + 1, plan, rating_local);
								} else {
									rating_local = -100;
								}
							} else {
								composite_stack->push(StackElem(EXPRESSION, am->get_composite_expr(), &binding, am->get_name(), 0));

								// if we have not probed composite actions yet in this
								if (composite_ratings.find(step) == composite_ratings.end()) {
									composite_ratings.insert(std::make_pair(step, std::map<std::string, int>()));
								}

								std::string what_in_stack = composite_stack->top().print();
								assert(composite_stack->size() == 1);
								assert(add_database_stack->size() == 0);

								// if applying action violated some constraints then do not continue
								if (!maybe_violated_constraints(belief, add_database_stack->size())) {
										traverse_composite_expr(composite_stack, add_database_stack,
													remove_database_stack, am, member, belief,
													action_modules, step, 1, plan, rating_local, 0, "");

								} else {
									rating_local = -100;
								}

								assert(add_database_stack->size() == 0);

								assert(composite_stack->size() == 1);
								assert(composite_stack->top().print() == what_in_stack);
								composite_stack->pop();
							}

							if (rating_local >= 0) {
								if (debug_probing) {
									printf("(Step %d) Adding action \"%s\" for belief \"%s\" and world \"%s\" is executable with rating %d and values: ", step,
											am->get_name().c_str(), belief->get_name().c_str(), member->get_name().c_str(), rating_local);
									for(int i = 0; i < vn.size(); i++) {
										printf("%s: %s, ", vn.at(i).c_str(), val->at(i)->print(true).c_str());
									}
									printf("\n");
								}

								std::vector<std::string>* variable_names = new std::vector<std::string>(vn.begin(), vn.end());
								action_sets.insert(new ActionEntry(belief, member, am, val, variable_names, rating_local));
							}

							// TODO: memory leak!!!
							member->set_add_database(to_add_snapshot);
							member->set_remove_database(to_remove_snapshot);
						} else {
							if (debug_probing | debug_planning) {
								printf("(Step %d) Adding action \"%s\" for belief \"%s\" and world \"%s\" is executable with rank 0 and values: ", step,
										am->get_name().c_str(), belief->get_name().c_str(), member->get_name().c_str());
								for(int i = 0; i < vn.size(); i++) {
									printf("%s: %s, ", vn.at(i).c_str(), val->at(i)->print(true).c_str());
								}
								printf("\n");
							}

							std::vector<std::string>* variable_names = new std::vector<std::string>(vn.begin(), vn.end());
							action_sets.insert(new ActionEntry(belief, member, am, val, variable_names, 0));
						}
					}
				}

				delete composite_actions;
				delete actions;
			}
			delete member_module_list;
		}
		delete belief_module_list;

		//printf("Plan execution - checking member %s from belief module %s\n", member->get_name().c_str(), belief->get_name().c_str());

		while (!action_sets.empty()) {
			ActionEntry* entry = *(action_sets.begin());
			action_sets.erase(action_sets.begin());

			if (rating < 0 && increase_rating_only && entry->get_rating() < selected_rating->top()) {
				continue;
			}
			if (rating < 0 && strict_increase_rating_only && entry->get_rating() <= selected_rating->top()) {
				continue;
			}

			ModuleGeneric* belief = entry->get_belief();
			ModuleGeneric* member = entry->get_world();
			ModuleAction* am = entry->get_action();
			std::vector<std::string>* vn = entry->get_variable_names();
			std::vector<Value *>* val = entry->get_valuation();

			std::map<std::string, Value*> binding;

			for(int i = 0; i < vn->size(); i++) {
				binding.insert({vn->at(i).c_str(), val->at(i)});
			}

			if (debug_planning | debug_probing) {
				printf("(Step %d) Action \"%s\" is executable with values: ", step, am->get_name().c_str());
				for(int i = 0; i < vn->size(); i++) {
					printf("%s: %s, ", vn->at(i).c_str(), val->at(i)->print(true).c_str());
				}
				printf("\n");
			}

			// We have TRUE valuation - let's apply it
			database* to_add_snapshot = make_snapshot(member->get_add_database());
			database* to_remove_snapshot = make_snapshot(member->get_remove_database());

			selected_rating->push(entry->get_rating());
			int selected_rating_size = selected_rating->size();

			bool should_return = false;
			if (!am->is_composite()) {
				process_non_composite(member, am, binding, debug_planning);

				int rating_local = -1;

				// if applying action violated some constraints then do not continue
				if (!maybe_violated_constraints(belief, add_database_stack->size())
						&& find_and_apply_action(step + 1, plan, rating_local)) {
					std::vector<Value*>* action_arguments = am->get_args();

					if (debug_planning) {
						printf("(Step %d) (planning) Planning success! %s(", step, am->get_name().c_str());
						for(int i = 0; i < vn->size(); i++) {
							bool is_argument = false;
							for (auto name : *action_arguments) {
								if (name->get_value_string() == vn->at(i).c_str())
									is_argument = true;
							}
							if (!is_argument)
								continue;
							printf("%s: %s, ", vn->at(i).c_str(), val->at(i)->print(true).c_str());
						}
						printf(")\n");
					}

					std::stringstream ss;
					ss << am->get_name().c_str() << "(";
					for(int i = 0; i < vn->size(); i++) {
						bool is_argument = false;
						for (auto name : *action_arguments) {
							if (name->get_value_string() == vn->at(i).c_str())
								is_argument = true;
						}
						if (!is_argument)
							continue;
						ss << vn->at(i).c_str() << ": " << val->at(i)->print(true).c_str();
						if (i < vn->size() - 1)
							ss << ", ";
					}
					ss << ")";
					plan.push_back(ss.str());

					if (composite_ratings.find(step) != composite_ratings.end()) {
						composite_ratings.erase(composite_ratings.find(step));
					}

					should_return = true;
				}
			} else {
				composite_stack->push(StackElem(EXPRESSION, am->get_composite_expr(), &binding, am->get_name(), 0));
				if (debug_planning_composite)
					printf("(Step %d) [composite] START composite! \"%s\" put on stack!\n", step, composite_stack->top().print().c_str());

				std::string what_in_stack = composite_stack->top().print();
				assert(composite_stack->size() == 1);
				assert(add_database_stack->size() == 0);

				// if applying action violated some constraints then do not continue
				if (!maybe_violated_constraints(belief, add_database_stack->size()) &&
						traverse_composite_expr(composite_stack, add_database_stack,
									remove_database_stack, am, member, belief,
									action_modules, step, 1, plan, rating,
									entry->get_rating(), "")) {
					std::vector<Value*>* action_arguments = am->get_args();

					if (debug_planning) {
						printf("(Step %d) (planning) Planning success! %s(", step, am->get_name().c_str());
						for(int i = 0; i < vn->size(); i++) {
							bool is_argument = false;
							for (auto name : *action_arguments) {
								if (name->get_value_string() == vn->at(i).c_str())
									is_argument = true;
							}
							if (!is_argument)
								continue;
							printf("%s: %s, ", vn->at(i).c_str(), val->at(i)->print(true).c_str());
						}
						printf(")\n");
					}

					std::stringstream ss;
					ss << am->get_name().c_str() << "(";
					for(int i = 0; i < vn->size(); i++) {
						bool is_argument = false;
						for (auto name : *action_arguments) {
							if (name->get_value_string() == vn->at(i).c_str())
								is_argument = true;
						}
						if (!is_argument)
							continue;

						ss << vn->at(i).c_str() << ": " << val->at(i)->print(true).c_str();
						if (i < vn->size() - 1)
							ss << ", ";
					}
					ss << "):";
					plan.push_back(ss.str());

					if (composite_ratings.find(step) != composite_ratings.end()) {
						composite_ratings.erase(composite_ratings.find(step));
					}

					should_return = true;
				}

				assert(add_database_stack->size() == 0);

				if (!should_return) {
					if (debug_planning_composite) printf("(Step %d) [composite] After exit stack size is %lu\n", step, composite_stack->size());
					assert(composite_stack->size() == 1);
					assert(composite_stack->top().print() == what_in_stack);
					if (debug_planning_composite) printf("(Step %d) [composite] Value on stack \"%s\" after exit!\n", step, composite_stack->top().print().c_str());
					composite_stack->pop();
				}
			}

			assert(selected_rating->size() == selected_rating_size);
			selected_rating->pop();

			// TODO: memory leak!!!
			member->set_add_database(to_add_snapshot);
			member->set_remove_database(to_remove_snapshot);

			delete vn;
			delete val;

			if (should_return) return true;
		}

		if (composite_ratings.find(step) != composite_ratings.end()) {
			composite_ratings.erase(composite_ratings.find(step));
		}

		return false;
	}

	void ModuleProblem::reason()
	{
		// In fact no reasoning here... Just planning...
		std::vector<std::string> plan;

		std::srand (time(NULL));

		auto start = std::chrono::high_resolution_clock::now();

		assert(add_database_stack->size() == 0);
		assert(composite_ratings.size() == 0);

		// push some small value...
		selected_rating->push(-200);

		assert(selected_rating->size() == 1);

		// Why step = 0 and not 1? Because checking goal is done in
		// additional step.
		int rating = -1;
		if (find_and_apply_action(/* step = */ 0, plan, rating)) {
			auto stop = std::chrono::high_resolution_clock::now();
			auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(stop - start);

			printf("\nPlan found:\n");
			int step = 1;
			for(auto it = plan.end() - 1; it >= plan.begin(); it--) {
				if (it->c_str()[0] == ' ') {
					printf("#> %s\n", it->c_str());
					continue;
				}
				printf("%d. %s\n", step++, it->c_str());
			}

			std::cout << std::endl << "Planning took " << duration.count() << " milliseconds." << std::endl;;
		} else {
			auto stop = std::chrono::high_resolution_clock::now();
			auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(stop - start);
			std::cout << "\nNo plan found (planning took " << duration.count() << " milliseconds)." << std::endl;
		}

		assert(selected_rating->size() == 1);
		selected_rating->pop();

		assert(add_database_stack->size() == 0);
		assert(composite_ratings.size() == 0);
	}

	std::vector<std::string>* ModuleProblem::get_beliefs() {
		return this->beliefs;
	}

	std::vector<std::string>* ModuleProblem::get_actions() {
		return this->actions;
	}

	Expression* ModuleProblem::get_goal() {
		return this->goal;
	}

	int ModuleProblem::get_max_depth() {
		return this->max_depth;
	}

	Result ModuleProblem::ask_for_result(Fact * term) const
	{
		return Result();
	}

	void ModuleProblem::add_belief_module(ModuleGeneric* module) {
		if (module->get_module_type() != Inter4ql::module_type::BELIEF) {
			std::string error_msg = "Tried to add a non-belief module '" + module->get_name() + "' to a planning problem. This is currently not allowed.\n";
			throw new Exception(error_msg);
		}

		belief_modules->push_back(module);
	}

	std::vector<ModuleGeneric*>* ModuleProblem::get_belief_modules() {
		return belief_modules;
	}

	void ModuleProblem::add_action_module(ModuleGeneric* module) {
		action_modules->push_back(module);
	}

	bool ModuleProblem::contains_action(std::string action_name) {
		return std::find(actions->begin(), actions->end(), action_name) != actions->end();
	}

	std::vector<ModuleGeneric*>* ModuleProblem::get_action_modules() {
		return action_modules;
	}

	void ModuleProblem::replace_actions(std::vector<std::string>* actions) {
		delete this->actions;
		this->actions = actions;
	}
}
